/* @(#) All Right Reserved, Copyright (c) 1997 Australian National University */
/* @(#)                     Copyright (c) 1997 Fujitsu Laboratories Ltd. */
/*  */
/*        */
/*                        Permission Notice */
/*  */
/*                Australian National University */
/*                Canberra ACT 0200 Australia */
/*                           and */
/*                Fujitsu Laboratories Ltd. */
/*                4-1-1 Kamikodanaka Nakahara */
/*                Kawasaki 211-88 Japan */
/*  */
/*   Permission to use, copy, modify and distribute this software (including  */
/*   its documentation) without fee is hereby granted under the following terms */
/*   and conditions: */
/*  */
/*   1. This permission notice and all of the copyright notices in this */
/*      software must appear in all copies and derived versions of all */
/*      or a part of this software. */
/*  */
/*   2. Sales of this software, whether incorporated in other products as a */
/*      part thereof or not, is prohibited. */
/*  */
/*   3. On publication or presentation in any form of the results obtained */
/*      by use of this software, an explicit reference to this software and */
/*      its copyright owner shall be made. */
/*  */
/*   4. THIS SOFTWARE IS PROVIDED 'AS IS' AND THE AUSTRALIAN NATIONAL UNIVERSITY */
/*      AND FUJITSU LABORATORIES LTD. MAKE NO WARRANTIES, EITHER EXPRESSED OR */
/*      IMPLIED, AS TO ANY MATTER WHATSOEVER REGARDING THIS SOFTWARE,  */
/*      INCLUDING, WITHOUT LIMITATION, WARRANTIES OF MERCHANTABILITY OR FITNESS */
/*      FOR ANY PARTICULAR PURPOSE. */
/*      IN NO EVENT SHALL THE AUSTRALIAN NATIONAL UNIVERSITY OR FUJITSU */
/*      LABORATORIES LTD. BE LIABLE FOR ANY CLAIM FROM A THIRD PARTY, OR */
/*      FOR SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES WHATSOEVER RESULTING */
/*      FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, */
/*      NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION */
/*      WITH THE USE, COPYING OR MODIFICATION OF THIS SOFTWARE. */
/*  */

/************************************************************************/
/*        FDATAAuxMatrix.c: FDATA Auxillary Matrix procedures           */
/************************************************************************/
/* SCCS INFO @(#)FDATAAuxMatrix.c 3.4         last modified P. Strazdins, 96/12/17
   first created: P. Strazdins 91/12/12
 */
/* set of matrix routines for matrix initialiation and result checking;
 * not called by DBLAS routines. 
 * compilation: macro FDATATYPE must be defined;
 *		if TRACE_ERRS defined, MatDiff() prints indices where error
 *		detected
 */ 
#include "FDATASetUp.h"
#include "FDATAOps.h"
#include "LocMat.h"
#include "stdlib.h"
#include "FDATACellMatrix.h"

#define TRACE_ERRS

/****************** matrix initialization & checking procedures *************/

#define F_RAND_MOD	10000
#define F_RAND		(((FD_BASE) ((rand() % F_RAND_MOD) - F_RAND_MOD/2)) \
			/((FD_BASE) F_RAND_MOD/2))	/* values in [-1..1] */

static void MatInit(RandElts, IMult, JMult, Offs, M, N, A)
  int RandElts, M, N; FDATA IMult, JMult, Offs; LocMat A;
/* pre:   LM_IncOK(M, N, A)  
 * post:  A[i,j] = (i*IMult + j*JMult + Offs), for 0 <= i < M and 0 <= j < N
 *	  where f(x) = x, !RandElts and f(x) = {srand(x); F_RAND}  ow.
 */
{ int i, j;
  for (i=0; i < M; i++)
    if (RandElts)
      for (j=0; j < N; j++) {
      	srand((int) (i*IMult + j*JMult + Offs)*F_RAND_MOD); 
        *LM_Index(A, i, j) = F_RAND;
      }
    else
      for (j=0; j < N; j++) 
	*LM_Index(A, i, j) = i*IMult + j*JMult + Offs;        
} /* MatInit() */
 

#define EPSILON			1.0E-10	/* macine precision */		
 
static int MatCheck(RandElts, IMult, JMult, Offs, M, N, A)
  int RandElts, M, N; FDATA IMult, JMult, Offs; LocMat A;
/* pre:   LM_IncOK(M, N, A)  
 * post: returns number of elements of M x N A differing from A' where
 *	 A' satisfies postcond. MatInit(RandElts, IMult, JMult, Offs, M, N, A') 
 */
{ int i, j, nerrs=0; FDATA e, v;
  for (i=0; i < M; i++) {
    int nej=0;	
    for (j=0; j < N; j++) {
      if (RandElts) {
      	srand((int) (i*IMult + j*JMult + Offs)*F_RAND_MOD);
        v = F_RAND;
      } else
        v = i*IMult + j*JMult + Offs;
      e = v - *LM_Index(A, i, j);	if (e < 0.0) e = -e;
      if (e > EPSILON) {
      	/* a threshold check required since the value of Offs for MatCheck()
      	   may not necessarily be that for MatInit() of the source matrix,
      	   eg. B[i, j] in MatCheck with Offs'=inc+o 
      	   comes from A[i, j+inc] produced by MatInit() with Offs = o
	*/
#ifdef TRACE_ERRS
	printf("(%d,%d: %.1f#%.1f) ", i, j, *LM_Index(A, i, j), v) ;
#endif    	
        nej++;
      } 
    } /* for (j...) */
#ifdef TRACE_ERRS
    if (nej)
      printf("\n");
    nerrs += nej;  
#endif    
  } /* for (i...) */ 
  return (nerrs);       
} /* MatCheck() */
 

static int MatDiff(M, N, A, B, MxDiff, AvDiff, MaxI, MaxJ)
  int M, N; LocMat A, B; FD_BASE *MxDiff, *AvDiff; int *MaxI, *MaxJ;
/* pre:  LM_IncOK(M, N, A) && LM_IncOK(M, N, B) 
 * post: *MxDiff = Max {0} U {0<=i<M, 0<=j<N: D(i,j)}
 * 	 *AvDiff = Sum {0<=i<M, 0<=j<N: D(i,j)}/M/N  , if M>0 & N>0
 *	  	 = 0					otherwise
 * 	  D(*MaxI,*MaxJ) = *MxDiff		      , if M>0 & N>0
 * 	  where D(i,j) = Abs(A[i,j]-B[i,j])/(Abs(A[i,j])+Abs(B[i,j])+1E-30)
 *	  returns the number of elements st D(i,j) > PRECISION_THRESH 
 */
# define PRECISION_THRESH 1.0E-7                /*set t to 2x precision limit */
{ int i, j, n_err=0; FD_BASE t, d; FDATA del;
  int errjmn;
  *MxDiff = 0.0; *MaxI = 0; *MaxJ = 0;
  *AvDiff = 0.0;
  for (i = 0; i < M; i++) {
#ifdef TRACE_ERRS
    int errjmn=-1, errjmax;
#endif    
    for (j = 0; j < N; j++) {
      FDATA Aij = *LM_Index(A,i,j), Bij = *LM_Index(B,i,j);
      FD_SET(del, =, Aij); FD_SET(del, -=, Bij);
      d = FD_ABS(del);
      t = PRECISION_THRESH + FD_ABS(Aij) + FD_ABS(Bij); 
      d = d / t;
      if (d > *MxDiff) { *MxDiff = d; *MaxI = i; *MaxJ = j;}
      if (d > PRECISION_THRESH) {
        n_err++;
#ifdef TRACE_ERRS
	if (errjmn<0) 
          errjmn = j;
        errjmax = j;  
#endif                        	  
      }
      *AvDiff = *AvDiff + d;
    } /* for (j...) */
#ifdef TRACE_ERRS    
   if (errjmn>0) 
      printf(" e[%d,%d..%d] ", i, errjmn, errjmax);    
#endif
  } /* for (i...) */
  if (M > 0 && N > 0) *AvDiff = *AvDiff / ((FD_BASE) M * N);
  return (n_err);
} /* MatDiff() */
	
