/* @(#) All Right Reserved, Copyright (c) 1997 Australian National University */
/* @(#)                     Copyright (c) 1997 Fujitsu Laboratories Ltd. */
/*  */
/*        */
/*                        Permission Notice */
/*  */
/*                Australian National University */
/*                Canberra ACT 0200 Australia */
/*                           and */
/*                Fujitsu Laboratories Ltd. */
/*                4-1-1 Kamikodanaka Nakahara */
/*                Kawasaki 211-88 Japan */
/*  */
/*   Permission to use, copy, modify and distribute this software (including  */
/*   its documentation) without fee is hereby granted under the following terms */
/*   and conditions: */
/*  */
/*   1. This permission notice and all of the copyright notices in this */
/*      software must appear in all copies and derived versions of all */
/*      or a part of this software. */
/*  */
/*   2. Sales of this software, whether incorporated in other products as a */
/*      part thereof or not, is prohibited. */
/*  */
/*   3. On publication or presentation in any form of the results obtained */
/*      by use of this software, an explicit reference to this software and */
/*      its copyright owner shall be made. */
/*  */
/*   4. THIS SOFTWARE IS PROVIDED 'AS IS' AND THE AUSTRALIAN NATIONAL UNIVERSITY */
/*      AND FUJITSU LABORATORIES LTD. MAKE NO WARRANTIES, EITHER EXPRESSED OR */
/*      IMPLIED, AS TO ANY MATTER WHATSOEVER REGARDING THIS SOFTWARE,  */
/*      INCLUDING, WITHOUT LIMITATION, WARRANTIES OF MERCHANTABILITY OR FITNESS */
/*      FOR ANY PARTICULAR PURPOSE. */
/*      IN NO EVENT SHALL THE AUSTRALIAN NATIONAL UNIVERSITY OR FUJITSU */
/*      LABORATORIES LTD. BE LIABLE FOR ANY CLAIM FROM A THIRD PARTY, OR */
/*      FOR SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES WHATSOEVER RESULTING */
/*      FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, */
/*      NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION */
/*      WITH THE USE, COPYING OR MODIFICATION OF THIS SOFTWARE. */
/*  */

/************************************************************************/
/*                              GMM.cell.c                              */
/************************************************************************/
/* SCCS INFO @(#)GMM.cell.c 3.5         last modified P. Strazdins, 97/05/23
   first created: P. Strazdins 93/01/20
*/
/* cell program for testing Ny x Nx Distributed Matrix Multiply:
 * (GMM.Parameters.c sets the varibales from the command line parameters)
 *       C = alpha * A' * B' + beta * C,
 *           A' = (TransA==NoTranspose? A: A^T) is m x k
 *           B' = (TransB==NoTranspose? B: B^T) is k x n
 * if the DoubleMult option is specified, it performs (for FormC != Rectangular):
 *       C = alpha * A' * B' +  alpha * (A')^T * (B')^T + beta * C,
 *	(provided TransA # TransB)
 * for i=0..NReps-1; where
 *	m = M+i*StepM; n = N+i*StepN; k = K+i*StepK
 *      A,B,C are the respective sub-matrices from positions
 *      (AI0,AJ0), (BI0,BJ0), (CI0,CJ0) of matrices A0, B0 & C0.
 * The global matrices A0 is MA x NA, B0 is MB x NB, C0 is MC x NC;
 * these are all stored in scattered storage across cells
 * (with element (0,0) in cell (0,0)), where:
 *      (MA,NA) = TransA=NoTranspose? (maxm+AI0,maxk+AJ0): (maxk+AI0,maxm+AJ0)
 *      (MB,NB) = TransB=NoTranspose? (maxk+BI0,maxn+Bj0): (maxn+BI0,maxk+BJ0)
 *      (MC,NC) = (manm+CI0,maxn+CJ0)
 *      maxm = StepM<=0? M: M+(NReps-1)StepM
 * and BM is block size across M dimension, and similarly for BN and BK.
 * (externally-defined) cpp FLAGs:
 *      FDATATYPE       SINGLE (default), DOUBLE
 */
/************************************************************************/
#define SYRK_SYNC 8		/* synchronize on each SYRK_SYNC reps for 
				   SYR(2)K; otherwise may get dramatic slowdowns
				   on square processor configurations as diag.
				   cells get too far ahead of the others */

#include "Cell.h"
#include "Timer.c"

#include "BLASOptions.h"
#include "FDATASetUp.h"
#include "FDATAOps.h"
#include "LocMat.h"
#include "MatDim.h"
#include "DistMat.h"
#include "FDATAMalloc.h"
#include "FDATACellMatrix.h"
#include "FDATACellTriMatrix.h"
#include "FDATACommun.h"
#include "D_BLAS.h"

#include "HostMsg.c"
#define PROGNAME "GMM.cell"
#include "GMM.Parameters.c"

/************ scattered matrix initialization /checking routines ************/

static void SetLinearScattered(ry, rx, my, mx, M, N, beta, A)
  DistMat A; FDATA *beta; FD_BASE *ry, *rx; int my, mx, M, N;
/* pre:  A is an M x N distributed matrix with block & cell offset (0,0)
 * post: (global) A[i,j] = ry*(i%my) + rx*(j%mx) for 0<=i<M, 0<=j<N
 *       ie. (local) A[i',j'] = ry*(L2Gd(i', DM_D0(A))%my) +
 *		                rx*(L2Gd(j', DM_D1(A))%mx) 
 *	 for 0<=i<G2Ld(M, DM_D0(A)), 0<=j<G2Ld(N, DM_D1(A))%mx)
 */
#define lA DM_Loc(A)
{ register FD_BASE ryb = *ry * *beta, 	rxb = *rx * *beta,   
		   Ai0;
  int i, j; 
  M = G2Ld(M, DM_D0(A)); 		N = G2Ld(N, DM_D1(A));
  if (M > 0) {
    for (j=0; j<N; j++)
      FD_SET_REAL(*LM_Index(lA, 0, j), rxb * (FD_BASE) (L2Gd(j, DM_D1(A))%mx));
    for (i=1; i<M; i++) { 
      FD_SET_REAL(Ai0,  ryb * (FD_BASE) (L2Gd(i, DM_D0(A)) % my));
      for (j=0; j<N; j++) 
        FD_SETOP(*LM_Index(lA, i, j), Ai0, +, *LM_Index(lA, 0, j));
    }
    FD_SET_REAL(Ai0,  ryb * (FD_BASE) (L2Gd(0, DM_D0(A)) % my));
    for (j=0; j<N; j++) 
      FD_SETOP(*LM_Index(lA, 0, j), Ai0, +, *LM_Index(lA, 0, j));
  }      
} /* SetLinearScattered() */
#undef lA


static int ModSumK(K, K0, m)
/* pre:  K0, K>=0; m>0  
 * post: returns sum: 0 <= k < K: (k + K0)%m 
 */
{ int k, s=0, km = K0%m;
  for (k=0; k < K; k++) {
   s += km;
   km++;
   if (km == m) 
     km = 0;
  }
  return (s);
} /* ModSumK() */

static double ModSumK2(K, K0y, K0x, my, mx)
/* pre:  K0y, K0x, K>=0; my, mx>0  
 * post: returns sum: 0 <= k < K: ((k + K0y)%my) * ((k + K0x)%mx) 
 */
{ int k; double s=0.0;  int kmx = K0x%mx, kmy = K0y%my;
  for (k=0; k < K; k++) {
    s += (double) kmy * (double) kmx;
    kmx++;
    if (kmx == mx) 
      kmx = 0; 
    kmy++;    
    if (kmy == my) 
      kmy = 0; 
  }
  return (s);
} /* ModSumK2()*/


static int CheckLinearScatteredMult(
	ry, rx, my, mx, DoubleMult, TransA, TransB, FormC, 
	M, N, K, alpha, AI0, AJ0, BI0, BJ0, beta, CI0, CJ0, C,
 	MxDiff, AvDiff, MaxI, MaxJ
)
  FD_BASE *ry, *rx; 
  int my, mx, DoubleMult, M, N, K, AI0, AJ0, BI0, BJ0, CI0, CJ0; 
  int TransA, TransB; char FormC;
  FDATA *alpha, *beta; DistMat C;
  FD_BASE *MxDiff, *AvDiff; int *MaxI, *MaxJ;
/* pre:  0 <= AI0, BI0, CI0; 0 <= AJ0, BJ0, CJ0
 *	the (i,j)th element of A = (*ry*((i+AI0)%my)+*rx*((j+AJ0)%mx)) 
 *	& similarly for B and C
 * post:sets C = (*beta)C + (*alpha)A'B' 
 *	where global A' = (TransA==NoTranspose? A: A^T) is M x K
 *		     B' = (TransA==NoTranspose? B: B^T) is K x N
 *	returns the number of local elements where C != C_pre
 *	with the maximum difference being *MxDiff at (*MaxI, *MaxJ)
 *	and the average difference being *AvDiff
 */
{ int i, j; 
  FDATA *jBx, *jAx, *jC0;			
  int NErrs;
  FD_BASE D00, RInc, CInc, RCInc, D002, RInc2, CInc2, RCInc2;
  FD_BASE RyA, RxA, RyB, RxB;
  int myA, mxA, myB, mxB;

  if (!TransA) { 
    RyA = *ry; RxA = *rx; myA = my; mxA = mx;
  } else { 
    int t=AI0; AI0 = AJ0; AJ0 = t; 
    RyA = *rx; RxA = *ry; myA = mx; mxA = my;
  }  
  if (!TransB) { 
    RyB = *ry; RxB = *rx; myB = my; mxB = mx; 
  } else {
    int t=BI0; BI0 = BJ0; BJ0 = t; 
    RyB = *rx; RxB = *ry; myB = mx; mxB = my;
  }
  D00 =   *alpha * RxA * RyB * (FD_BASE) ModSumK2(K, AJ0, BI0, mxA, myB);
  RInc =  *alpha * RyA * RyB * (FD_BASE) ModSumK(K, BI0, myB);
  CInc =  *alpha * RxA * RxB * (FD_BASE) ModSumK(K, AJ0, mxA);
  RCInc = *alpha * RyA * RxB * (FD_BASE) K;  
/*printf("D00=%f, RInc=%f, CInc=%f, RCInc=%f\n", D00, RInc, CInc, RCInc);*/
  if (DoubleMult) {	
    D002 =   *alpha * RyB * RxA * (FD_BASE) ModSumK2(K, BI0, AJ0, mxB, myA);
    RInc2 =  *alpha * RxB * RxA * (FD_BASE) ModSumK(K, AJ0, myA);
    CInc2 =  *alpha * RyB * RyA * (FD_BASE) ModSumK(K, BI0, mxB);
    RCInc2 = *alpha * RxB * RyA * (FD_BASE) K;  
/*printf("2: D00=%f, RInc=%f, CInc=%f, RCInc=%f\n", D002, RInc2, CInc2, RCInc2);*/
  }

  M = G2Ld(M, DM_D0(C));		N = G2Ld(N, DM_D1(C));
  DoubleMult = DoubleMult!=0;		/* ensure 0 or 1 */
  if (N > 0)
    MatMalloc(PROGNAME, 1, &jBx, (2+DoubleMult)*N);
  jAx = jBx + N;
  jC0 = jAx + (DoubleMult? N: 0);
  for (j=0; j < N; j++) {
    int jG = L2Gd(j, DM_D1(C));
    jBx[j] = (FD_BASE) ((jG + BJ0) % mxB);
    if (DoubleMult)
      jAx[j] = (FD_BASE) ((jG + AI0) % mxA);
    jC0[j] = *rx * (FD_BASE) ((jG + CJ0) % mx);
  }

  NErrs = *MaxI = *MaxJ = 0;		*AvDiff = *MxDiff = 0.0;   
  for (i=0; i<M; i++) {
    register FD_BASE DI0, DIInc, iC0, DIInc2;
    int iG = L2Gd(i, DM_D0(C)), 	jL = 0, jU = N,
        iAM = (iG + AI0) % myA, 	iBM = (iG + BJ0) % myB; 
    int ejmn, ejmx, nej;
    DI0 = D00 + RInc * (FD_BASE) iAM;
    DIInc = CInc + RCInc * (FD_BASE) iAM;
/*printf("DI0=%f, DIInc=%f\n", DI0, DIInc);*/
    if (DoubleMult) {
      DI0 += D002 + RInc2 * (FD_BASE) iBM;
      DIInc2 = CInc2 + RCInc2 * (FD_BASE) iBM;
/*printf("+DI0=%f, DIInc2=%f\n", DI0, DIInc2);*/
    }
    iC0 = *ry * (FD_BASE) ((iG + CI0) % my);

    if (FormC==Lower) jU=G2Ld(iG+1, DM_D1(C)); 
    if (FormC==Upper) jL=G2Ld(iG  , DM_D1(C));
    ejmn=-1; nej = 0;
    for (j=0; j<N; j++) {
      register FDATA Cij = *LM_Index(DM_Loc(C), i, j);
      register FDATA Dij = iC0 + jC0[j];/* initial value of C[i,j] */
      register FDATA Eij;
      if (jL<=j && j < jU) {
        Dij = (*beta) * Dij + DI0 + DIInc*jBx[j];
	if (DoubleMult)
          Dij += DIInc2*jAx[j];
      }
      Eij = Cij - Dij;      		Eij = FD_ABS(Eij);
      if (Eij > 1.0E-10) {
	if (ejmn==-1) {
          ejmn = j; 
	  if (!nej) 
            printf(" %d:", i);
	} 
	nej++; ejmx = j; 
	NErrs++;
	*AvDiff += Eij;
	if (FD_ABS(*MxDiff) < Eij) {
	  *MxDiff = Cij - Dij;
	  *MaxI = i;			*MaxJ = j;
        }
      } /* if (Eij...) */
      else if (ejmn >= 0) {
	printf(" %d-%d", ejmn, ejmx); 	
	ejmn = -1;
      }
      *LM_Index(DM_Loc(C), i, j) = Dij;
    }
    if (nej) {
      if (ejmn >= 0) 
        printf("%d-%d ", ejmn, ejmx);
      printf("(%d);", nej);
    }
  } /* for (i...) */
  if (NErrs)
    printf("\n");
  if (M*N > 0)
    *AvDiff /= (M*N);
  if (N > 0) 
    MatFree(jBx);
  return(NErrs);
} /* CheckLinearScatteredMult() */

/***************************** main program *******************************/

void main (argc, argv)
  int argc; char *argv[]; 
{
  /* matrices - dynamically allocated using WS*/
  DistMat_RVAlloc(3);
  DistMat A0, B0, C0;			/* matrices which contain A, B, C as submatrices*/
  int MA, NA, MB, NB, MC, NC;		/* dimesions of A0, B0, C0*/
  int MaxM, MaxN, MaxK;			/* maximum values of M, N, K */
  int DoSYRK, CommonBA, BASize;		/* flag set if matrices large so that
				   	    A, B forced to occupy same space */ 
  int TrA, TrB;				/* shorhands for TransA!=NoTranspose...*/
  					/* timing & result checking variables */
  double flops, speed, tcomp;
  char Date[64], LogFileName[64], ResultMsg[256];

  FD_BASE DAv=0.0, DMax=0.0; 		/* av. & max. differences */ 
  int NErr=0, DMaxI=0, DMaxJ=0;		/* # errors; position of max. difference */
  int i, rptm;				/* index for repeating multiply */

  CELL_INIT;
  STDOUT_SETUP(STDOUT);
  InitArgs(argc, argv);			
  GetDate(Date);
  
  COMMUN_INIT(Ncely, Ncelx);
  HPRINT((STDOUT, "%s", Date));
  GetLogFileName(argv[0], Date, LogFileName);
  sprintf(ResultMsg, "\n%.20s # %dx%d ", Date, Ncely, Ncelx);	
  LOGFILE_APPEND(LogFileName, ResultMsg);
  Chv2Str(argv, 0, argc, ' ', ResultMsg, i/*dummy*/);
  printf("%s cell (%d,%d) of %dx%d: %s\n", 
	  Date, CID_I, CID_J, NP_I, NP_J, ResultMsg);
  LOGFILE_APPEND(LogFileName, ResultMsg);

  GetParameters(argc, argv);
  CheckParameters();
  PrintParameters();	

  /* set up matrices */
  MaxM = M + (NReps-1)*MAX(StepM, 0);	MaxN = N + (NReps-1)*MAX(StepN, 0);
  MaxK = K + (NReps-1)*MAX(StepK, 0);

  TrA = TransA!=NoTranspose;		TrB = TransB!=NoTranspose;

  A0 = InitDimMat(TrA, !TrA? BM: BK, !TrA? BK: BM, CID_I, Ncely, CID_J, Ncelx, 0, 0);
  B0 = InitDimMat(TrB, !TrB? BK: BN, !TrB? BN: BK, CID_I, Ncely, CID_J, Ncelx, 0, 0);  
  C0 = InitDimMat(0, BM, BN, CID_I, Ncely, CID_J, Ncelx, 0, 0);
  MA = G2Ld((!TrA? MaxM: MaxK) + AI0, DM_D0( A0));
  NA = G2Ld((!TrA? MaxK: MaxM) + AJ0, DM_D1( A0));
  MB = G2Ld((!TrB? MaxK: MaxN) + BI0, DM_D0( B0));
  NB = G2Ld((!TrB? MaxN: MaxK) + BJ0, DM_D1( B0));
  MC = G2Ld(MaxM + CI0, DM_D0( C0)); 
  NC = G2Ld(MaxN + CJ0, DM_D1( C0));

  if (DoubleMult)			/* SYR2K special case */
    B0 = Trans(B0);  

  /* save space by B occupying same memory as A where possible*/
  DoSYRK = (FormC!=Rectangular && !DoubleMult);
  CommonBA = DoSYRK;
  BASize = MAX(MA,MB)*MAX(NA,NB);	/* shared A B size */
  if (!DoSYRK && FormA==Rectangular && FormB==Rectangular && 
       ColMajB==ColMajA &&
       Dim_BCompat(DM_D0(A0), DM_D0(B0)) && 
       Dim_BCompat(DM_D1(A0), DM_D1(B0))) {
    CommonBA = (MA*NA + MB*NB) > BASize;
  }
  if (!CommonBA)
    BASize = MA*NA + MB*NB;


  LM_Init(DM_Loc(C0), NULL,  ColMajC && MC>1? 1: MAX(NC,1), 
			    !ColMajC && NC>1? 1: MAX(MC,1));
  MatMalloc(PROGNAME, 1, &LM_ADR(DM_Loc(C0)), MC*NC);
  MatMalloc(PROGNAME, 1, &LM_ADR(DM_Loc(A0)), BASize);
  if (CommonBA) {
    int n = MAX(MAX(NA,NB),1),		m = MAX(MAX(MA,MB),1);
    LM_Init(DM_Loc(A0), LM_ADR(DM_Loc(A0)),  ColMajA && m>1? 1: n, 
	    				    !ColMajA && n>1? 1: m);
    LM_Copy(0, DM_Loc(B0), DM_Loc(A0));
  } else {
    LM_Init(DM_Loc(A0), LM_ADR(DM_Loc(A0)),  ColMajA && MA>1? 1: MAX(NA,1), 
					    !ColMajA && NA>1? 1: MAX(MA,1));
    LM_Init(DM_Loc(B0), LM_ADR(DM_Loc(A0))+MA*NA, 
	    ColMajB && MB>1? 1: MAX(NB,1), !ColMajB && NB>1? 1: MAX(MB,1));
  }

  for (i=0; i < NReps; i++) {
    DistMat_RVAlloc(5);
    DistMat A, B, C;			int mC, nC;
    MA = (!TrA? M: K) + AI0;		NA = (!TrA? K: M) + AJ0;
    MB = (!TrB? K: N) + BI0;		NB = (!TrB? N: K) + BJ0;
    MC = M + CI0; 			NC = N + CJ0;

					/* set up data for each repetition */
    SetLinearScattered(&ry, &rx, my, mx, MC, NC, &FD_1, C0);
    if (CommonBA) 
      SetLinearScattered(&ry, &rx, my, mx, MAX(MA,MB), MAX(NA,NB), &FD_1, A0);
    else {
      SetLinearScattered(&ry, &rx, my, mx, MA, NA, &FD_1, A0);
      SetLinearScattered(&ry, &rx, my, mx, MB, NB, &FD_1, B0);
    }
    A = SubMatTr(DM_Trans(A0), A0, AI0, AJ0);
    B = SubMatTr(DM_Trans(B0), B0, BI0, BJ0);
    C = SubMat(C0, CI0, CJ0);
    if (FormA != Rectangular) { /* zero elts of A that shouldnt be accessed   */
      DistMat AT;			AT = SubMatTr(FormA!=Upper, A, 1, 0); 
      TriMatScale(G2Ld(MA-AI0-1, DM_D0(AT)), G2Ld(NA-AJ0-1, DM_D1(AT)),
		  &FD_0, TransForm(FormA), AT);
    }
    if (FormB != Rectangular) { /* zero elts of B that shouldnt be accessed   */
      DistMat BT;			BT = SubMatTr(FormB!=Upper, B, 1, 0); 
      TriMatScale(G2Ld(MB-BI0-1, DM_D0(BT)), G2Ld(NB-BJ0-1, DM_D1(BT)),
		  &FD_0, TransForm(FormB), BT);
    }
    MatPrint("A", MIN(G2Ld(MA-AI0, DM_D0(A)), PrintRes), 
	          MIN(G2Ld(NA-AJ0, DM_D1(A)), PrintRes), DM_Loc(A));
    MatPrint("B", MIN(G2Ld(MB-BI0, DM_D0(B)), PrintRes), 
	          MIN(G2Ld(NB-BJ0, DM_D1(B)), PrintRes), DM_Loc(B));

    start_timer();
    if (FormC != Rectangular) {
      if (FormA != Rectangular || FormB != Rectangular) {
	HPRINT((STDOUT, "GMM.cell: if C of form %c, A & B must be of rectangular form for SYRK/2K\n", 
		FormC));
	exit(2);
      }
      if (TransA == TransB) {
	HPRINT((STDOUT, "%s: (form C = %c) exactly one of A or B must be trans. for SYRK/2K\n", 
		  PROGNAME, FormC));
	exit(2);
      }
      if (!DoubleMult) {
        if (AI0!=BI0 || AJ0!=BJ0) {
	  HPRINT((STDOUT, "%s: for SYRK, offsets of A & B must be identical\n", 
		  PROGNAME, FormC));
	  exit(2);
        }
        if (Level<=2 && beta==1.0 && K==1)
	  for (rptm=0; rptm < rptmult; rptm++) {
	    if (rptm%SYRK_SYNC==0)
	      BarrierSync(1-TrA);
            D_SYR(&FormC, N, alpha, A, C);
	  }
        else
	  for (rptm=0; rptm < rptmult; rptm++) {
	    if (rptm%SYRK_SYNC==0)
	      BarrierSync(1-TrA);
            D_SYRK(&FormC, N, K, alpha, A, rptm? FD_1: beta, C);
	  }
      } else
        if (Level<=2 && beta==1.0 && K==1) 
	  for (rptm=0; rptm < rptmult; rptm++) {
	    if (rptm%SYRK_SYNC==0)
	      BarrierSync(1-TrA);
            D_SYR2(&FormC, N, alpha, A, B, C);
	  }
        else
          for (rptm=0; rptm < rptmult; rptm++) {
	    if (rptm%SYRK_SYNC==0)
	      BarrierSync(1-TrA);
            D_SYR2K(&FormC, N, K, alpha, A, B, rptm? FD_1: beta, C);
	  }
    } else if (DoubleMult) {	
	HPRINT((STDOUT, "%s: for C of form %c, -2 (double mult, ie SYR2K) option is not implemented.\n", 
		PROGNAME, FormA));
	exit(2);
    } else if (FormA != Rectangular) {      
      if (FormB != Rectangular) {
	HPRINT((STDOUT, "%s: if A of form %c, B must be of rectangular for SYMM\n", 
		PROGNAME, FormA));
	exit(2);
      }
      if (Level<=2 && N==1) 
        for (rptm=0; rptm < rptmult; rptm++)
          D_SYMV(&FormA, M, alpha, A, B, rptm? FD_1: beta, C);
      else
        for (rptm=0; rptm < rptmult; rptm++)
          D_SYMM(&FormA, M, N, alpha, A, B, rptm? FD_1: beta, C);
    } else if (FormB != Rectangular) {
      char RevFormB = TransForm(FormB);
      if (Level<=2 && M==1) 
        for (rptm=0; rptm < rptmult; rptm++)
          D_SYMV(&RevFormB, N, alpha, Trans(B), Trans(A), rptm? FD_1: beta, Trans(C));
      else
        for (rptm=0; rptm < rptmult; rptm++)
#if 0
          D_SYMM(&RevFormB, N, M, alpha, Trans(B), Trans(A), rptm? FD_1: beta, Trans(C));
#else
          D_SYMM(&FormB, N, M, alpha, B, Trans(A), rptm? FD_1: beta, Trans(C));
#endif
    } else {
      if (Level==1 && K>1 && TrA!=TrB) {
	FDATA dot = 0.0;
        for (rptm=0; rptm < rptmult; rptm++) 
	  dot += D_DOT(K, Trans(A), B);	
        D_SetElt(C, beta * D_GetElt(C) + alpha * dot);
      } else if (Level==1 && N>1 && TrB==0) {
        FDATA a = alpha * D_GetElt(A);
        D_SCAL(N, beta, Trans(C));
        for (rptm=0; rptm < rptmult; rptm++) 
          D_AXPY(N, a, Trans(B), Trans(C));
      } else if (Level==1 && M>=1 && K==1 && TrA==0) {
        FDATA b = alpha * D_GetElt(B);
        D_SCAL(M, beta, C);
        for (rptm=0; rptm < rptmult; rptm++) 
          D_AXPY(M, b, A, C);	
      } else if (Level<=2 && K==1 && beta==1.0) {
        for (rptm=0; rptm < rptmult; rptm++) 
          D_GER(M, N, alpha, A, Trans(B), C);
      } else if (Level<=2 && N==1)	
        for (rptm=0; rptm < rptmult; rptm++) 
          D_GEMV(M, K, alpha, A, B, rptm? FD_1: beta, C);
      else if (Level<=2 && M==1)	
        for (rptm=0; rptm < rptmult; rptm++) 
          D_GEMV(N, K, alpha, Trans(B), Trans(A), rptm? FD_1: beta, Trans(C));
      else
        for (rptm=0; rptm < rptmult; rptm++) 
          D_GEMM(M, N, K, alpha, A, B, rptm? FD_1: beta, C);
    }
    tcomp = end_timer(1);

    flops = (double) M*N*(2*K+(alpha!=FD_1 && alpha!=FD_M1));
    if (FormC!=Rectangular && !DoubleMult) flops /= 2.0;
    speed = rptmult * flops / 1.0E6 /tcomp;
    HPRINT((STDOUT, "Computed %d reps of %dx%dx%d multiply (%d FLOPs) in %Gs at %G MFLOPs/cell\n",
            rptmult, M, K, N, (int) flops, tcomp, speed/Ncelx/Ncely));    
    printf("Computed %dx%dx%d multiply (%d FLOPs) in %Gs at %G MFLOPs/cell\n",
            M, K, N, (int) flops, tcomp, speed/Ncelx/Ncely);    
    sprintf(ResultMsg, " #s: %.3f %.3f", tcomp, speed/Ncelx/Ncely);
    LOGFILE_APPEND(LogFileName, ResultMsg);

    mC = G2Ld(M, DM_D0( C));		nC = G2Ld(N, DM_D1( C));
    MatPrint("C", MIN(mC,PrintRes), MIN(nC,PrintRes), DM_Loc(C));
    alpha *= (FDATA) rptmult;
    NErr = CheckLinearScatteredMult(&ry, &rx, my, mx, DoubleMult, 
	TrA, TrB, FormC, M, N, K, &alpha, AI0, AJ0, BI0, BJ0, 
	&beta, CI0, CJ0, C, &DMax, &DAv, &DMaxI, &DMaxJ);
    MatPrint("D", MIN(mC,PrintRes), MIN(nC,PrintRes), DM_Loc(C));
    if (mC > 0 && nC > 0)
#define E_MAX *LM_Index(DM_Loc(C), DMaxI, DMaxJ)
      printf("%d errs; Av: %f; Max Error: C[%d,%d] = %f; true value = %f\n", 
	     NErr, DAv, DMaxI, DMaxJ, M*N>0? E_MAX+DMax: 0.0,  M*N>0? E_MAX: 0.0);
#undef E_MAX

    GatherSend(NErr, sizeof(int));
    GatherSend(DMax, sizeof(FD_BASE));
    GatherSend(DAv,  sizeof(FD_BASE));
    GatherSend(DMaxI, sizeof(int));
    GatherSend(DMaxJ, sizeof(int));

    if (CID==0) {			/* accumulate error results */
      int *NErrs, *MxDiffIs, *MxDiffJs; 	FD_BASE *MxDiffs, *AvDiffs;
      NErrs = (int *) malloc(3*Ncely*Ncelx*sizeof(int));
      MxDiffIs = NErrs + Ncely*Ncelx;	MxDiffJs = MxDiffIs + Ncely*Ncelx;
      MxDiffs = (FD_BASE *) memalign(sizeof(FD_BASE), 2*Ncely*Ncelx*sizeof(FD_BASE));
      AvDiffs = MxDiffs + Ncely*Ncelx;

      NErrs[0] = NErr; 			/* needed for non-parallel programs */
      MxDiffs[0] = DMax;		AvDiffs[0] = DAv;
      MxDiffIs[0] = DMaxI;		MxDiffJs[0] = DMaxJ;

      GatherMsg(Ncely, Ncelx,  NErrs, sizeof(int));
      NErr = SumIntMat(Ncely, Ncelx, NErrs);
      GatherMsg(Ncely, Ncelx,  MxDiffs, sizeof(FD_BASE));
      DMax = MaxMatrix(MxDiffs, Ncely, Ncelx); 
      GatherMsg(Ncely, Ncelx,  AvDiffs, sizeof(FD_BASE));
      DAv = SumMatrix(AvDiffs, Ncely, Ncelx)/(Ncely*Ncelx);
      HPRINT((STDOUT, "%d global errors, max=%G, av=%G\n", NErr, DMax, DAv));
      sprintf(ResultMsg, " #e: %d %.3f %.3f", NErr, DMax, DAv);
      LOGFILE_APPEND(LogFileName, ResultMsg);

      GatherMsg(Ncely, Ncelx,  MxDiffIs, sizeof(int));
      GatherMsg(Ncely, Ncelx,  MxDiffJs, sizeof(int));
      if (NErr) {
        PrintIntMat(Ncely, Ncelx, NErrs);
	PrintMatrix('a', MxDiffs, Ncely, Ncelx, Ncelx);
        PrintMatrix('v', AvDiffs, Ncely, Ncelx, Ncelx);
        PrintCoords(Ncely, Ncelx, MxDiffIs, MxDiffJs);
      }	
      free(NErrs);			free(MxDiffs);
    }
    HPRINT((STDOUT, "\n\n"));

    M += StepM; N += StepN; K += StepK;

    if (NErr)
      exit(1);
  } /* for (i...) */
  if (BASize)
    MatFree(LM_ADR(DM_Loc(A0)));
  if (LM_ADR(DM_Loc(C0))!=NULL)
    MatFree(LM_ADR(DM_Loc(C0)));
  exit (0);
} /* main() */

