/* @(#) All Right Reserved, Copyright (c) 1997 Australian National University */
/* @(#)                     Copyright (c) 1997 Fujitsu Laboratories Ltd. */
/*  */
/*        */
/*                        Permission Notice */
/*  */
/*                Australian National University */
/*                Canberra ACT 0200 Australia */
/*                           and */
/*                Fujitsu Laboratories Ltd. */
/*                4-1-1 Kamikodanaka Nakahara */
/*                Kawasaki 211-88 Japan */
/*  */
/*   Permission to use, copy, modify and distribute this software (including  */
/*   its documentation) without fee is hereby granted under the following terms */
/*   and conditions: */
/*  */
/*   1. This permission notice and all of the copyright notices in this */
/*      software must appear in all copies and derived versions of all */
/*      or a part of this software. */
/*  */
/*   2. Sales of this software, whether incorporated in other products as a */
/*      part thereof or not, is prohibited. */
/*  */
/*   3. On publication or presentation in any form of the results obtained */
/*      by use of this software, an explicit reference to this software and */
/*      its copyright owner shall be made. */
/*  */
/*   4. THIS SOFTWARE IS PROVIDED 'AS IS' AND THE AUSTRALIAN NATIONAL UNIVERSITY */
/*      AND FUJITSU LABORATORIES LTD. MAKE NO WARRANTIES, EITHER EXPRESSED OR */
/*      IMPLIED, AS TO ANY MATTER WHATSOEVER REGARDING THIS SOFTWARE,  */
/*      INCLUDING, WITHOUT LIMITATION, WARRANTIES OF MERCHANTABILITY OR FITNESS */
/*      FOR ANY PARTICULAR PURPOSE. */
/*      IN NO EVENT SHALL THE AUSTRALIAN NATIONAL UNIVERSITY OR FUJITSU */
/*      LABORATORIES LTD. BE LIABLE FOR ANY CLAIM FROM A THIRD PARTY, OR */
/*      FOR SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES WHATSOEVER RESULTING */
/*      FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, */
/*      NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION */
/*      WITH THE USE, COPYING OR MODIFICATION OF THIS SOFTWARE. */
/*  */

/************************************************************************/
/*         D_BLAS.c: Level 3 Distributed BLAS implementation file	*/
/************************************************************************/
/* SCCS INFO @(#)D_BLAS.c 4.4         last modified P. Strazdins, 97/05/14
   first created: P. Strazdins 93/03/23
 */
/* implementation file for Distributed Level-3 BLAS procedures; 
   for compilation, the macro FDATATYPE (having the values SINGLE or DOUBLE)
   must be defined:
	if FDATATYPE=SINGLE, FDATA=float
	if FDATATYPE=DOUBLE, FDATA=double
   Below, the underscores in the procedure name should be replaced by 'S',
   if FDATATYPE=SINGLE, and 'D', if FDATATYPE=DOUBLE. 
   eg. for FDATATYPE=SINGLE, D_GEMM becomes DSGEMM.

   See DBLAS.h for a general description of this module.
   
   (externally-defined) cpp FLAGs:
        FDATATYPE       SINGLE (default), DOUBLE, COMPLEX or DOUBLE_COMPLEX
        ASSERT, TRACE   turn assertion checking and procedure tracing on
                        in major internal procedure calls
*/   
#include "D_BLAS.h"
#include "D_BLASCheck.h"
#include "BLASOptions.h"
#include "FDATASetUp.h"
#include "FDATAAssert.h"

#include "LocMat.h"
#include "MatDim.h"
#include "DistMat.h"
#include "FDATACommun.h"  		/* CommSetUp() */

/****************** DBLAS procedures *****************************************/

#ifdef COMP_D_GEMM
#include "GenMult.h"

void D_GEMM(M, N, K, alpha, A, B, beta, C)
  int M; int N; int K;
  FDATA alpha; DistMat A; DistMat B; FDATA beta; DistMat C;
{ int AlignM, AlignN, AlignK;
  int ParamErr=0;
  AlphaBetaDecln();

  CheckAlign(ParamErr, AlignM, M, DM_Inc(1-Tr(A), A),   DM_DM(A), 
  	     DM_Inc(1-Tr(C), C),   DM_DM(C), 0x158);
  CheckParam(ParamErr, 
  	     PAlignCheck(d_P(DM_DM(A)), DM_P(Tr(A), C), &AlignM), 0x58);
  CheckAlign(ParamErr, AlignN, N, DM_Inc(Tr(B), B), DM_DN(B), 
  	     DM_Inc(Tr(C), C), DM_DN(C), 0x268);  
  CheckParam(ParamErr, 
             PAlignCheck(d_P(DM_DN(B)), DM_P(1-Tr(B), C), &AlignM), 0x68);
  CheckAlign(ParamErr, AlignK, K, DM_Inc(Tr(A), A), DM_DN(A), 
  	     DM_Inc(1-Tr(B), B),   DM_DM(B), 0x356);  
  CheckParam(ParamErr, 
   	     PAlignCheck(d_P(DM_DN(A)), DM_P(1-Tr(A), C), &AlignK), 0x58);
  CheckParam(ParamErr, 
   	     PAlignCheck(d_P(DM_DM(B)), DM_P(Tr(B), C), &AlignK), 0x68);

  if (ForceAlgVers==1) 		AlignM = 0;
  else if (ForceAlgVers==2) 	AlignN = 0;
  else if (ForceAlgVers==3) 	AlignK = 0;
           	     
  CheckParam(ParamErr, AlignM+AlignN+AlignK>=2, 0x568);
  
  SetUpComm(C);				
  ERROR_CHECK_TR(
    (TraceStr, "(%d,%d,%d,%1.1f,%s,%s,%1.1f,%s)", 
      M, N, K, alpha, DM_TRACE(0, A), DM_TRACE(1, B), beta, DM_TRACE(2, C)),
    ParamErr, "GEMM", 0
  );

  if (AlignM+AlignN+AlignK==3) 		/* chose version simply on dim. size */	
    if (M < N && M < K)
      AlignM = 0;
    else if (N < M && N < K)
      AlignN = 0;         	
    else 
     AlignK = 0;  
  
  AlphaSet(alpha, pAlpha);	 	BetaSet(beta, pBeta);
  if (!AlignK)
    DistRankKMult(M, N, K, pAlpha, A, B, pBeta, C);
  else if (!AlignM)
    DistVecMatMult(M, N, K, pAlpha, A, B, pBeta, C);
  else 
    DistVecMatMult(N, M, K, pAlpha, Trans(B), Trans(A), pBeta, Trans(C));    
  TRACE_EXIT;
} /* D_GEMM() */
#endif /* COMP_D_GEMM */


#ifdef COMP_D_SYMM
#include "SYMMult.h"

void D_SYMM (UpLo, M, N, alpha, A, B, beta, C)
  char *UpLo; int M; int N; 
  FDATA alpha; DistMat A; DistMat B; FDATA beta; DistMat C;
{ char UpLoA = UpperCase(*UpLo); 
  int ParamErr=0;  			
  AlphaBetaDecln();
  int AlignM=1, AlignN=1;

  if (!NoOuterErrCheck) {      
    CheckParam(ParamErr, UpLoOK(UpLoA), 0x1);  
    CheckParam(ParamErr, Dim_BCompat(DM_D0(A), DM_D1(A)), 0x55);
    CheckAlign(ParamErr, AlignM, M, DM_Inc(1-Tr(A), A),   DM_DM(A),
               DM_Inc(1-Tr(C), C),   DM_DM(C), 0x258);
    CheckParam(ParamErr, AlignM && d_P(DM_DM(A))==DM_P(Tr(A), C) && 
    			           d_P(DM_DN(A))==DM_P(1-Tr(A), C), 0x58);
    CheckAlign(ParamErr, AlignM, M, DM_Inc(Tr(A), A),     DM_DN(A),
               DM_Inc(1-Tr(B), B),  DM_DM(B), 0x256);
    CheckParam(ParamErr, AlignM, 0x57);
    CheckAlign(ParamErr, AlignN, N, DM_Inc(Tr(B), B),     DM_DN(B),
               DM_Inc(Tr(C), C),    DM_DN(C), 0x368);
    CheckParam(ParamErr, 
               PAlignCheck(d_P(DM_DN(B)), DM_P(1-Tr(B), C), &AlignN), 0x68);
  }

  SetUpComm(C);		
  ERROR_CHECK_TR(
    (TraceStr, "(%c,%d,%d,%1.1f,%s,%s,%1.1f,%s)", 
     *UpLo, M, N, alpha, DM_TRACE(0, A), DM_TRACE(1, B), beta, DM_TRACE(2, C)),
    ParamErr, "SYMM", 0
  );

  AlphaSet(alpha, pAlpha);	 	BetaSet(beta, pBeta);
  SYMMult(M, N, pAlpha, UpLoA, A, B, pBeta, C);
  TRACE_EXIT;
} /* D_SYMM() */
#endif /*COMP_D_SYMM*/


#ifdef COMP_D_SYRK
#include "SYRKMult.h"

void D_SYRK (UpLo, N, K, alpha, A, beta, C)
  char *UpLo; int N; int K;  FDATA alpha; DistMat A; FDATA beta; DistMat C;
{ char UpLoC = UpperCase(*UpLo); 
  int ParamErr=0;  			
  AlphaBetaDecln();
  int AlignM=1;

  if (!NoOuterErrCheck) {      
    CheckParam(ParamErr, UpLoOK(UpLoC), 0x1);  
    CheckAlign(ParamErr, AlignM, N, DM_Inc(1-Tr(A), A),   DM_DM(A),
               DM_Inc(1-Tr(C), C),   DM_DM(C), 0x257);
    CheckParam(ParamErr, AlignM && d_P(DM_DM(A))==DM_P(Tr(A), C) && 
    			 P_OK(d_P(DM_DN(A)), DM_P(1-Tr(A), C)), 0x57);
    CheckDim(ParamErr, K, DM_Inc(Tr(A), A), DM_DN(A), 0x35);
    CheckDim(ParamErr, N, DM_Inc(Tr(C), C), DM_DN(C), 0x27);
    CheckParam(ParamErr, Dim_BCompat(DM_D0(C), DM_D1(C)), 0x77);
  }

  SetUpComm(C);
  ERROR_CHECK_TR(
    (TraceStr, "(%c,%d,%d,%1.1f,%s,%1.1f,%s)", 
     *UpLo, N, K, alpha, DM_TRACE(0, A), beta, DM_TRACE(1, C)), 
    ParamErr, "SYRK", 0
  );

  AlphaSet(alpha, pAlpha);              BetaSet(beta, pBeta);
  DistSYRKMult(N, K, pAlpha, A, pBeta, UpLoC, C);
  TRACE_EXIT;
} /* D_SYRK()*/
#endif /*COMP_D_SYRK*/


#ifdef COMP_D_SYR2K
#include "SYR2KMult.h"

void D_SYR2K (UpLo, N, K, alpha, A, B, beta, C)
  char *UpLo; int N; int K;
  FDATA alpha; DistMat A; DistMat B; FDATA beta; DistMat C;
{ char UpLoC = UpperCase(*UpLo); 
  int ParamErr=0;
  AlphaBetaDecln();
  int AlignM=1;

  if (!NoOuterErrCheck) {
    CheckParam(ParamErr, UpLoOK(UpLoC), 0x1);
    CheckParam(ParamErr, Tr(A)==Tr(B), 0x56);
    CheckAlign(ParamErr, AlignM, N, DM_Inc(1-Tr(A), A),   DM_DM(A),
               DM_Inc(1-Tr(C), C),   DM_DM(C), 0x258);
    CheckDim(ParamErr, K, DM_Inc(Tr(A), A), DM_DN(A), 0x35);
    CheckParam(ParamErr, AlignM &&
			 d_P(DM_DM(A))==DM_P(Tr(A), C) &&
			 P_OK(d_P(DM_DN(A)), DM_P(1-Tr(A), C)), 0x58); 
    CheckDim(ParamErr, N, DM_Inc(1-Tr(B), B),   DM_DM(B), 0x26);
    CheckDim(ParamErr, K, DM_Inc(Tr(B), B), DM_DN(B), 0x36);
    CheckParam(ParamErr, Dim_BCompat(DM_DM(B), DM_DM(C)) &&
			 d_P(DM_DM(B))==DM_P(Tr(B), C) &&
			 P_OK(d_P(DM_DN(B)), DM_P(1-Tr(B), C)), 0x68);
    CheckDim(ParamErr, N, DM_Inc(Tr(C), C), DM_DN(C), 0x28);
    CheckParam(ParamErr, Dim_BCompat(DM_D0(C), DM_D1(C)), 0x88);
  }  
    

  SetUpComm(C);  
  ERROR_CHECK_TR(   
    (TraceStr, "(%c,%d,%d,%1.1f,%s,%s,%1.1f,%s)", *UpLo, 
     N, K, alpha, DM_TRACE(0, A), DM_TRACE(1, B), beta, DM_TRACE(2, C)),
    ParamErr, "SYR2K", 0
  );
  
  AlphaSet(alpha, pAlpha);              BetaSet(beta, pBeta);   
  DistSYR2KMult(N, K, pAlpha, A, B, pBeta, UpLoC, C);
  TRACE_EXIT;
} /* D_SYR2K()*/

#endif /* D_SYR2K */

#ifdef COMP_D_TRMM
#include "TRSM.h"
#include "TriMatUpdate.h"

void D_TRMM( UpLo, Diag, M, N, alpha, A, B)
  char *UpLo; char *Diag; int M; int N; FDATA alpha; DistMat A;  DistMat B;
{ char UpLoA = UpperCase(*UpLo), 	DiagA = UpperCase(*Diag); 
  int AlignM=1,				AlignK = CompatBlk(DM_DM(B), DM_DN(A));
  int ParamErr=0;  
  AlphaDecln();
  
  CheckAlign(ParamErr, AlignM, M, DM_Inc(1-Tr(A), A),   DM_DM(A),
             DM_Inc(1-Tr(B), B),  DM_DM(B), 0x367);
  if (!NoOuterErrCheck) {  
    CheckParam(ParamErr, UpLoOK(UpLoA), 0x1);  
    CheckParam(ParamErr, DiagOK(DiagA), 0x2);  
    CheckDim(ParamErr, M, DM_Inc(Tr(A), A), DM_DN(A), 0x36);
    CheckDim(ParamErr, N, DM_Inc(Tr(B), B), DM_DN(B), 0x47);
    CheckParam(ParamErr, (AlignM | AlignK) && d_P(DM_DM(A))==DM_P(Tr(A), B) &&
               P_OK(d_P(DM_DN(A)), DM_P(1-Tr(A), B)), 0x67);
    CheckDim(ParamErr, M, DM_Inc(Tr(A), A), DM_DN(A), 0x36);
    CheckDim(ParamErr, N, DM_Inc(Tr(B), B), DM_DN(B), 0x47);
  }

  SetUpComm(B);
  ERROR_CHECK_TR(   
    (TraceStr, "(%c,%c,%d,%d,%1.1f,%s,%s)",
    *UpLo, *Diag, M, N, alpha, DM_TRACE(0, A), DM_TRACE(1, B)),
    ParamErr, "TRMM", 0
  );

  if (DM_Trans(A)^DM_Trans(B))
    UpLoA = TransForm(UpLoA);
  AlphaSet(alpha, pAlpha); 
  if (ForceAlgVers==2) AlignM = 0;
  else if (ForceAlgVers==1) AlignK = 0;
  if (!AlignM || (AlignK && 4*N<=M))
    TriMatUpdate(UpLoA, DiagA, M, N, pAlpha, A, B);    
  else
    TR_M(0, UpLoA, DiagA, M, N, pAlpha, A, B);
  TRACE_EXIT;
} /* D_TRMM() */

#endif /* COMP_D_TRMM */

#ifdef COMP_D_TRSM
#include "TRSM.h"
#include "TriMatSolve.h"

void D_TRSM( UpLo, Diag, M, N, alpha, A, B)
  char *UpLo; char *Diag; int M; int N; FDATA alpha; DistMat A;  DistMat B;
{ char UpLoA = UpperCase(*UpLo), 	DiagA = UpperCase(*Diag); 
  int AlignM=1,				AlignK = CompatBlk(DM_DM(B), DM_DN(A));
  int ParamErr=0;  
  AlphaDecln();
  
  CheckAlign(ParamErr, AlignM, M, DM_Inc(1-Tr(A), A),   DM_DM(A),
             DM_Inc(1-Tr(B), B),  DM_DM(B), 0x367);
  if (!NoOuterErrCheck) {  
    CheckParam(ParamErr, UpLoOK(UpLoA), 0x1);  
    CheckParam(ParamErr, DiagOK(DiagA), 0x2);  
    CheckDim(ParamErr, M, DM_Inc(Tr(A), A), DM_DN(A), 0x36);
    CheckDim(ParamErr, N, DM_Inc(Tr(B), B), DM_DN(B), 0x47);
    CheckParam(ParamErr, (AlignM|AlignK) && d_P(DM_DM(A))==DM_P(Tr(A), B) &&
               P_OK(d_P(DM_DN(A)), DM_P(1-Tr(A), B)), 0x67);
    CheckDim(ParamErr, M, DM_Inc(Tr(A), A), DM_DN(A), 0x36);
    CheckDim(ParamErr, N, DM_Inc(Tr(B), B), DM_DN(B), 0x47);
  }

  SetUpComm(B);
  ERROR_CHECK_TR(   
    (TraceStr, "(%c,%c,%d,%d,%1.1f,%s,%s)",
    *UpLo, *Diag, M, N, alpha, DM_TRACE(0, A), DM_TRACE(1, B)),
    ParamErr, "TRSM", 0
  );

  if (DM_Trans(A)^DM_Trans(B))
    UpLoA = TransForm(UpLoA);
  AlphaSet(alpha, pAlpha);
  if (ForceAlgVers==2) AlignM = 0;
  else if (ForceAlgVers==1) AlignK = 0;
  if (!AlignM || (AlignK && 4*N<=M))
    if (Blk1d(N, DM_DN(B)) || !AlignM)
      TriMatSolve1Blk(UpLoA, DiagA, M, N, pAlpha, A, B);    
    else  
      TriMatSolve(UpLoA, DiagA, M, N, pAlpha, A, B);    
  else
    TR_M(1, UpLoA, DiagA, M, N, pAlpha, A, B);
  TRACE_EXIT;
} /* D_TRSM() */

#endif /* COMP_D_TRSM */

