/* @(#) All Right Reserved, Copyright (c) 1997 Australian National University */
/* @(#)                     Copyright (c) 1997 Fujitsu Laboratories Ltd. */
/*  */
/*        */
/*                        Permission Notice */
/*  */
/*                Australian National University */
/*                Canberra ACT 0200 Australia */
/*                           and */
/*                Fujitsu Laboratories Ltd. */
/*                4-1-1 Kamikodanaka Nakahara */
/*                Kawasaki 211-88 Japan */
/*  */
/*   Permission to use, copy, modify and distribute this software (including  */
/*   its documentation) without fee is hereby granted under the following terms */
/*   and conditions: */
/*  */
/*   1. This permission notice and all of the copyright notices in this */
/*      software must appear in all copies and derived versions of all */
/*      or a part of this software. */
/*  */
/*   2. Sales of this software, whether incorporated in other products as a */
/*      part thereof or not, is prohibited. */
/*  */
/*   3. On publication or presentation in any form of the results obtained */
/*      by use of this software, an explicit reference to this software and */
/*      its copyright owner shall be made. */
/*  */
/*   4. THIS SOFTWARE IS PROVIDED 'AS IS' AND THE AUSTRALIAN NATIONAL UNIVERSITY */
/*      AND FUJITSU LABORATORIES LTD. MAKE NO WARRANTIES, EITHER EXPRESSED OR */
/*      IMPLIED, AS TO ANY MATTER WHATSOEVER REGARDING THIS SOFTWARE,  */
/*      INCLUDING, WITHOUT LIMITATION, WARRANTIES OF MERCHANTABILITY OR FITNESS */
/*      FOR ANY PARTICULAR PURPOSE. */
/*      IN NO EVENT SHALL THE AUSTRALIAN NATIONAL UNIVERSITY OR FUJITSU */
/*      LABORATORIES LTD. BE LIABLE FOR ANY CLAIM FROM A THIRD PARTY, OR */
/*      FOR SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES WHATSOEVER RESULTING */
/*      FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, */
/*      NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION */
/*      WITH THE USE, COPYING OR MODIFICATION OF THIS SOFTWARE. */
/*  */

/************************************************************************/
/*         D_BLAS1.c: Level 1 Distributed BLAS implementation file	*/
/************************************************************************/
/* SCCS INFO @(#)D_BLAS1.c 4.5         last modified P. Strazdins, 97/04/25
   first created: P. Strazdins 93/03/23
 */
/* implementation file for Distributed Level-1 BLAS procedures; 
   See D_BLAS.h for a general description of this module.
   (externally-defined) cpp FLAGs:
        FDATATYPE       SINGLE (default), DOUBLE, COMPLEX or DOUBLE_COMPLEX
        ASSERT, TRACE   turn assertion checking and procedure tracing on
                        in major internal procedure calls
*/   
#include "D_BLAS.h"
#include "D_BLASCheck.h"

#include "BLASOptions.h"
#include "FDATASetUp.h"
#include "FDATAAssert.h"
#include "FDATAMalloc.h"

#include "LocMat.h"
#include "MatDim.h"
#include "DistMat.h"

#include "FDATACellMatrix.h"	/* MatScale(), MatCopy/Add(), MatMult() */
#include "FDATACommun.h"


static inline void SendVec(SwapDIJ, DI, DJ, nx, Xp)
  int SwapDIJ, DI, DJ, nx; LocMat Xp; 
/* post: send nx x 1 Xp to cell (SwapDIJ? (DI,DJ): (DJ:DI))   
 */
{ int IncX = LM_Inc(1, Xp);
  FDATA *b = LM_ADR(Xp);
  if (IncX != 1) {			/* must pack before sending */
    MatMalloc_FT("SendVec", 1, &b, nx);
    VecCopy(nx, LM_ADR(Xp), IncX, b, 1);
  }  
  if (SwapDIJ)
    SWAP(int, DI, DJ);
  SendMat(DI, DJ, 1, nx, b, nx, GENMSGTTYPE);  
  if (IncX != 1)
     MatFree(b);
} /* SendVec() */


#ifdef COMP_D_SWAP
#include "FDATACellVector.h"		/* VecSwap() */

void D_SWAP (N, X, Y) 
  int N; DistMat X; DistMat Y;
{ int IncX = DM_Inc(DM_Trans(X), X),	IncY = DM_Inc(DM_Trans(Y), Y);
  int ParamErr=0;			int AlignN=1;

  if (d_P0(DM_DN(X))!=0 && d_P0(DM_DN(Y))!=0) 
    return;				/* fast exit */
 
  if (!NoOuterErrCheck) {
    CheckAlign(ParamErr, AlignN, N, DM_Inc(1-Tr(X), X), DM_DM(X),
               DM_Inc(1-Tr(Y), Y), DM_DM(Y), 0x123);
    CheckParam(ParamErr, AlignN && P_Same(X, Y) && DM_Trans(X)==DM_Trans(Y), 
	       0x23);
  }

  SetUpComm(X);
  ERROR_CHECK_TR(
    (TraceStr, "(%d,%s,%s)", N, DM_TRACE(0, X), DM_TRACE(1, Y)),
    ParamErr, "SWAP", 1
  );


  if (d_P0(DM_DN(X))==0 && d_P0(DM_DN(Y))==0 && d_P0(DM_DM(X))==d_P0(DM_DM(Y)))
    VecSwap(G2Ld(N, DM_DM(X)), LM_ADR(DM_Loc(X)), IncX, LM_ADR(DM_Loc(Y)), IncY);
  else if (d_P0(DM_DN(X))==0 || d_P0(DM_DN(Y))==0) {
    int DIy = d_P0(DM_DM(Y))-d_P0(DM_DM(X)), 
        DJy = -d_P0(DM_DN(Y))-d_P0(DM_DN(X)),
        DIx = -DIy, 			DJx = DJy;
    FDATA *b = NULL;			
    int nx = G2Ld(N, DM_DM(X)),		ny = DIy==0? nx: G2Ld(N, DM_DM(Y));
	      
    if (DM_Trans(X)) {
      SWAP(int, DIy, DJy);		SWAP(int, DIx, DJx);
    }
    if (IncX != 1 || IncY != 1)
      MatMalloc_FT("D_SWAP", 1, &b, MAX(nx,ny));
      
    if (d_P0(DM_DN(X))==0 && nx>0) {
      FDATA *Xp = LM_ADR(DM_Loc(X));	      
      if (IncX != 1)
        VecCopy(nx, Xp, IncX, b, 1);
      SendMat(DIx, DJx, 1, nx, IncX!=1? b: Xp, nx, GENMSGTTYPE); 
    }
    if (d_P0(DM_DN(Y))==0 && ny>0) {
      FDATA *Yp = LM_ADR(DM_Loc(Y));	
      if (IncY != 1) 
        VecCopy(nx, Yp, IncY, b, 1);
      SendMat(DIy, DJy, 1, ny, IncX!=1? b: Yp, ny, GENMSGTTYPE); 
    }

    if (d_P0(DM_DN(Y))==0 && ny>0) {
      FDATA *Yp = LM_ADR(DM_Loc(Y));	
      RecvMat(DIy, DJy, 1, ny, IncY==1? Yp: b, ny, GENMSGTTYPE);  
      if (IncY != 1)         
        VecCopy(ny, b, 1, Yp, IncY);                  
    }
    if (d_P0(DM_DN(X))==0 && nx>0) {
      FDATA *Xp = LM_ADR(DM_Loc(X));	
      RecvMat(DIx, DJx, 1, nx, IncX==1? Xp: b, nx, GENMSGTTYPE);  
      if (IncX != 1)         
        VecCopy(nx, b, 1, Xp, IncX);                  
    }
    
    if (b!=NULL)
      MatFree(b);
 }
  TRACE_EXIT;
} /* D_SWAP() */

#endif /*COMP_D_SWAP*/


#ifdef COMP_D_SCAL

void D_SCAL (N, beta, X) 
  int N; FDATA beta; DistMat X; 
{ int ParamErr=0;
  BetaDecln();

  if (!NoOuterErrCheck) {
    CheckDim(ParamErr, N, DM_Inc(1-Tr(X), X), DM_DM(X), 0x13);
  }
  
  SetUpComm(X);	
  ERROR_CHECK_TR(
    (TraceStr, "(%d,%1.1f,%s)", N, beta, DM_TRACE(0, X)),
    ParamErr, "SCAL", 1
  );

  BetaSet(beta, pBeta);
  if (d_P0(DM_DN(X))==0) {
    LocMat Xp;
    LM_Copy(1-DM_Trans(X), Xp, DM_Loc(X));	
    MatScale(1, G2Ld(N, DM_DM(X)), pBeta, &Xp);
  }  	
       
  TRACE_EXIT;
} /* D_SCAL() */

#endif /*COMP_D_SCAL*/


#ifdef COMP_D_COPY
#include "FDATADistMatrix.h"            /* TransMat(), RotMat()  */

void D_COPY (N, X, Y)
  int N; DistMat X; DistMat Y;
{ int ParamErr=0;                       int AlignN=1;
  LocMat Xp, Yp;

  if (!NoOuterErrCheck) {
    CheckAlign(ParamErr, AlignN, N, DM_Inc(1-Tr(X), X), DM_DM(X),
               DM_Inc(1-Tr(Y), Y), DM_DM(Y), 0x123);
    CheckParam(ParamErr, AlignN && P_Same(X, Y), 0x23);
  }
      
  SetUpComm(X);
  ERROR_CHECK_TR(
    (TraceStr, "(%d,%s,%s)", N, DM_TRACE(0,X), DM_TRACE(1,Y)),
    ParamErr, "COPY", 1
  );

  LM_Copy(1-DM_Trans(X), Xp, DM_Loc(X));/* Xp, Yp are 1 x n */
  LM_Copy(1-DM_Trans(Y), Yp, DM_Loc(Y));	

  if (DM_Trans(X)!=DM_Trans(Y))
    TransMat(1-DM_Trans(X), 0, 1, N, &Xp, DM_DN(X), DM_DM(X), 1,
             NULL, &Yp, d_P0(DM_DM(Y)), d_P0(DM_DN(Y)), 1);
  else if (d_P0(DM_DN(X))==0 && d_P0(DM_DN(Y))==0 && 
	   d_P0(DM_DM(X))==d_P0(DM_DM(Y)))
    MatCopy(1, G2Ld(N, DM_DM(Y)), &Xp, NULL, &Yp);
  else if (d_P0(DM_DN(X))==0 || d_P0(DM_DN(Y))==0) {
    int n = G2Ld(N, DM_DM(X));	
    if (d_P0(DM_DN(X))==0 && n>0) 
      SendVec(DM_Trans(X), d_P0(DM_DM(X))-d_P0(DM_DM(Y)), -d_P0(DM_DN(Y)), n, Xp);
    if (d_P0(DM_DN(Y))==0) {
      int IncY = LM_Inc(1, Yp);
      int DI = d_P0(DM_DM(Y))-d_P0(DM_DM(X)),	DJ = -d_P0(DM_DN(X));
      int ny = DI==0? n: G2Ld(N, DM_DM(Y));
      FDATA *b = LM_ADR(Yp);			
      if (DM_Trans(X))
        SWAP(int, DI, DJ);
      if (ny > 0) {  
        if (IncY != 1)          		/* must receive into a buffer */
          MatMalloc_F("D_COPY", 1, &b, ny);
        RecvMat(DI, DJ, 1, ny, b, ny, GENMSGTTYPE);  
        if (IncY != 1) {
          VecCopy(ny, b, 1, LM_ADR(Yp), IncY);                  
          MatFree(b);
        }    
      }
    }    
  }   
  TRACE_EXIT;
} /* D_COPY() */

#endif /* COMP_D_COPY */


#ifdef COMP_D_AXPY
#include "FDATADistMatrix.h"            /* TransMat(), RotMat()  */

void D_AXPY (N, beta, X, Y) 
  int N; FDATA beta; DistMat X; DistMat Y;
{ LocMat Xp, Yp;			BetaDecln();
  int ParamErr=0;                       int AlignN;

 if (!NoOuterErrCheck) {
    CheckAlign(ParamErr, AlignN, N, DM_Inc(1-Tr(X), X), DM_DM(X),
               DM_Inc(1-Tr(Y), Y), DM_DM(Y), 0x123);
    CheckParam(ParamErr, AlignN && P_Same(X, Y), 0x34);
 }    
  
  SetUpComm(X);				
  ERROR_CHECK_TR(
    (TraceStr, "(%d,%1.1f,%s,%s)", N, beta, DM_TRACE(0, X), DM_TRACE(1, Y)),
    ParamErr, "AXPY", 1
  );
  
  BetaSet(beta, pBeta);

  LM_Copy(1-DM_Trans(X), Xp, DM_Loc(X));	/* Xp, Yp are 1 x n */
  LM_Copy(1-DM_Trans(Y), Yp, DM_Loc(Y));	

  if (DM_Trans(X)!=DM_Trans(Y)) 
    TransMat(1-DM_Trans(X), 1, 1, N, &Xp, DM_DN(X), DM_DM(X), 1,
             pBeta, &Yp, d_P0(DM_DM(Y)), d_P0(DM_DN(Y)), 1);
  else if (d_P0(DM_DN(X))==0 && d_P0(DM_DN(Y))==0 && 
	   d_P0(DM_DM(X))==d_P0(DM_DM(Y))) {
    MatAdd(1, G2Ld(N, DM_DM(X)), &Xp, pBeta, &Yp);
  } else if (d_P0(DM_DN(X))==0 || d_P0(DM_DN(Y))==0) {
    int n = G2Ld(N, DM_DM(X));	
    if (d_P0(DM_DN(X))==0 && n>0) {
      SendVec(DM_Trans(X), d_P0(DM_DM(X))-d_P0(DM_DM(Y)), -d_P0(DM_DN(Y)), 
	      n, Xp);
    }
    if (d_P0(DM_DN(Y))==0) {
      int IncY = LM_Inc(1, Yp);
      int DI = d_P0(DM_DM(Y))-d_P0(DM_DM(X)),	DJ = -d_P0(DM_DN(X));
      int ny = DI==0? n: G2Ld(N, DM_DM(Y));
      if (DM_Trans(X))
        SWAP(int, DI, DJ);  
      if (IncY != 1 && ny>0) { 
        FDATA *b = NULL;	
        MatMalloc_FT("D_AXPY", 1, &b, ny);
        LM_Init(Xp, b, ny, 1);
        RecvMat(DI, DJ, 1, ny, LM_ADR(Xp), ny, GENMSGTTYPE);  
        MatAdd(1, ny, &Xp, pBeta, &Yp);
        MatFree(b);
      } else if (ny>0) 
        RecvBlkMat(DI, DJ, 1, 1, 0, 1, 1, ny, pBeta, LM_ADR(Yp), GENMSGTTYPE);
    }    
  }
  TRACE_EXIT;
} /* D_AXPY() */

#endif /*COMP_D_AXPY*/


#ifdef COMP_D_DOT

FDATA D_DOT (N, X, Y) 
  int N; DistMat X; DistMat Y;
{ LocMat Xp, Yp, Sp; 			FDATA prod; 
  int nx;
  int DI=d_P0(DM_DM(Y))-d_P0(DM_DM(X)), DJ=d_P0(DM_DN(Y))-d_P0(DM_DN(X));
  int ParamErr=0;                       int AlignN;

 if (!NoOuterErrCheck) {
    CheckAlign(ParamErr, AlignN, N, DM_Inc(1-Tr(X), X), DM_DM(X),
               DM_Inc(1-Tr(Y), Y), DM_DM(Y), 0x123);
    CheckParam(ParamErr, AlignN && P_Same(X, Y) && DM_Trans(X)==DM_Trans(Y),
               0x24);
 }
    
  SetUpComm(X);  
  ERROR_CHECK_TR(
    (TraceStr, "(%d,%s,%s)", N, DM_TRACE(0, X), DM_TRACE(1, Y)),
    ParamErr, "DOT", 1
  );
    
  nx = G2Ld(N, DM_DM(X));
  LM_Copy(1-DM_Trans(X), Xp, DM_Loc(X));	/* Xp is 1 x nx */
  if (d_P0(DM_DN(X))==0 && (DI!=0 || DJ!=0) && nx>0) 
    SendVec(DM_Trans(X), -DI, -DJ, nx, Xp);
  if (d_P0(DM_DN(Y))==0) {
    int ny = DI==0? nx: G2Ld(N, DM_DM(Y));
    if ((DI!=0 || DJ!=0) && ny>0) {
      FDATA *b;		
      MatMalloc_FT("D_DOT", 1, &b, ny);
      if (DM_Trans(Y))
        SWAP(int, DI, DJ);
      RecvMat(DI, DJ, 1, ny, b, ny, GENMSGTTYPE);
      LM_Init(Xp, b, ny, 1);
    }  
    LM_Copy(DM_Trans(Y), Yp, DM_Loc(Y));	/* Yp is ny x 1 */
    LM_Init(Sp, &prod, 1, 1);
    prod = 0.0;
    MatMult(1, 1, ny, NULL, &Xp, &Yp, NULL, &Sp);
    if (DI!=0 || DJ!=0)
      MatFree(LM_ADR(Xp));
    SumMat(DM_Trans(Y), d_P0(DM_DM(Y)), d_P(DM_DM(Y)), 1, 1, &Sp);
     if (d_P(DM_DM(Y)) > 1)
      BrdMat(DM_Trans(Y), d_P0(DM_DM(Y)), 1, 1, &prod, 1);
  } 
  if (d_P(DM_DN(Y)) > 1)
    BrdMat(1-DM_Trans(Y), d_P0(DM_DN(Y)), 1, 1, &prod, 1); 
  TRACE_RETURN(prod);  
} /* D_DOT() */

#endif /* COMP_D_DOT */


#ifdef COMP_D_NRM2

FDATA D_NRM2 (N, X)
  int N; DistMat X;	
{ LocMat Xp, XpT, Sp;			FDATA prod;
  double sqrt();
  int ParamErr = 0;

  if (!NoOuterErrCheck) 
    CheckDim(ParamErr, N, DM_Inc(1-Tr(X), X), DM_DM(X), 0x12);
    
  SetUpComm(X);  
  ERROR_CHECK_TR(
    (TraceStr, "(%d,%s)", N, DM_TRACE(0, X)),
    ParamErr, "NRM2", 1
  );
  
  if (d_P0(DM_DN(X))==0) {
    int n = G2Ld(N, DM_DM(X));
    LM_Copy(DM_Trans(X), Xp, DM_Loc(X));	/* Xp is ny x 1 */
    LM_Copy(1-DM_Trans(X), XpT, DM_Loc(X));	/* XpT is 1 x ny */
    LM_Init(Sp, &prod, 1, 1);
    prod = 0.0;
    MatMult(1, 1, n, NULL, &XpT, &Xp, NULL, &Sp);
    SumMat(DM_Trans(X), d_P0(DM_DM(X)), d_P(DM_DM(X)), 1, 1, &Sp);
    if (d_P(DM_DM(X)) > 1)
      BrdMat(DM_Trans(X), d_P0(DM_DM(X)), 1, 1, &prod, 1);
    prod = (FDATA) sqrt((double) prod); 
  } 
  if (d_P(DM_DN(X)) > 1)
    BrdMat(1-DM_Trans(X), d_P0(DM_DN(X)), 1, 1, &prod, 1); 
  TRACE_RETURN(prod);  
} /* D_NRM2() */

#endif /* COMP_D_NRM2 */


#ifdef COMP_D_ASUM
#include "FDATACellVector.h"		/* VecAbsSum() */

FDATA D_ASUM (N, X) 
  int N; DistMat X; 
{ FDATA asum = 0.0;
  int ParamErr = 0;

  if (!NoOuterErrCheck)  
    CheckDim(ParamErr, N, DM_Inc(1-Tr(X), X), DM_DM(X), 0x12);
   
  SetUpComm(X);
  ERROR_CHECK_TR(
    (TraceStr, "(%d,%s)", N, DM_TRACE(0,X)),
    ParamErr, "ASUM", 1
  );
  
  if (d_P0(DM_DN(X))==0) {
    int IncX = DM_Inc(DM_Trans(X), X);	LocMat Ap;
    asum = VecAbsSum(G2Ld(N, DM_DM(X)), LM_ADR(DM_Loc(X)), IncX); 
    LM_Init(Ap, &asum, 1, 1);
    SumMat(DM_Trans(X), d_P0(DM_DM(X)), d_P(DM_DM(X)), 1, 1, &Ap);
    BrdMat(DM_Trans(X), d_P0(DM_DM(X)), 1, 1, &asum, 1);
  } 
  if (d_P(DM_DN(X)) > 1)
    BrdMat(1-DM_Trans(X), d_P0(DM_DN(X)), 1, 1, &asum, 1);      
  TRACE_RETURN(asum);
} /* D_ASUM() */

#endif /* COMP_D_ASUM */


#ifdef COMP_D_AMAX
#include "FDATACellVector.h"            /* VecAbsMaxPos() */

int D_AMAX (N, X, Max) 
  int N; DistMat X; FDATA *Max;
{ FDATA amax[2];			int imax = -1;
  int ParamErr=0;
  FDATA MsgMaxAbsValTmp = MsgMaxAbsVal;

  if (!NoOuterErrCheck)   
    CheckDim(ParamErr, N, DM_Inc(1-Tr(X), X), DM_DM(X), 0x12);
    
  SetUpComm(X);
  ERROR_CHECK_TR(
    (TraceStr, "(%d,%s,_)", N, DM_TRACE(0, X)),
    ParamErr, "AMAX", 1
  );

  MsgMaxAbsVal = 0.0; 		
  /* suppress received element data element checking in FDATACommun;
     here we are sending index position in messages, which may be
     > expected value of any matrix element */

  if (d_P0(DM_DN(X))==0) {
    int nx = G2Ld(N, DM_DM(X));		int IncX = DM_Inc(DM_Trans(X), X);
    imax = VecAbsMaxPos(nx, LM_ADR(DM_Loc(X)), IncX);
    if (imax==-1) {			/* assert nx = 0 */
      amax[0] = 0.0;			imax = 0;
    } else  
      amax[0] = *(LM_ADR(DM_Loc(X)) + FAST_MUL2(imax, IncX));
    imax = L2Gd(imax, DM_DM(X));	/* convert to a global index */
    AbsMaxElt(DM_Trans(X), &amax[0], &imax);
    if (amax[0] == 0.0)			/* take top element as being the max */
      imax = 0;
    amax[1] = (FDATA) imax;  
/*  BrdMat(DM_Trans(X), d_P0(DM_DM(X)), 1, 2, amax, 2);*/
  }
  if (d_P(DM_DN(X)) > 1)
    BrdMat(1-DM_Trans(X), d_P0(DM_DN(X)), 1, 2, amax, 2); 
  *Max = amax[0];    

  MsgMaxAbsVal = MsgMaxAbsValTmp;  	/* restore value for FDATACommun */

  TRACE_RETURN((int) amax[1]);        
} /* D_AMAX() */
#endif /*COMP_D_AMAX*/
 
/********************** new parallel level 1 procedures  ********************/

#ifdef COMP_D_ScaleAdd
#include "FDATADistMatrix.h"		/* TransMat(), RotMat()  */

void D_ScaleAdd (M, N, alpha, A, beta, C)
  int M; int N; FDATA alpha; DistMat A; FDATA beta; DistMat C;
{ int m = -1, n = -1;  
  int Beta0 = beta==FD_0;
  int ParamErr=0;                       
  int AlignM, AlignN;
  AlphaBetaDecln();

  if (!NoOuterErrCheck) {
    CheckAlign(ParamErr, AlignM, M, DM_Inc(1-Tr(A), A),   DM_DM(A),
               DM_Inc(1-Tr(C), C),   DM_DM(C), 0x146);
    CheckParam(ParamErr, AlignM && P_Same(A, C), 0x46);
    CheckAlign(ParamErr, AlignN, N, DM_Inc(Tr(A), A),     DM_DN(A),
               DM_Inc(Tr(C), C), DM_DN(C), 0x245);
  }

  SetUpComm(C);		
  ERROR_CHECK_TR(
    (TraceStr, "(%dx%d,%1.1f,%s,%1.1f,%s)",
                M, N, alpha, DM_TRACE(0,A), beta, DM_TRACE(1, C)),
    ParamErr, "ScaleAdd", 1
  );

  AlphaSet(alpha, pAlpha);              BetaSet(beta, pBeta);
  if (!Beta0 && beta!=1.0) {	
    m = G2Ld(Tr(C)? N: M, DM_D0(C));	n = G2Ld(Tr(C)? M: N, DM_D1(C));  
    MatScale(m, n, pBeta, DM_pLoc(C));
  }
  if (DM_Trans(A)==DM_Trans(C)) {
    if (DM_Trans(C)) 
      SWAP(int, M, N);    	
    if (m==-1) {
      m = G2Ld(M, DM_D0(C));    	n = G2Ld(N, DM_D1(C));
    } 
    if (d_P0(DM_D0(A))==d_P0(DM_D0(C)) && d_P0(DM_D1(A))==d_P0(DM_D1(C)))
      (Beta0? MatCopy: MatAdd)(m, n, DM_pLoc(A), pAlpha, DM_pLoc(C));
    else {
      int D0_dP0 = d_P0(DM_D0(A))-d_P0(DM_D0(C)),
          D1_dP0 = d_P0(DM_D1(A))-d_P0(DM_D1(C));
      int mA = D0_dP0?    G2Ld(M, DM_D0(A)): m,
          m_ = D0_dP0? G2Ld_Mx(M, DM_D0(A)): m,
          nA = D1_dP0?    G2Ld(N, DM_D1(A)): n,
          n_ = D1_dP0? G2Ld_Mx(N, DM_D1(A)): n;
      LocMat lA;			LM_Init(lA, NULL, n_, 1);
      MatMalloc_F("ScaleAdd", 1, &LM_ADR(lA), m_*n_);    	
      MatCopy(mA, nA, DM_pLoc(A), pAlpha, &lA);
      RotMat(0, D0_dP0, m_, &UnitDim, n_, LM_ADR(lA));
      RotMat(1, D1_dP0, m_, &UnitDim, n_, LM_ADR(lA));
      (Beta0? MatCopy: MatAdd)(m, n, &lA, NULL, DM_pLoc(C));     
      MatFree(LM_ADR(lA));
    }      
  } else if (M >= N) {			/* partition along longest dim. */
    LM_Trans(1-DM_Trans(A), DM_Loc(A));	LM_Trans(1-DM_Trans(C),  DM_Loc(C));
    TransMat(1-DM_Trans(A), 1-Beta0, N, M, DM_pLoc(A), DM_DN(A), DM_DM(A), 1, 
 	     pAlpha, DM_pLoc(C), d_P0(DM_DM(C)), d_P0(DM_DN(C)), 1);
    LM_Trans(1-DM_Trans(A), DM_Loc(A));	LM_Trans(1-DM_Trans(C),  DM_Loc(C));
  } else {
    LM_Trans(DM_Trans(A), DM_Loc(A));	LM_Trans(DM_Trans(C), DM_Loc(C));
    TransMat(DM_Trans(A), 1-Beta0, M, N, DM_pLoc(A), DM_DM(A), DM_DN(A), 1, 
 	     pAlpha, DM_pLoc(C), d_P0(DM_DN(C)), d_P0(DM_DM(C)), 1); 
    LM_Trans(DM_Trans(A), DM_Loc(A));	LM_Trans(DM_Trans(C), DM_Loc(C));
  }	     
  TRACE_EXIT; 	    
} /* D_ScaleAdd() */
        
#endif /* COMP_D_ScaleAdd */


#ifdef COMP_D_GetElt

FDATA D_GetElt (A) 
  DistMat A;
{ FDATA v;
  int ParamErr=0;

  if (!NoOuterErrCheck) 
    CheckParam(ParamErr, DM_OK(A), 0x1);
 
  SetUpComm(A);
  ERROR_CHECK_TR( (TraceStr, "(%s)", DM_TRACE(0, A)), ParamErr, "GetElt", 2);

  if (!d_P0(DM_D0(A)) && !d_P0(DM_D1(A))) 
    v = *LM_ADR(DM_Loc(A));
  if (d_P(DM_D0(A))>1)  
    BrdMat(0, d_P0(DM_D0(A)), 1, 1, &v, 1);
  if (d_P(DM_D1(A))>1)  
    BrdMat(1, d_P0(DM_D1(A)), 1, 1, &v, 1);
  TRACE_RETURN(v);  
} /* D_GetElt() */

#endif /* COMP_D_GetElt */


#ifdef COMP_D_SetElt
void D_SetElt (A, v)
  DistMat A; FDATA v;
{ int ParamErr=0;

  if (!NoOuterErrCheck)
    CheckParam(ParamErr, DM_OK(A), 0x1);
  
  ERROR_CHECK_TR(
    (TraceStr, "(%s,%1.1f)", DM_TRACE(0,A), v),
    ParamErr, "SetElt", 2
  );

  if (!d_P0(DM_D0(A)) && !d_P0(DM_D1(A))) 
    *LM_ADR(DM_Loc(A)) = v;    
  TRACE_EXIT;
} /* D_SetElt() */

#endif /* COMP_D_SetElt */


#ifdef COMP_D_CellHolds

int D_CellHolds (M, N, A) 
  int M; int N; DistMat A; 
{ int ParamErr=0;

  if (!NoOuterErrCheck) {
    CheckParam(ParamErr, DimOK(M), 1.0);
    CheckParam(ParamErr, DimOK(N), 2.0);
    CheckParam(ParamErr, DM_OK(A), 0x3);
  }

  ERROR_CHECK_TR(
    (TraceStr, "(%dx%d,%s)", M, N, DM_TRACE(0, A)),
    ParamErr, "CellHolds", 2
  );
    
  TRACE_RETURN ((d_P0(DM_DM(A)) < NBLKSd(M, DM_DM(A))) && 
  	        (d_P0(DM_DN(A)) < NBLKSd(N, DM_DN(A))));
} /* D_CellHolds() */

#endif /*COMP_D_CellHolds*/
