/* @(#) All Right Reserved, Copyright (c) 1997 Australian National University */
/* @(#)                     Copyright (c) 1997 Fujitsu Laboratories Ltd. */
/*  */
/*        */
/*                        Permission Notice */
/*  */
/*                Australian National University */
/*                Canberra ACT 0200 Australia */
/*                           and */
/*                Fujitsu Laboratories Ltd. */
/*                4-1-1 Kamikodanaka Nakahara */
/*                Kawasaki 211-88 Japan */
/*  */
/*   Permission to use, copy, modify and distribute this software (including  */
/*   its documentation) without fee is hereby granted under the following terms */
/*   and conditions: */
/*  */
/*   1. This permission notice and all of the copyright notices in this */
/*      software must appear in all copies and derived versions of all */
/*      or a part of this software. */
/*  */
/*   2. Sales of this software, whether incorporated in other products as a */
/*      part thereof or not, is prohibited. */
/*  */
/*   3. On publication or presentation in any form of the results obtained */
/*      by use of this software, an explicit reference to this software and */
/*      its copyright owner shall be made. */
/*  */
/*   4. THIS SOFTWARE IS PROVIDED 'AS IS' AND THE AUSTRALIAN NATIONAL UNIVERSITY */
/*      AND FUJITSU LABORATORIES LTD. MAKE NO WARRANTIES, EITHER EXPRESSED OR */
/*      IMPLIED, AS TO ANY MATTER WHATSOEVER REGARDING THIS SOFTWARE,  */
/*      INCLUDING, WITHOUT LIMITATION, WARRANTIES OF MERCHANTABILITY OR FITNESS */
/*      FOR ANY PARTICULAR PURPOSE. */
/*      IN NO EVENT SHALL THE AUSTRALIAN NATIONAL UNIVERSITY OR FUJITSU */
/*      LABORATORIES LTD. BE LIABLE FOR ANY CLAIM FROM A THIRD PARTY, OR */
/*      FOR SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES WHATSOEVER RESULTING */
/*      FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, */
/*      NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION */
/*      WITH THE USE, COPYING OR MODIFICATION OF THIS SOFTWARE. */
/*  */

/************************************************************************/
/*	DistMat: distributed block-cyclic matrices definition		*/
/************************************************************************/
/* SCCS INFO @(#)DistMat.h 4.4         last modified P. Strazdins, 97/04/18
   first created: P. Strazdins 94/05/17
*/
/* contains:
 *	* distributed matrix dimension and structure definitions
 *	* macros/procedures for computing cell dimensions,
 *	  converting local <-> global indices etc.
 *	* getting sub-matrices of a distributed matrix structure
 *  
 *  compilation: macro FDATATYPE must be defined;
 *		 if CC defined, compile for single 1x1 cell (workstation)
 *		 (some optimiations can be performed)
 */

#ifndef __DistMat_INCL__
#define __DistMat_INCL__

#include "FDATASetUp.h"
#include "LocMat.h"
#include "MatDim.h"
#include "GenAssert.h"
#include "FDATACellMatrix.h"

/********************* exported symbols  ***********************************/

#if FDATATYPE==SINGLE
#define InitDimMat_	DSInitDimMat_
#define SubMat_		DSSubMat_
#define NewMat_		DSNewMat_
#define ReduceDimMat_	DSReduceDimMat_
#define Spread2DistMat_ DSSpread2DistMat_
#define PrintDistMat	DSPrintDistMat
#else
#define InitDimMat_	DSInitDimMat_
#define SubMat_		DDSubMat_
#define ReduceDimMat_	DDReduceDimMat_
#define NewMat_		DDNewMat_
#define Spread2DistMat_ DSSpread2DistMat_
#define PrintDistMat	DDPrintDistMat
#endif 

/********************* structure definitions ******************************/

struct DistMatStruct {		/* cell component of a distributed matrix */
				/* using block scattered storage */
  MatDim D[2];			/* cell row & column storage descriptors */
  				/* nb. this should force double alignment */
  LocMat L;			/* local cell matrix descriptor */
};
				
typedef struct DistMatStruct DistMatStore, *DistMat;
				/* Transpose bit implemented in lsb of the pointer */
				
#define DM_Ptr(A)	((DistMat) ((unsigned) (A) & ~0x1))
#define DM_Trans(A)	((unsigned) (A) & 0x1)

/**************** DistMat selectors *********************************/ 

#define DM_Loc(A) 	(DM_Ptr(A)->L)
#define DM_pLoc(A) 	(&DM_Loc(A))
#define DM_D(t, A)	(&(DM_Ptr(A)->D[t]))
#define DM_D0(A)	DM_D(0, A)
#define DM_D1(A)	DM_D(1, A)
#define DM_DM(A)	DM_D(DM_Trans(A), A)
#define DM_DN(A)	DM_D(1-DM_Trans(A), A)

/**************** DistMat constructors *********************************/

/* InitDimMat(), SubMat(), ReduceDimMat(), Spread2DistMat() and NewMat() are 
 * the main constructors. To make these calls efficient (minimize structure 
 * copying) and yet clean to use, the DistMat_RVAlloc(n); macro must be   
 * called at the start of any C block where these macros appear.
 * The value of n must be at least the number of these calls
 * within that block (excluding any nested blocks with their own
 * DistMat_RVAlloc(n')). However, with ASSERT defined, this can be checked.
 * The bodies of any loops nested within this block MUST have
 * a suitable DistMat_RVAlloc(n'); call at the block starting the loop.
 * This is enough to ensure that any such call gets a unique 
 * element of DM_RV_[]. 
 */

#define DistMat_RVAlloc(n)	DistMatStore DM_RV_[n]; \
				int DM_RV_ix = 0, DM_RV_mx = (n) 

#ifdef ASSERT
#define	DistMat_RVOK		COMMON_ASSERT(DM_RV_ix<DM_RV_mx, 0, "DistMat_RV_Allocation", 10, 0)
#else
#define	DistMat_RVOK		1
#endif
#define DistMat_RVNext		(&DM_RV_[DM_RV_ix++])

#define InitDimMat(Trans, B_0, B_1, P0_0, P_0, P0_1, P_1, di, dj)	\
				(DistMat_RVOK, 				\
				InitDimMat_(DistMat_RVNext, 		\
					    Trans, B_0, B_1, P0_0, P_0, P0_1, P_1, di, dj))

#define SubMat(A, i, j)		(DistMat_RVOK, 				\
				SubMat_(DistMat_RVNext, A, i, j))
		
#define ReduceDimMat(RowCol, A)	(DistMat_RVOK, 				\
				ReduceDimMat_(DistMat_RVNext, RowCol, A))

#define Spread2DistMat(RowCol, Trans, DN, sB)				\
				(DistMat_RVOK, 				\
				 Spread2DistMat_(DistMat_RVNext, RowCol, Trans, DN, sB))

		
#define NewMat(M, N, A, WS)	(DistMat_RVOK, 				\
				NewMat_(DistMat_RVNext, M, N, A, WS))

DistMat InitDimMat_ STDC_PARAMS( 
	(DistMat A, int Trans, int B_0, int B_1, int P0_0, int P_0, int P0_1,
	 int P_1, int di, int dj) 
);
/* post: returns A = SubMat({TransB, _, {0, B_0, P0_0, P_0}, {0, B_1, P0_1, P_1}}, di, dj)
 *	 local matrix part of A is not initialized
 */	


DistMat SubMat_ STDC_PARAMS( (DistMat RV, DistMat A, int i, int j) );
/* post: returns RV, where RV is the distributed Matrix structure 
 *	 corresponding to global A[i, j]
 */	

#define SubMatTr( Tr, A, i, j ) SubMat(A, (Tr)?(j):(i), (Tr)?(i):(j)) 	

#define	TransSubMat(A, i, j)	Trans(SubMat(A, i, j))
#define Trans(A)		((DistMat) ((unsigned) (A) ^ 0x1))
#define TransposeMat(A)		Trans(A)
#define DM_SetTrans(t, A)	A = ((DistMat) ((t)? ((unsigned) (A) |  0x1):\
						     ((unsigned) (A) & ~0x1)))
#define DM_Transpose(A) 	(A) = Trans(A)



#define DM_DeclareAlloc(sA, A)	struct DistMatStruct sA;  DistMat A = &sA
#define DM_Init(A, t, lA, IB0, IB, IP0, IP, JB0, JB, JP0, JP)		\
	{ pInitDim(DM_D0(A), IB0, IB, IP0, IP);	DM_Loc(A) = (lA); 	\
	  pInitDim(DM_D1(A), JB0, JB, JP0, JP); DM_SetTrans(t, A); }


DistMat ReduceDimMat_ STDC_PARAMS( 
	(DistMat RV, int RowCol, DistMat A) 
);
/* pre: RowCol = 0, 1, -1
 * post: returns RV where RV = A with the processor sub-field of the dimension 
 *	 RowCol!=DM_Trans(A) reduced to 1. If RowCol==-1, both dimensions
 *	 are reduced 
 */	


#define MkVec(RC, A)		ReduceDimMat(RC, A)
#define RowVec(A)		MkVec(0, A)
#define ColVec(A)		MkVec(1, A)
#define UnitVec(A)		MkVec(-1, A)


DistMat Spread2DistMat_ STDC_PARAMS(
        (DistMat B, int RowCol, int Trans, pMatDim DN, pLocMat sB)
);
/* post: returns B st: DM_Loc(B)=sB, DM_Trans(B)=Trans,
 *       DM_Dim(!RowCol, B) = DN,  DM_Dim(RowCol, B) = {0,1,0,1}
 *       This is the DistMat representation of a spread matrix.
 */
 

DistMat NewMat_ STDC_PARAMS( 
	(DistMat RV, int M, int N, DistMat A, FDATA *WSA)
);
/* pre:  DistMatOK(M, N, A)
 * post: returns A' where A`=A except local storage  of A
 *       points to WSA' and strides correspond to this cell's local
 *       portions of M x N matrix A.
 *       where WSA' = WSA!=NULL? WSA: <malloc of an appropriately sized area>
 */
        

/****************** DistMat selector macros ******************************/

#define DisposeMat(A) MatFree(LM_ADR(DM_Loc(A)))
/* pre:  cell storage of A has been dynamically allocated
 * post: cell storage of A has bewen freed. 
 * note: should be paired with a call to NewMat(..., NULL)
 */

/******************* high-level DistMat predicates **********************/
 
#define InVecScope(RC, A)	Dim_Owns(DM_D(((RC)^DM_Trans(A)), A))
#define InColScope(A)		InVecScope(1, A)
#define InRowScope(A)		InVecScope(0, A)
#define InEltScope(A)		(InRowScope(A) && InColScope(A))

/******************** defensive programming ***********************/

#define DM_OK(A)	/* check internal consistency of A */\
	(Dim_OK(DM_D(0, A)) && Dim_OK(DM_D(1, A)))

char *PrintDistMat STDC_PARAMS( ( DistMat A, char *pA ) );
/* post: as for sprintf(pA, DM_SPRINTF_FORMAT, DM_SPRINTF_PARAMS(A)); returns pA
 */
					/* shorthands for tracing */
#define DM_SPRINTF_FORMAT	"{%d,a%d %d %d,%d %d:%d %d,%d %d:%d %d}" 
#define DM_SPRINTF_PARAMS(A)						\
	DM_Trans(A), 							\
	FD_ADR(LM_ADR(DM_Loc(A))), LM_Inc(0, DM_Loc(A)), LM_Inc(1, DM_Loc(A)),\
	d_B0(DM_D0(A)), d_B(DM_D0(A)), d_P0(DM_D0(A)), d_P(DM_D0(A)),\
	d_B0(DM_D1(A)), d_B(DM_D1(A)), d_P0(DM_D1(A)), d_P(DM_D1(A))

#endif /*__DistMat_INCL__*/
