/* @(#) All Right Reserved, Copyright (c) 1997 Australian National University */
/* @(#)                     Copyright (c) 1997 Fujitsu Laboratories Ltd. */
/*  */
/*        */
/*                        Permission Notice */
/*  */
/*                Australian National University */
/*                Canberra ACT 0200 Australia */
/*                           and */
/*                Fujitsu Laboratories Ltd. */
/*                4-1-1 Kamikodanaka Nakahara */
/*                Kawasaki 211-88 Japan */
/*  */
/*   Permission to use, copy, modify and distribute this software (including  */
/*   its documentation) without fee is hereby granted under the following terms */
/*   and conditions: */
/*  */
/*   1. This permission notice and all of the copyright notices in this */
/*      software must appear in all copies and derived versions of all */
/*      or a part of this software. */
/*  */
/*   2. Sales of this software, whether incorporated in other products as a */
/*      part thereof or not, is prohibited. */
/*  */
/*   3. On publication or presentation in any form of the results obtained */
/*      by use of this software, an explicit reference to this software and */
/*      its copyright owner shall be made. */
/*  */
/*   4. THIS SOFTWARE IS PROVIDED 'AS IS' AND THE AUSTRALIAN NATIONAL UNIVERSITY */
/*      AND FUJITSU LABORATORIES LTD. MAKE NO WARRANTIES, EITHER EXPRESSED OR */
/*      IMPLIED, AS TO ANY MATTER WHATSOEVER REGARDING THIS SOFTWARE,  */
/*      INCLUDING, WITHOUT LIMITATION, WARRANTIES OF MERCHANTABILITY OR FITNESS */
/*      FOR ANY PARTICULAR PURPOSE. */
/*      IN NO EVENT SHALL THE AUSTRALIAN NATIONAL UNIVERSITY OR FUJITSU */
/*      LABORATORIES LTD. BE LIABLE FOR ANY CLAIM FROM A THIRD PARTY, OR */
/*      FOR SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES WHATSOEVER RESULTING */
/*      FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, */
/*      NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION */
/*      WITH THE USE, COPYING OR MODIFICATION OF THIS SOFTWARE. */
/*  */

/************************************************************************/
/*       	FDATACellMatrix: FDATA Matrix procedures for cells    	*/
/************************************************************************/
/* SCCS INFO @(#)FDATACellMatrix.c 4.6         last modified P. Strazdins, 97/05/22
   first created: P. Strazdins 93/03/23 
*/
/* Set of support matrix routines used for DBLAS, independent of distributed
 * memory issues.
 * Efficiency: the use of the following techniques for scalar arhcitectures 
 * are used (if EXT_FTN_BLAS and VANILLA not defined):
 *  	1. use of a local *FDATA pointer (eg ax) to access a column of a matrix;
 *	   incrementing by last dimension avoids expensive int multiply
 *	2. for procedures scaling by FDATA *alpha, the macro
 *		FD_ALPHA_REGS(*alpha)
 *	   sets up fpu registers (as many as FDATA needs) to store *alpha
 *	   then, the other FD_ALPHA... macros use these registers
 *	   to perform fast scaling
 *	3. the matrices are MxN; these procedures typically iterate
 *	   over these dimensions; if M_INNER(...), the order of the loops
 *	   are interchanged
 *	4. 4 way loop unrolling and software pipelining where appropriate   
 *    generally, the SPARC C compiler does well with these optimizations.
 * compilation: macro FDATATYPE must be defined;
 *
 *              if macro VANILLA defined, `vanilla' versions of the routines
 *              are used (slower, but useful for debugging or vectorization.
 *              if macro NO_PARTITION is defined, suppress partitoning methods
 *              wherever possible    
 *              if macro EXT_FTN_BLAS defined, Fortran BLAS is used whenever
 *              possible; if EXT_FTN_BLAS_3 defined, do so for only level 3 call
 */
#ifdef TEST_MALLOC
# include <stdio.h>
# define CONST
# include "test/malloctest/mem_man.h"
  /* useful to call: mem_write_verbose(stdout);*/
#else 
# include <stdlib.h>			/* memalign(), free() */
#endif 

#include "FDATACellMatrix.h"
#include "BLASOptions.h"
#include "FDATAOps.h"
#include "FDATAAssert.h"

#include "MachineParams.h"
	
/********************** matrix versions of level 1 routines *****************/

void MatZero(M, N, A)
  int M, N; pLocMat A;
{ int i, RIncA=pLM_RInc(A), CIncA=pLM_CInc(A); FDATA *Ax = pLM_ADR(A);
  COMMON_ASSERT_TR(
    (TraceStr, "(%dx%d,a%d %d %d)", M, N, FD_ADR(Ax), RIncA, CIncA),
    IncOK(M, N, RIncA, CIncA), 0, "MatZero", 4, TRACE_COMPUTE
  );
  if M_INNER(M, N, RIncA, CIncA) { /* interchange so larger index on inner loop*/
      SWAP(int, M, N); SWAP(int, RIncA, CIncA);
  }
#ifdef VANILLA
  for (i = 0; i < M; i++) {
    int j; register FDATA *ax = Ax;	  	  
    for (j = 0; j < N; j++) {
      FD_SET_REAL(*ax, 0.0);
      FD_SET_IMAG(*ax, 0.0);
      ax += CIncA;
    }
    Ax += RIncA;
  } /* for (i...) */
#elif defined(EXT_FTN_BLAS)
    i=0;
    if (M==0 || N==0)		/* avoid possible illegal strides in FTN_GEMM()*/
      return;
    if (RIncA==1) {
      if (CIncA==1 && N==1)
        CIncA = M;	    
      FTN_GEMM("N", "T", &M, &N, &i, &FD_0, Ax, &M, Ax, &N, &FD_0, Ax, &CIncA);
    } else if (CIncA==1)
      FTN_GEMM("N", "T", &N, &M, &i, &FD_0, Ax, &N, Ax, &M, &FD_0, Ax, &RIncA);
    else    
      for (i=0; i < M; i++) {
      	/* bug in the vanilla blas: if A[i,j]=NaN before call, A[i,j]=NaN after.
      	   ie. its tries to mulitply A[i,j] by 0.0 */
        FTN_SCAL(&N, &FD_0, Ax, &CIncA);      
        Ax += RIncA;
      } /* for (i...) */
#else    
  {
  register int CIncA2=CIncA+CIncA, CIncA3=CIncA2+CIncA, CIncA4=CIncA2+CIncA2;
  register int n=4*(N/4), j; 
  for (i = 0; i < M; i++) {
    register FDATA *ax = Ax;
    for (j=0; j < n; j+= 4) {       /*4 way loop unrolling*/
      FD_SET_REAL(*(ax), 0.0);		FD_SET_IMAG(*(ax), 0.0);
      FD_SET_REAL(*(ax+CIncA), 0.0);	FD_SET_IMAG(*(ax+CIncA), 0.0);
      FD_SET_REAL(*(ax+CIncA2), 0.0);	FD_SET_IMAG(*(ax+CIncA2), 0.0);
      FD_SET_REAL(*(ax+CIncA3), 0.0);	FD_SET_IMAG(*(ax+CIncA3), 0.0);
      ax += CIncA4;	
    }       
    for (j=n; j < N; j++) {
      FD_SET_REAL(*ax, 0.0);      	FD_SET_IMAG(*ax, 0.0);
      ax += CIncA;
    }
    Ax += RIncA;
  } /* for (i...)*/
  } 
#endif /*VANILLA*/
} /* MatZero() */


void MatScale(M, N, alpha, C)
  int M, N; FDATA *alpha; pLocMat C; 
{ int i; int RIncC=pLM_RInc(C), CIncC=pLM_CInc(C); FDATA *Cx = pLM_ADR(C);
  COMMON_ASSERT_TR_E(
    (TraceStr, "(%dx%d,%1.1f,a%d %d %d)", 
     M, N, FD_REAL(pFD_VAL(alpha)), FD_ADR(Cx), RIncC, CIncC),
     IncOK(M, N, RIncC, CIncC), 0, "MatScale", 3, TRACE_COMPUTE
  );

  if (pFD_UNITY(alpha)) {			/* fast exit*/
    TRACE_RETURN;
  } else if (FD_EQ(*alpha, FD_0))		/* *alpha==0; fpu can be avoided */
    MatZero(M, N, C);
  else {
    if (M_INNER(M, N, RIncC, CIncC)) {/* interchange so larger dim. on inner loop*/
      SWAP(int, M, N);     SWAP(int, RIncC, CIncC);
    }
#if defined(VANILLA) 
    for (i=0; i < M; i++) {
      register FDATA a = pFD_VAL(alpha);
      register FDATA *cx = Cx; int j;	
      for (j=0; j < N; j++) {
        FD_SETOP(*cx, a, *, *cx);
        cx += CIncC;
      }  
      Cx += RIncC;
    } /* for (i...) */
#elif defined(EXT_FTN_BLAS)
    for (i=0; i < M; i++) {
      FTN_SCAL(&N, alpha, Cx, &CIncC);      
      Cx += RIncC;
    } /* for (i...) */
#else
    {
    register int j, n, CIncC2, CIncC3, CIncC4, CIncC5, CIncC6;
    register FDATA t0, t1, t2, t3, c0, c1, c2, c3; 	
    register FDATA a = pFD_VAL(alpha);
     
    CIncC2=CIncC+CIncC; CIncC3=CIncC2+CIncC; CIncC4=CIncC2+CIncC2;
    CIncC5=CIncC3+CIncC2; CIncC6=CIncC4+CIncC2;
    n = 4*(N/4);

    for (i=0; i < M; i++) {
      register FDATA *cx = Cx;	
      if (n >= 4) {
                t0 = *(cx);    
        j=4;
                t1 = *(cx+CIncC);
                                FD_SETOP(c0, a, *, t0);
                t2 = *(cx+CIncC2);      
                                FD_SETOP(c1, a, *, t1);
        if (j < n) do {         	
                t3 = *(cx+CIncC3);
                                FD_SETOP(c2, a, *, t2);
                                                        *(cx) = c0;
                t0 = *(cx+CIncC4);    
                                FD_SETOP(c3, a, *, t3);
                                                        *(cx+CIncC) = c1;
          j += 4;                                        
                t1 = *(cx+CIncC5);
                                FD_SETOP(c0, a, *, t0);
                                                        *(cx+CIncC2) = c2;
                t2 = *(cx+CIncC6);
                                FD_SETOP(c1, a, *, t1);
                                                        *(cx+CIncC3) = c3;
	  cx += CIncC4;		
        } while (j < n);
		t3 = *(cx+CIncC3);
				FD_SETOP(c2, a, *, t2);
							*(cx) = c0;
				FD_SETOP(c3, a, *, t3);
							*(cx+CIncC) = c1;
							*(cx+CIncC2) = c2;
							*(cx+CIncC3) = c3;
	  cx += CIncC4;		
      } /* if (n...) */  
      for (j=n; j < N; j++) {
	  	t0 = *(cx); 	FD_SETOP(c0, a, *, t0);	*(cx) = c0;
	  cx += CIncC;
      } 
      Cx += RIncC;
    } /* for (i...)*/
    }
#endif /*VANILLA*/
  }
  TRACE_EXIT;
} /* MatScale() */
            

FDATA *PackMMOpds(M, N, K, alphaf, A, B)
  int M, N, K; FDATA **alphaf; pLocMat *A, *B; 
{ int PackB = LM_PackUseful(K, N, **B, M), 
      PackA = LM_PackUseful(M, K, **A, N);
  static LocMat A1, B1; 
  FDATA *WS = NULL, *Ax = NULL, *Bx = NULL;
  
  if (PackA && PackB && MatMalloc_FT("MatPartMult", 0, &WS, K*(M+N))) {
    Bx = WS+M*K;
    Ax = WS;	
  } else if (((PackA && K<=M) || (!PackB && M<N && !pFD_UNITY(*alphaf) && K<=N))
    	      && MatMalloc_FT("MatPartMult", 0, &WS, K*M))  {
    Ax = WS;  
  } else if ((PackB || (!pFD_UNITY(*alphaf) && K<=N && M>0)) 
  	     && MatMalloc_FT("MatPartMult", 0, &WS, K*N)) {
    Bx = WS;
  }  
  if (Bx != NULL) {
    LM_Init(B1, Bx, N, 1); 
    MatCopy(K, N, *B, *alphaf, &B1);
    *B = &B1;	    	*alphaf = NULL;
  } 
  if (Ax != NULL) {
    LM_Init(A1, Ax, 1, M); 	  
    MatCopy(M, K, *A, *alphaf, &A1);
    *A = &A1;     	*alphaf = NULL;
  }  

  return(WS);
} /* PackMMOpds() */


void MatPartMult(M, N, K, alpha, A, B, beta, C)
  int M, N, K; FDATA *alpha, *beta; pLocMat A, B, C;
{ COMMON_ASSERT_TR_E(
   (TraceStr, "(%dx%dx%d,%1.1f,a%d %d %d,a%d %d %d,%1.1f,a%d %d %d)", 
    M, N, K, pFD_VAL(alpha), FD_ADR(pLM_ADR(A)), pLM_RInc(A), pLM_CInc(A),
    FD_ADR(pLM_ADR(B)), pLM_RInc(B), pLM_CInc(B), pFD_VAL(beta),
    FD_ADR(pLM_ADR(C)), pLM_RInc(C), pLM_CInc(C)), 
   (N==0 || pLM_IncOKUnit(M, K, A)) && (M==0 || pLM_IncOKUnit(K, N, B)) &&
   pLM_IncOKUnit(M, N, C) && 0<=M && 0<=N && 0<=K, 
   MatOverlap(M, K, A, M, N, C) + 2*MatOverlap(K, N, B, M, N, C),
   "MatPartMult", 2, TRACE_COMPUTE
  );
#if defined( VANILLA ) || defined(NO_PARTITION)
  MatMult(M, N, K, alpha, A, B, beta, C);
#else
  {   
  int k; int mn = MIN(M, N); 
  LocMat A_, B_, C_; 			/* used to prevent side effects *A,..*/

  if (!(CopyEfficient(mn) && PartitionUseful(M, N, K))) {
    /* small computation - quick fall through */
    FDATA * WS = PackMMOpds(M, N, K, &alpha, &A, &B);
    MatMult(M, N, K, alpha, A, B, beta, C);
    TRACE_EXIT;
    if (WS != NULL)
      MatFree(WS);
    return;
  }
      
  if (M < N) {			/* make A (M) largest, B (N) smallest */
    SWAP(int, M, N);	
    LM_Copy(1, A_, *A);	LM_Copy(1, B_, *B);	LM_Copy(1, C_, *C);
    A = &B_;		B = &A_;		C = &C_;
  } else {
    LM_Copy(0, A_, *A);	LM_Copy(0, B_, *B);	
    A = &A_;		B = &B_;		
  }      

  k=0;
  while (k < K) {
    int dK; 
    int DoPartition = 0;
    FDATA *WS = NULL; int Mw, Ldw;	/* partitioning workspace parameters */
 
    dK = K - k;
    if (2*dK > 3*INNER_VECSIZE && PartitionUseful(M, N, dK))
      dK = INNER_VECSIZE;

    if (CopyEfficient(mn) && PartitionUseful(M, N, dK)) {
      /* try partitioning over M & N; a time-space tradeoff */
      Ldw = Mw = (CACHESIZE/2) / dK;   
      DoPartition = MatMalloc("MatPartMult", 0, &WS, CACHESIZE*((mn+Ldw-1)/Ldw));

      if (! DoPartition) {/* partition fairly efficiently with 1 CACHESIZE WS */
        Ldw = Mw = (CACHESIZE - MINOR_OUTER_VECSIZE*dK)/dK;
        DoPartition = - MatMalloc("MatPartMult", 0, &WS, CACHESIZE);
      }
    }
  
    if (! DoPartition ) {
      FDATA *alphaf = alpha; pLocMat Af = A, Bf = B;	
      WS = PackMMOpds(M, N, dK, &alphaf, &Af, &Bf);
      MatMult(M, N, dK, alphaf, Af, Bf, beta, C);
    } else if (DoPartition == -1) {
      int i, j, m, n; LocMat WA, WB, Aik, Bkj, Cij;
      LM_Init(WA, WS, 1, MINOR_OUTER_VECSIZE);
      LM_Init(WB, WS + MINOR_OUTER_VECSIZE*dK, Ldw, 1);
      for (j=0; j < N; j += Mw) {
        n = MIN(Mw, N-j);	    	 
        MatCopy(dK, n, pLocSubMat_(&Bkj, B, 0, j), alpha, &WB);
        pLocSubMat_(&Aik, A, 0, 0);	pLocSubMat_(&Cij, C, 0, j);
        for (i=0; i < M; i += MINOR_OUTER_VECSIZE) {
      	  m = MIN(MINOR_OUTER_VECSIZE, M-i);
          MatCopy(m, dK, &Aik, NULL, &WA);
          MatMult(m, n, dK, NULL, &WA, &WB, beta, &Cij);
          LM_SubMat(Aik, m, 0);		LM_SubMat(Cij, m, 0); 
        }
      } /* for (i...) */ 
    } else {   
      int i, m, n, j=0, jInc=Mw; LocMat WA, WB, Aik, Bkj, Cij; 
      LM_Init(WA, WS+(CACHESIZE/2), 1, Ldw);
      LM_Init(WB, WS, Ldw, 1);
      for (i=0; i < M; i+= Mw) {
        m = MIN(Mw, M-i);	
        MatCopy(m, dK, pLocSubMat_(&Aik, A, i, 0), NULL, &WA);
        for (; (0 <= j) && (j < N); j += jInc) {
          n = MIN(Mw, N-j);
          if (i==0) 	
            MatCopy(dK, n, pLocSubMat_(&Bkj, B, 0, j), alpha, &WB);
            MatMult(m, n, dK, NULL, &WA, &WB, beta, pLocSubMat_(&Cij, C, i, j));
            LM_ADR(WB) += (jInc>0)? CACHESIZE: -CACHESIZE;
        } /* for (j...) */
        j -= jInc;			jInc = -jInc; 
        LM_ADR(WB) += (jInc>0)? CACHESIZE: -CACHESIZE;
      } /* for (i...) */
    } /* if (DoPartition) */

    if (WS != NULL)
      MatFree(WS);

    k += dK;
    beta = NULL;		/* must be unity for next iteration */
    pLocSubMat(A, 0, dK);		pLocSubMat(B, dK, 0);	
  } /* while (k<K) */
  }
#endif /*VANILLA*/
  TRACE_EXIT;
} /* MatPartMult() */

#if defined(EXT_FTN_BLAS) || defined(VANILLA) || defined(EXT_FTN_BLAS_3)

void MatMult(M, N, K, alpha, A, B, beta, C)
  int M, N, K; FDATA *alpha, *beta;  pLocMat A, B, C;
{ int i;
  COMMON_ASSERT_TR_E(
   (TraceStr, "(%dx%dx%d,%1.1f,a%d %d %d,a%d %d %d,%1.1f,a%d %d %d)", 
    M, N, K, pFD_VAL(alpha), FD_ADR(pLM_ADR(A)), pLM_RInc(A), pLM_CInc(A),
    FD_ADR(pLM_ADR(B)), pLM_RInc(B), pLM_CInc(B), pFD_VAL(beta),
    FD_ADR(pLM_ADR(C)), pLM_RInc(C), pLM_CInc(C)),
   (N==0 || pLM_IncOKUnit(M, K, A)) && (M==0 || pLM_IncOKUnit(K, N, B)) &&
   pLM_IncOKUnit(M, N, C) && 0<=M && 0<=N && 0<=K, 
   MatOverlap(M, K, A, M, N, C) + 2*MatOverlap(K, N, B, M, N, C),
   "MatMult", 3, TRACE_COMPUTE
  );

  if (alpha==NULL) alpha = &FD_1;	if (beta==NULL) beta = &FD_1;		
#if defined( VANILLA )
  { int i, j, k; FDATA *Ax = pLM_ADR(A), *Cx = pLM_ADR(C);
  for (i = 0; i < M; i++) {
    FDATA *b = pLM_ADR(B), *cx = Cx;	
    for (j=0; j < N; j++) {
      FDATA *ax=Ax, *bx=b; FDATA c = FD_0;
      for (k=0; k < K; k++) {
        FD_MULSET(c, +=, *ax, *bx);
        ax += pLM_CInc(A); bx += pLM_RInc(B);
      }  
      FD_MULSET(*cx, =, *cx, *beta);
      FD_MULSET(*cx, +=, c, *alpha);
      b += pLM_CInc(B);  cx += pLM_CInc(C);
    } /* for (j...) */
    Cx += pLM_RInc(C); Ax+= pLM_RInc(A); 
  } /* for (i...) */ 
  }
#elif defined(EXT_FTN_BLAS) || defined (EXT_FTN_BLAS_3)
 { int LdA, LdB, LdC = pLM_Ld(C);
   int tC = (pLM_CInc(C) == 1); 
   char TrA, TrB;
#if 0
   MatPrint("B", K, N, *B);
   MatPrint("C", M, N, *C);
#endif  
   if (tC) {			/* interchange elements of computations */
     SWAP(int, M, N);
     SWAP(pLocMat, A, B);
     if (LdC==1 && N==1)	/* remove degenerate case */
       LdC = M;
   }  
   TrA = (pLM_Inc(tC, A)==1 && !(pLM_Inc(1-tC, A)==1 && K==1))? 
         NoTranspose: Transpose,
   TrB = (pLM_Inc(tC, B)==1 && !(pLM_Inc(1-tC, B)==1 && N==1))? 
         NoTranspose: Transpose;
   LdA = pLM_Ld(A), LdB = pLM_Ld(B);      
   if (K==0)
     LdA = TrA!=NoTranspose? 1: M, LdB = TrB!=NoTranspose? N: 1;

   COMMON_ASSERT_TR(
   	(TraceStr, "(%c,%c,%dx%dx%d,%1.1f,a%d %d,a%d %d,%1.1f,a%d %d)", 
  	 TrA, TrB, M, N, K, *alpha, FD_ADR(pLM_ADR(A)), LdA, 
  	 FD_ADR(pLM_ADR(B)), LdB, *beta, FD_ADR(pLM_ADR(C)), LdC),
     	1, 0, "GEMM", 4, TRACE_COMPUTE);

   if ((M>0)&(N>0)) {   
     FTN_GEMM(&TrA, &TrB, &M, &N, &K, alpha, pLM_ADR(A), 
   	      &LdA, pLM_ADR(B), &LdB, beta, pLM_ADR(C), &LdC);
   }	      
 }
# endif /*VANILLA*/
  TRACE_EXIT;
} /* MatMult() */
#else 
# include "FDATACellMatMult.c"
#endif  

