/* @(#) All Right Reserved, Copyright (c) 1997 Australian National University */
/* @(#)                     Copyright (c) 1997 Fujitsu Laboratories Ltd. */
/*  */
/*        */
/*                        Permission Notice */
/*  */
/*                Australian National University */
/*                Canberra ACT 0200 Australia */
/*                           and */
/*                Fujitsu Laboratories Ltd. */
/*                4-1-1 Kamikodanaka Nakahara */
/*                Kawasaki 211-88 Japan */
/*  */
/*   Permission to use, copy, modify and distribute this software (including  */
/*   its documentation) without fee is hereby granted under the following terms */
/*   and conditions: */
/*  */
/*   1. This permission notice and all of the copyright notices in this */
/*      software must appear in all copies and derived versions of all */
/*      or a part of this software. */
/*  */
/*   2. Sales of this software, whether incorporated in other products as a */
/*      part thereof or not, is prohibited. */
/*  */
/*   3. On publication or presentation in any form of the results obtained */
/*      by use of this software, an explicit reference to this software and */
/*      its copyright owner shall be made. */
/*  */
/*   4. THIS SOFTWARE IS PROVIDED 'AS IS' AND THE AUSTRALIAN NATIONAL UNIVERSITY */
/*      AND FUJITSU LABORATORIES LTD. MAKE NO WARRANTIES, EITHER EXPRESSED OR */
/*      IMPLIED, AS TO ANY MATTER WHATSOEVER REGARDING THIS SOFTWARE,  */
/*      INCLUDING, WITHOUT LIMITATION, WARRANTIES OF MERCHANTABILITY OR FITNESS */
/*      FOR ANY PARTICULAR PURPOSE. */
/*      IN NO EVENT SHALL THE AUSTRALIAN NATIONAL UNIVERSITY OR FUJITSU */
/*      LABORATORIES LTD. BE LIABLE FOR ANY CLAIM FROM A THIRD PARTY, OR */
/*      FOR SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES WHATSOEVER RESULTING */
/*      FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, */
/*      NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION */
/*      WITH THE USE, COPYING OR MODIFICATION OF THIS SOFTWARE. */
/*  */

/************************************************************************/
/*       	FDATACellMatrix: FDATA Matrix procedures for cells    	*/
/************************************************************************/
/* SCCS INFO @(#)FDATACellMatrix.h 4.3         last modified P. Strazdins, 97/02/19
   first created: P. Strazdins 93/03/23 
*/
/* Set of support matrix routines used for DBLAS, independent of distributed
 * memory issues.
 * Most of these can be implemented via level 1 BLAS, although for convenience
 * and efficiency, they are implemented as proper matrix operations,
 * using local matrix dimensions and strides.
 * These refer to local cell view of matrices only with generalized row-column 
 * strides (see LocMat.h).
 *
 */
#ifndef __FDATACellMatrix_INCL__
#define __FDATACellMatrix_INCL__  

#include "FDATASetUp.h"
#include "LocMat.h"

#include "FDATACellMatCopy.h"
#include "FDATAMalloc.h"

/********************* exported symbols  ***********************************/

#ifndef CELLBLAS

#if FDATATYPE==SINGLE
#define MatZero		DSMatZero   
#define MatScale	DSMatScale  
#define PackMMOpds	DSPackMMOpds
#define MatPartMult	DSMatPartMult   
#define MatMult		DSMatMult   
#else
#define MatZero		DDMatZero   
#define MatScale	DDMatScale  
#define PackMMOpds	DDPackMMOpds
#define MatPartMult	DDMatPartMult   
#define MatMult		DDMatMult   
#endif

#else	/* rename routines to aviod name clash */

#if FDATATYPE==SINGLE
#define MatZero		SMatZero   
#define MatScale	SMatScale  
#define PackMMOpds	SPackMMOpds
#define MatPartMult	SMatPartMult   
#define MatMult		SMatMult   
#else
#define MatZero		DMatZero   
#define MatScale	DMatScale  
#define PackMMOpds	DPackMMOpds
#define MatPartMult	DMatPartMult   
#define MatMult		DMatMult   
#endif

#endif /*CELLBLAS*/

#ifdef EXT_FTN_BLAS             /* define Fortran BLAS procedure names called */
#include "FtnBLAS.h"
#endif /*EXT_FTN_BLAS*/

/************************ optimization-related macros ***********************/
 
#define IncCacheClash(N, Inc) 						\
        (((Inc) > 1) && (((N)-1)*(Inc) >= CACHESIZE))
        /*= cache confict between elements of vector of length N stride Inc;
            more precisely, but less efficiently: 
            ((N)-1)*GCD(Inc, CACHESIZE) >= CACHESIZE*CACHE_ASSOC */
                 
#define M_INNER(M, N, RInc, CInc) 					\
	((M)>=MIN_VECSIZE && ABS(RInc) <= ABS(CInc))
        /* choose vector (M, RInc) over (N, CInc) as the inner vector 	*/
        /* note: IncOK(M, N, RInc, CInc) can be assumed; in this case,
           one of |RInc|,|CInc| must be much greater than the other
           (unless both are unity)  					*/
	
#define CopyEfficient( N ) 						\
	( (N) > 2*MINOR_OUTER_VECSIZE)
	/* = whether efficent to copy MxK A being re-used by a factor of N */ 
	
#define PackUseful(M, K, RInc, CInc, N)					\
	(CopyEfficient(N) && /*(K)*(M)<=CACHESIZE &&*/ 			\
	 ((RInc) >= 2*(K) || (CInc) >= 2*(M)) &&			\
	 (IncCacheClash(M, RInc) || IncCacheClash(K, CInc)))
	/* = whether useful to pack M x K matrix with strides 		*/
	/*   (RInc, CInc) and a reuse factor of N 			*/

#define PartitionUseful(M, N, K)					\
	((K>1) && (2*((M)+(N))*(K) >= 3*CACHESIZE))	
	/* = whether (M+N)*K exceeds CACHESIZE sufficiently to justify 	*/ 
	/*   partitioning of computation to be applied 			*/
		
#define LM_PackUseful(M, K, A, N) PackUseful(M, K, LM_RInc(A), LM_CInc(A), N)

#define LM_Contiguous(RC, K, N, A) 					\
	(LM_Inc(1-(RC), A)==1 && ((K)==1 || (N)==LM_Inc(RC, A)))
	/* = whether (RC? K x N: N x K) A is contiguous.		*/
	
/********************** matrix routines *********************************/
    
void MatZero STDC_PARAMS( (int M, int N, pLocMat A) );
/* pre:  pLM_IncOK(M, N, A)
 * post: A[i,j] = 0 for 0<=i<M, 0<=j<N
 */

void MatScale STDC_PARAMS( (int M, int N, FDATA *alpha, pLocMat A ) );
/* pre:  pLM_IncOK(M, N, A)
 *	 A[i,j] = a(i,j)  for 0<=i<M, 0<=j<N and some function a
 * post: A[i,j] = *alpha * a(i,j) for 0<=i<M, 0<=j<N
 * nb. optimized for *alpha == 0.0, 1.0
 */
            
FDATA *PackMMOpds STDC_PARAMS( 
	(int M, int N, int K, FDATA **alphaf, pLocMat *A, pLocMat *B)
);
/* pre:  *A (*B) represents an M x K (K x N) matrix
 * post: if it is determined efficient (in context of matrix multiply *A * *B)
 *       & possible to pack *A and/or *B (via PackUseful()),
 *       and *A_pre and/or B*_pre are packed into workspaces  malloced from W,
 *       and A and/or B point to the copies.
 *       In this case, W is returned (for future disposal) and one of
 *       *A or *B will be scaled by **alphaf and alphaf will be set to NULL.
 *       otherwise returns NULL. 
 */

void MatPartMult STDC_PARAMS( 
	(int M, int N, int K, FDATA *alpha, pLocMat A, pLocMat B, FDATA *beta, 
	 pLocMat C)
);
/* pre:  pLM_IncOKUnit(M, K, A) && pLM_IncOKUnit(K, N, B) &&
 * 	 pLM_IncOKUnit(M, N, C) && 0 <= M, N, K, 
 *	 !Overlap(M, K, A, M, N, C) && !Overlap(K, N, B, M, N, C)
 * post: C := (*beta)C + (*alpha)A*B;  A is MxK, B is KxN, C is MxN; 
 * 	 applying partitioning methods & packing to reduce cache conflicts
 *	 assumes the underlying MM procedure (MatMult) can not (efficiently)
 *	 perform the scaled computation (*alpha != 1), so scaling is performed
 *	 whenever workspace is allocated.
 */  	

void MatMult STDC_PARAMS(
	(int M, int N, int K, FDATA *alpha, pLocMat A, pLocMat B, FDATA *beta, 
	 pLocMat C)
);
/* pre/post: as for MatPartMult(), except 
 *	(currently) for !defined(VANILLA or EXT_FTN_BLAS) *alpha = 1.0
 * note: no partitioning methods are applied 	      
 */

#endif /*__FDATACellMatrix_INCL__*/
