/* @(#) All Right Reserved, Copyright (c) 1997 Australian National University */
/* @(#)                     Copyright (c) 1997 Fujitsu Laboratories Ltd. */
/*  */
/*        */
/*                        Permission Notice */
/*  */
/*                Australian National University */
/*                Canberra ACT 0200 Australia */
/*                           and */
/*                Fujitsu Laboratories Ltd. */
/*                4-1-1 Kamikodanaka Nakahara */
/*                Kawasaki 211-88 Japan */
/*  */
/*   Permission to use, copy, modify and distribute this software (including  */
/*   its documentation) without fee is hereby granted under the following terms */
/*   and conditions: */
/*  */
/*   1. This permission notice and all of the copyright notices in this */
/*      software must appear in all copies and derived versions of all */
/*      or a part of this software. */
/*  */
/*   2. Sales of this software, whether incorporated in other products as a */
/*      part thereof or not, is prohibited. */
/*  */
/*   3. On publication or presentation in any form of the results obtained */
/*      by use of this software, an explicit reference to this software and */
/*      its copyright owner shall be made. */
/*  */
/*   4. THIS SOFTWARE IS PROVIDED 'AS IS' AND THE AUSTRALIAN NATIONAL UNIVERSITY */
/*      AND FUJITSU LABORATORIES LTD. MAKE NO WARRANTIES, EITHER EXPRESSED OR */
/*      IMPLIED, AS TO ANY MATTER WHATSOEVER REGARDING THIS SOFTWARE,  */
/*      INCLUDING, WITHOUT LIMITATION, WARRANTIES OF MERCHANTABILITY OR FITNESS */
/*      FOR ANY PARTICULAR PURPOSE. */
/*      IN NO EVENT SHALL THE AUSTRALIAN NATIONAL UNIVERSITY OR FUJITSU */
/*      LABORATORIES LTD. BE LIABLE FOR ANY CLAIM FROM A THIRD PARTY, OR */
/*      FOR SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES WHATSOEVER RESULTING */
/*      FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, */
/*      NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION */
/*      WITH THE USE, COPYING OR MODIFICATION OF THIS SOFTWARE. */
/*  */

/************************************************************************/
/*              FDATAMalloc: FDATA workspace meory allocation           */
/************************************************************************/
/* SCCS INFO @(#)FDATAMalloc.c 1.2         last modified P. Strazdins, 97/03/10
   first created: P. Strazdins 97/02/23 
*/
/* compilation: macro FDATATYPE must be defined;
 */
  

#include <stdlib.h>			/* memalign(), free() */
#include "FDATAMalloc.h"
#include "FDATAAssert.h"
#include "MachineParams.h"

/* Motivation: 
 * W is an Abstract Data Object to enable allocation of
 * small matrices or `blocks' (eg. Size<=FastWSSize, or possibly Size<SizeW)
 * with minimal software overheads. Larger blocks will be allocated
 * via the standard malloc library (and with greater overheads), and will be
 * outside W.
 *
 * It assumes a `stack-like' pattern of MatMalloc()/MatFree() calls
 * for this efficiency (other patterns will result in greater overheads
 * and/or poor utilization of W). However, if the bottom block is freed,
 * a pointer is kept to the next unfreed block.  
 * Such patterns occurs within level 1 and 2 DBLAS operations.
 *
 * Implementation:
 * A block of size S is allocated in W from words Wa..Wa+S+1;
 * the last word (the `size field') contains the value of S. 
 * This is sufficient to track backwards all blocks in W from Wa. 
 * Thus, only Wa and the word containing the size in W need be changed
 * upon a memory allocation or free.
 *	nb. this scheme assumes sizeof(int) <= sizeof(FDATA)
 *
 * Defensive Programming:
 * 2 types of error can occur when using dynamic memory allocation
 * in an application:
 *	1. blocks are over-written, resulting in a corrupted W
 *	   (ie. the size fileds becomes corrupted)
 *	2. a bad pointer is supplied with MatFree().
 * Using standard malloc library, both kinds of error can cause highly
 * unpredicatable results, and can be extremely difficult to find.
 * In this module, both types of error can be detected within MatFree()
 * provided the block was allocated from within W.
 */
 	
static struct {
  FDATA * W0,			/* start of W */
  	* W0_u,			/* bottom of unfrred space in W */
  	* W1,			/* end of W */
	* Wa;			/* points to next available space in W */
  int 	Size,			/* size of W */
  	FastWSSize;		/* max size of blocks in W */
} W = { 
  NULL, NULL, NULL, NULL, 4*FastWSSize_DEF, FastWSSize_DEF 
};		
/* invariants: 	W0 <= W0_u <= Wa <= W1; 0 <= FastWSSize <= Size;
 *		if W0!=NULL, FastWSSize=W1-W0
 */
 


void MatMallocInit (ParamType, ParamVal) 
  int ParamType, ParamVal;
{ if (ParamType==0) {
    if (W.W0==W.Wa) {		/* W is not currently in use */
      MatFreeBuf();
      if (ParamVal >= 0)
        W.Size = ParamVal;
    }
  } else {
    if (ParamVal>=0 && ParamVal<=W.Size)
      W.FastWSSize = ParamVal;
  }
} /* MatMallocInit() */

void MatFreeBuf()
{ if (W.W0!=NULL) {		/* free W for latter allocation */ 
    free ((void *) W.W0);
    W.W0 = W.W0_u = W.W1 = W.Wa = NULL;
  }	
} /* MatFreeBuf() */


int MatMalloc_f(Where, CrucialAlloc, A, Size, UseFastMalloc, ImmedFree) 
  char *Where; FDATA **A; int CrucialAlloc, Size, UseFastMalloc, ImmedFree;
{ FDATA *W1 = W.W1;			FDATA *Wa;
  int FastWSSize;
  if (Size <= 0) {
    *A = NULL;
    return 0;
  }    	
  
  if (W1==NULL && UseFastMalloc==FD_ALLOC_TRUE) {
    if (MatMalloc("MatMalloc_f", 0, &W.W0, W.Size)) {
      W.Wa = W.W0_u = W.W0;		W1 = W.W1 = W.W0 + W.Size;
    }  
  }  

  Wa = W.Wa;				FastWSSize = W.FastWSSize;
  if ((Size+1 <= ((W1-Wa)&UseFastMalloc)) && 
      ((Size&ImmedFree) <= FastWSSize)) {	/* use W for fast malloc */
     *A = Wa;
      Wa += Size+1;
      *((int *) (Wa-1)) = Size+1; 
      W.Wa = Wa;
  } else 
    *A = (FDATA *) memalign(MAX(sizeof(FDATA), CACHE_LINE_SIZE), 
    			    Size*sizeof(FDATA));

  COMMON_ASSERT_TR(
    (TraceStr, "(%s,%d,a%d,%d,%d,%d)", Where, CrucialAlloc,
                 FD_ADR(*A), Size, UseFastMalloc, ImmedFree),
    (UseFastMalloc==FD_ALLOC_FALSE ||  UseFastMalloc==FD_ALLOC_TRUE) &&
    (ImmedFree==FD_ALLOC_FALSE ||  ImmedFree==FD_ALLOC_TRUE), 0,
    "MatMalloc_f", 4, TRACE_MALLOC
  );
    
  if (*A==NULL && CrucialAlloc) {
    CommonAssert(1, Size, "MatMalloc_f", 4, TRACE_MALLOC);
  } 	

  return (*A != NULL);  
} /* MatMalloc() */

void MatFree(A) 
  FDATA *A;
{ 
  COMMON_ASSERT_TR(
    (TraceStr, "(a%d)", FD_ADR(A)), 1, 0, "MatFree", 4, TRACE_MALLOC
  );
  if (A != NULL) { 
    FDATA *W0_u = W.W0_u; 		 FDATA *Wa_ = W.Wa;
    int HeapSize = Wa_ - W0_u;
    if (W.W0_u<=A && A<W.W1) {		/* A is within (unfreed blocks of) W */
      int Size = *((int *) (Wa_-1));
      Wa_ = Wa_ - Size;
      if (A==Wa_) {			/* A is top block in W */
        W.Wa = Wa_;  
        if (Wa_ == W0_u)		/* A is also only unfreed block in W */ 
          W.W0_u = W.W0;
#if 0
printf("Wa_=a%d W0_u=a%d\n", FD_ADR(W.Wa), FD_ADR(W.W0_u));         
#endif
      } else {  	
        FDATA *Wa_l; 
        do {				/* search back from Wa_ for A */
#if 0        
printf("HeapSize=%d, Size=%d, Wa_=a%d W0_u=a%d\n", HeapSize, Size, 
	FD_ADR(Wa_), FD_ADR(W0_u));    
#endif 
	  if (Size<=1 || Size>HeapSize)
            CommonAssert(0, 0, "MatFree: corrupted heap", 4, TRACE_MALLOC);
          if (A > Wa_)
            CommonAssert(0, 0, "MatFree: bad pointer", 4, TRACE_MALLOC);
          if (A == Wa_) break;   
          Wa_l = Wa_;
          HeapSize -= Size;
          Size = *((int *) (Wa_-1));   
          Wa_ -= Size;
        } while (1);
        if (A == W0_u)			/* A is bottom unfreed block in W */
          W.W0_u += Size;
        else {  			/*merge block at A with one below*/
          Size += *((int *) (Wa_-1));
          *((int *) (Wa_l-1)) = Size; 
        }
#if 0
printf("W0_u=a%d *(a%d-1)=%d\n", FD_ADR(W.W0_u), FD_ADR(Wa_l), Size);
#endif
      }
    } else if (W.W0<=A && A<W.W0_u)
      CommonAssert(0, 0, "MatFree: bad pointer", 4, TRACE_MALLOC);      
    else				/* A is outside W */
      free ((char *) A);
  }            
} /* MatFree() */

