/* @(#) All Right Reserved, Copyright (c) 1997 Australian National University */
/* @(#)                     Copyright (c) 1997 Fujitsu Laboratories Ltd. */
/*  */
/*        */
/*                        Permission Notice */
/*  */
/*                Australian National University */
/*                Canberra ACT 0200 Australia */
/*                           and */
/*                Fujitsu Laboratories Ltd. */
/*                4-1-1 Kamikodanaka Nakahara */
/*                Kawasaki 211-88 Japan */
/*  */
/*   Permission to use, copy, modify and distribute this software (including  */
/*   its documentation) without fee is hereby granted under the following terms */
/*   and conditions: */
/*  */
/*   1. This permission notice and all of the copyright notices in this */
/*      software must appear in all copies and derived versions of all */
/*      or a part of this software. */
/*  */
/*   2. Sales of this software, whether incorporated in other products as a */
/*      part thereof or not, is prohibited. */
/*  */
/*   3. On publication or presentation in any form of the results obtained */
/*      by use of this software, an explicit reference to this software and */
/*      its copyright owner shall be made. */
/*  */
/*   4. THIS SOFTWARE IS PROVIDED 'AS IS' AND THE AUSTRALIAN NATIONAL UNIVERSITY */
/*      AND FUJITSU LABORATORIES LTD. MAKE NO WARRANTIES, EITHER EXPRESSED OR */
/*      IMPLIED, AS TO ANY MATTER WHATSOEVER REGARDING THIS SOFTWARE,  */
/*      INCLUDING, WITHOUT LIMITATION, WARRANTIES OF MERCHANTABILITY OR FITNESS */
/*      FOR ANY PARTICULAR PURPOSE. */
/*      IN NO EVENT SHALL THE AUSTRALIAN NATIONAL UNIVERSITY OR FUJITSU */
/*      LABORATORIES LTD. BE LIABLE FOR ANY CLAIM FROM A THIRD PARTY, OR */
/*      FOR SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES WHATSOEVER RESULTING */
/*      FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, */
/*      NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION */
/*      WITH THE USE, COPYING OR MODIFICATION OF THIS SOFTWARE. */
/*  */

/************************************************************************/
/*	LocMat: definitions for cell matrix data structures		*/
/************************************************************************/
/* SCCS INFO @(#)LocMat.h 4.3         last modified P. Strazdins, 97/04/22
   first created: P. Strazdins 96/01/15
*/
/* this module facilitiates easy local matrix manipulations
 * by generalized element row/column increments. 
 *  
 * Generally, Inc[0] (Inc[1]) will be associated with a row (col) incremement.
 * Thus the address of A[i,j] = LM_ADR(A) + i*LM_RInc(A) + j*LM_CInc(A).  
 *
 * Typically, MIN(Inc[0..1]) = 1; this is required for matrix multiply at least.
 *
 * notes: a problem with the defnition of LocMat is that LM_ADR() field
 *	  is less `stable' than the others, which means it is likely
 *	  to vary over a computation. For this reason  LocMat SubLocMat()
 *	  must be a proper function.
 *  
 *  compilation: macro FDATATYPE must be defined;
 */

#ifndef __LocMat_INCL__

#define __LocMat_INCL__

#include "FDATASetUp.h"

#if FDATATYPE==SINGLE
#define LocSubMat	SLocSubMat
#define MkLocMat	SMkLocMat
#define LocMatTrans	SLocMatTrans
#define MatPrint        SMatPrint
#else
#define LocSubMat	DLocSubMat
#define MkLocMat	DMkLocMat
#define LocMatTrans	DLocMatTrans
#define MatPrint        DMatPrint
#endif

typedef short int IPair[2];	/* 2-dim integer params, ie. (I,J), (RInc,CInc)*/
	
typedef struct LocMatStruct {
  FDATA 	*x;		/* points to local element [0,0] in cell */
  IPair		Inc;		/* row/column strides for elements */
} LocMat;

typedef LocMat *pLocMat;

#define IncOK(M, N, RInc, CInc)  /*ensures no 2 elts of matrix map to 1 addr*/ \
        (ABS(RInc) > ((N)-1)*ABS(CInc) || ABS(CInc) > ((M)-1)*ABS(RInc))
         
#define IncOKUnit(M, N, RInc, CInc)  \
        ((MIN(RInc, CInc) == 1) && IncOK(M, N, RInc, CInc))
         
#define LM_IncOK(M, N, A)       IncOK(M, N, LM_RInc(A), LM_CInc(A))
 
#define LM_IncOKUnit(M, N, A)   IncOKUnit(M, N, LM_RInc(A), LM_CInc(A))


#define LM_ADR(M) 	((M).x)
#define LM_RInc(M)	((M).Inc[0])
#define LM_CInc(M)	((M).Inc[1])
#define LM_Inc(t, M)	((M).Inc[t])
#define LM_Ld(M)	(MAX(LM_RInc(M), LM_CInc(M)))
#define LM_Eq(L1, L2)	((L1).x==(L2).x && 		\
			 (L1).Inc[0]==(L2).Inc[0] && (L1).Inc[1]==(L2).Inc[1])
	 
#define LM_Index(M, i, j) 	\
	(LM_ADR(M)+FAST_MUL2(i, LM_RInc(M))+ FAST_MUL2(j, LM_CInc(M)))

#define LM_Init( M, x_, Inc0, Inc1 ) \
	(M).x = x_, (M).Inc[0] = Inc0, (M).Inc[1] = Inc1 

#define LM_Copy( Tr, MT, M ) \
	{ int t=(M).Inc[1-(Tr)]; (MT).x = (M).x; \
	  (MT).Inc[0] = (M).Inc[Tr]; (MT).Inc[1] = t;}

#define LM_Trans( Tr, M ) \
	if (Tr) SWAP(int, (M).Inc[0], (M).Inc[1]);

#define LM_SubMat(A, i, j)	LM_ADR(A) = LM_Index(A, i, j)

LocMat MkLocMat STDC_PARAMS( (FDATA *A, int RInc, int CInc) );
/* post: returns {A, RInc, CInc};
 */

LocMat LocSubMat STDC_PARAMS( (LocMat M, int i, int j) );
/* post: returns {LM_Index(M, i, j), LM_RInc(M), LM_CInc(M)};
 */

LocMat LocMatTrans STDC_PARAMS( (LocMat M) );
/* post: returns {LM_ADR(M), LM_CInc(M), LM_RInc(M)};
 */

#define LocMatTr(Tr, M)		(!Tr? M: LocMatTrans(M))

#define LocSubMatTr( Tr, M, i, j ) LocSubMat(M, (Tr)? (j):(i), (Tr)? (i):(j))  

void MatPrint STDC_PARAMS(
        (char *Name, int M, int N, LocMat A)
 );
/* pre:   LM_IncOK(M, N, A)
 * post:  prints A[0..M-1][0..N-1] in row-major form to stdout
 */

#define MatOverlap( MA, NA, A, MB, NB, B) \
        0
/*        ((MA)>0 && (NA)>0 && (MB)>0 && (NB)>0 &&                      \
        (((LM_ADR(A) < LM_ADR(B)) &&                                    \
          (LM_ADR(A)+((MA)-1)*LM_RInc(A)+((NA)-1)*LM_CInc(A) >= LM_ADR(B)))||\
         ((LM_ADR(B) < LM_ADR(A)) &&                                    \
          (LM_ADR(B)+((MB)-1)*LM_RInc(B)+((NB)-1)*LM_CInc(B) >= LM_ADR(A)))))
*/
/* pre:  MIN(RIncA, CIncA) = MIN(RIncB, CIncB) = 1
 * post: returns TRUE iff A[0..MA-1,0..NA-1] & B[0..MB-1,0..NB-1] overlap
 */

/******************** macros with pLocMat values *******************/ 

#define pLM_ADR(pM) 	LM_ADR(*(pM))
#define pLM_RInc(pM)	LM_RInc(*(pM))
#define pLM_CInc(pM)	LM_CInc(*(pM))
#define pLM_Inc(t, pM)	LM_Inc(t, *(pM))
#define pLM_Ld(pM)	(MAX(pLM_RInc(pM), pLM_CInc(pM)))

#define pLM_IncOK(M, N, pA)  LM_IncOK(M, N, *(pA))
 
#define pLM_IncOKUnit(M, N, pA)   LM_IncOKUnit(M, N, *(pA))

#define pLM_Index(pM, i, j)	LM_Index(*(pM), i, j) 	

#define	pLM_Trans(A)	SWAP(int, pLM_RInc(A), pLM_CInc(A))


#define pLocSubMat(pA, di, dj)		\
	(pLM_ADR(pA) = pLM_Index(pA, di, dj), (pA))

#define pLocSubMat_x(pA, A, i, j) 	\
	(pLM_ADR(pA) = pLM_Index(A, i, j), (pA))

#define pLocSubMat_(pA, A, i, j) 	\
	pMkLocMat(pA, pLM_Index(A, i, j), pLM_RInc(A), pLM_CInc(A))

#define pMkLocMat( pA, Ax, RInc, CInc )	\
	((pA)->x=(Ax), (pA)->Inc[0] = (RInc), (pA)->Inc[1] = (CInc), (pA))

#define pLocSubMatTr(Tr, pA, di, dj)          \
	pLocSubMat(pA, Tr? (dj): (di), r? (di): (dj))
        
#define pLocSubMat_Tr(Tr, pA, A, i, j) 	\
	pLocSubMat_(pA, A, Tr? (dj): (di), r? (di): (dj))
	
#endif /*__LocMat_INCL__*/
