/* @(#) All Right Reserved, Copyright (c) 1997 Australian National University */
/* @(#)                     Copyright (c) 1997 Fujitsu Laboratories Ltd. */
/*  */
/*        */
/*                        Permission Notice */
/*  */
/*                Australian National University */
/*                Canberra ACT 0200 Australia */
/*                           and */
/*                Fujitsu Laboratories Ltd. */
/*                4-1-1 Kamikodanaka Nakahara */
/*                Kawasaki 211-88 Japan */
/*  */
/*   Permission to use, copy, modify and distribute this software (including  */
/*   its documentation) without fee is hereby granted under the following terms */
/*   and conditions: */
/*  */
/*   1. This permission notice and all of the copyright notices in this */
/*      software must appear in all copies and derived versions of all */
/*      or a part of this software. */
/*  */
/*   2. Sales of this software, whether incorporated in other products as a */
/*      part thereof or not, is prohibited. */
/*  */
/*   3. On publication or presentation in any form of the results obtained */
/*      by use of this software, an explicit reference to this software and */
/*      its copyright owner shall be made. */
/*  */
/*   4. THIS SOFTWARE IS PROVIDED 'AS IS' AND THE AUSTRALIAN NATIONAL UNIVERSITY */
/*      AND FUJITSU LABORATORIES LTD. MAKE NO WARRANTIES, EITHER EXPRESSED OR */
/*      IMPLIED, AS TO ANY MATTER WHATSOEVER REGARDING THIS SOFTWARE,  */
/*      INCLUDING, WITHOUT LIMITATION, WARRANTIES OF MERCHANTABILITY OR FITNESS */
/*      FOR ANY PARTICULAR PURPOSE. */
/*      IN NO EVENT SHALL THE AUSTRALIAN NATIONAL UNIVERSITY OR FUJITSU */
/*      LABORATORIES LTD. BE LIABLE FOR ANY CLAIM FROM A THIRD PARTY, OR */
/*      FOR SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES WHATSOEVER RESULTING */
/*      FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, */
/*      NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION */
/*      WITH THE USE, COPYING OR MODIFICATION OF THIS SOFTWARE. */
/*  */

/************************************************************************/
/*	MatDim: distributed block-cyclic matrices definition		*/
/************************************************************************/
/* SCCS INFO @(#)MatDim.h 4.6         last modified P. Strazdins, 97/04/18
   first created: P. Strazdins 94/05/17
*/
/* contains:
 *	* distributed matrix dimension and structure definitions
 *	* macros/procedures for computing cell dimensions,
 *	  converting local <-> global indices etc.
 *  
 *  compilation: macro FDATATYPE must be defined;
 *		 if CELLBLAS defined, compile for single 1x1 cell (workstation)
 *		 (some optimiations can be performed)
 */

#ifndef __MatDim_INCL__
#define __MatDim_INCL__

#include "GenMacros.h"

/********************* structure definitions ******************************/
#define D_INVAL_LG	255	/* signifies invalied lgB, lpP fields */
#define VALID_LG(x)	((x)!=D_INVAL_LG)

struct MatDimStruct {		/* distribution parameters in a particular dim.*/
  unsigned char 	
	 	B0,		/* block offset */
  		B,		/* block size */
  		lgB, lgP;	/* if != D_INVAL_LG, use for bit opn arithmetic */
  short unsigned int 	
  		P0,		/* P0 is (P + I - I0)%P where:
				   I0 is id of cell containing 1st block
				   I is id of current cell  
				*/
  		P; 		/* number of processors in dimension */ 	
	/* invariants: 	0 <= B0 < B; 0 <= P0 < P; 
		       	B=2^lgB || lgB=D_INVAL_LG; P=2^lgP || P=D_INVAL_LG*/
};

union MatDimUnion {		/* for efficiency, esp in copying */
  	struct MatDimStruct d;
  	double align;	
};
	
typedef union MatDimUnion MatDim;
typedef MatDim *pMatDim;

/************** MatDim constructors **************************************/

#define MatDimInit(DI, b0, b, p0, p) \
	pInitDim(&(DI), b0, b, p0, p)
	        	
#define MatDimCp(D1, D2)                                                \
        if (sizeof((D1)->align) >= sizeof((D1)->d))			\
          (D1)->align = (D2)->align;                                    \
        else                                                            \
          (D1)->d = (D2)->d;
                                    
#define MatDimUnit(D1)		MatDimCp(D1, &UnitDim)

/* SetUpDim() used where I0 is combined block-cell offset */
#define SetUpDim(DI, I, I0, Bl, Pr) 				\
	{ int P0 = (I)-(I0)/(Bl); while (P0 < 0) P0 += (Pr);	\
	  pInitDim(&(DI), (I0)%(Bl), Bl, P0, Pr);	}
	  
/************** pMatDim selectors **************************************/

#ifdef CELLBLAS
# define d_B0(D)	0
# define d_B(D)		1
# define d_P0(D)	0
# define d_P(D)		1

# define d_lgB(D)  	0
# define d_lgP(D)  	0

# else
# define d_B0(D)	((D)->d.B0)
# define d_B(D)		((D)->d.B)
# define d_P0(D)	((D)->d.P0)
# define d_P(D)		((D)->d.P)

# define d_lgB(D)  	((D)->d.lgB)
# define d_lgP(D)  	((D)->d.lgP)

#endif /* CELLBLAS */
#define D_VAL_lgB(D)  	VALID_LG(d_lgB(D))
#define D_VAL_lgP(D)  	VALID_LG(d_lgP(D))


/************** pMatDim predicates **************************************/

#define Dim_Eq(D1, D2) 		((d_B0(D1)==d_B0(D2)) & (d_B(D1)==d_B(D2)) & \
				 (d_P0(D1)==d_P0(D2)) & (d_P(D1)==d_P(D2)))

#define Dim_OK(D)		(/*(0<=d_B0(D)) &*/ (d_B0(D)<d_B(D)) && 	\
		         	 /*(0<=d_P0(D)) &*/ (d_P0(D)<d_P(D)) && 	\
				(1<<d_lgB(D)==d_B(D) || d_lgB(D)==D_INVAL_LG)&&\
				(1<<d_lgP(D)==d_P(D) || d_lgP(D)==D_INVAL_LG))

/* Dim_PCompat(D1, D2) determines if the processor field in D1 is compatable
   to that in D2, ie they are either equal or d_P(D1)==1 
   (ie. there has been a `spread' across D1) 			*/
#define Dim_PCompat(D1, D2)	(d_P(D1)==d_P(D2) || d_P(D1)==1)

#define Dim_PEq(D1, D2)		(d_P(D1)==d_P(D2))

#define Dim_Aligned(D1, D2) 	(Dim_BCompat(D1, D2) && Dim_PEq(D1, D2))

/* Dim_BCompat(D1, D2) = block sizes and offsets are compatible across D1 & D2*/
#define Dim_BCompat(D1, D2) 	((d_B0(D1)==d_B0(D2)) & (d_B(D1)==d_B(D2)))
#define CompatBlk 		Dim_BCompat

/* determine whether current cell owns element 0 of dimension */
#define Dim_Owns(D)		(d_P0(D)==0)

/************* pMatDim front-ends to indexing procedures ******************/

#define G2L(i, B0, B, P0, P)	G2Llg(i, B0, B, P0, P, D_INVAL_LG, D_INVAL_LG)

#define L2G(i, B0, B, P0, P)	L2Glg(i, B0, B, P0, P, D_INVAL_LG, D_INVAL_LG)

				/* the motivation of the following is to use
				   the fast arithmetic enabled 
				   from d_lgB(D), d_lgP(D)) whenever possible */
				   

#define G2Ld(i, D)		G2Llg(i, d_B0(D), d_B(D), d_P0(D), d_P(D), \
				      d_lgB(D), d_lgP(D))

#define G2Ld_P0(i, D, P0)	G2Llg(i, d_B0(D), d_B(D), P0, d_P(D),	\
				      d_lgB(D), d_lgP(D))

#define G2Ld_(i, D, B0, P0)	G2Llg(i, B0, d_B(D), P0, d_P(D),	\
				      d_lgB(D), d_lgP(D))

#define G2Ld2(i, DB, DP)	G2Ld2_(i, DB, DP, d_B0(DB), d_P0(DP))
				      
#define G2Ld2_(i, DB, DP, B0, P0)					\
				G2Llg(i, B0, d_B(DB), P0, d_P(DP), 	\
				      d_lgB(DB), d_lgP(DP))

#define G2LdB(i, DB, P, B0, P0) G2Llg(i, B0, d_B(DB), P0, P, 		\
					d_lgB(DB), D_INVAL_LG)
					
#define G2Ld_Mx(i, D)		G2Llg(i, 0, d_B(D), 0, d_P(D),		\
				       d_lgB(D), d_lgP(D))

#define L2Gd(i, D) 		L2Glg(i, d_B0(D), d_B(D), d_P0(D), d_P(D), \
                                      d_lgB(D), d_lgP(D))

#define L2Gd_(i, D, B0, P0) 	L2Glg(i, B0, d_B(D), P0, d_P(D), 	\
                                      d_lgB(D), d_lgP(D))

/************* minor indexing macros ************************************/

				/* number of partial blocks for global length N
				   across dimension D */ 
#define NBLKSd(N, D) 		(D_VAL_lgB(D)?                          \
                                 (((N) + d_B0(D) + d_B(D)-1) >> d_lgB(D)):  \
                                 NBLKS(N, d_B0(D),  d_B(D)))
#define NBLKS(N, B0, B) 	(((N) + (B0) + (B) - 1)/ (B))

				/* whether  NBLKSd(N, D) <= 1 */
#define Blk1d(N, D)		(d_P(D)==1 || Blk1(N, d_B0(D),  d_B(D)))
#define Blk1d_(N, D)		Blk1(N, d_B0(D),  d_B(D))
#define Blk1(N, B0, B)		((N) + (B0) <= (B))

				/* return block-offset into (this) cell's 
				   portion of matrix (only >0 if leading cell)*/
#define BlkB0d(i, D)		BlkB0(i, d_P0(D)==0? d_B0(D): 0)
#define BlkB0(i, B0)		((i)==0? (B0): 0)

				/* return value of local index corresp. to start
				   of (local) block i in cell portion of matrix.
				   BlkIxP0d() & BlkIxd_() are typically used for
				   `spread' workspaces; here the block offset
				   is always taken into account because on this
				   cell ALL blocks are contained. BlkIxd() is
				   used for cell portion of a DistMat */
#define BlkIxP0d(D)		BlkIx_(d_P0(D), d_B0(D), d_B(D), d_lgB(D))
#define BlkIxd_(i, D)		BlkIx_(i, d_B0(D), d_B(D), d_lgB(D))
#define BlkIxd(i, D)		BlkIx_(i, d_P0(D)==0? d_B0(D): 0, d_B(D), d_lgB(D))
#define BlkIx_(i, B0, B, lgB)	((i)==0? 0: ((VALID_LG(lgB)? ((i)<<(lgB)): (i)*(B)) - (B0)))
#define BlkIx(i, B0, B) 	((i)==0? 0: (i)*(B) - (B0))

				/* return relative co-ordinate P0 of cell CidI
				   for dimension beginning in cell CidI0 */							
#define Cid2P0(CidI, CidI0, P) ((CidI)<(CidI0)? (CidI)-(I0) + (P): (CidI)-(I0))


#ifndef CELLBLAS
/* SubMatP0(i, B0, B, P0, P) is cell offset of global elt 0 for i-th sub-vector
 * of block size B and block offset B0 and relative cell offset P0
 */
#define SubMatP0(i, B0, B, P0, P)	\
		((P)-1 - FAST_MODP2(((i)+(B0))/(B) + (P)-1 - (P0), P))

#define SubMatP0d(i, D)						\
	(d_P(D)-1 - FAST_MODP2(					\
				(D_VAL_lgB(D)? (((i)+d_B0(D))>>d_lgB(D)): \
			        		((i)+d_B0(D))/d_B(D)) 	  \
			         + d_P(D)-1 - d_P0(D), d_P(D)))

/* SubMatB0(i, B0, B) is block offset of global elt 0 for i-th sub-vector
 * of block size B and block offset B0
 */
#define SubMatB0(i, B0, B)	FAST_MODP2((i)+(B0), B)

#else

#define SubMatP0d(i, D)			0
#define	SubMatP0(i, B0, B, P0, P)	0
#define SubMatB0(i, B0, B)		0
#endif /*CELLBLAS*/

#define SubMatB0d(i, D)		SubMatB0(i, d_B0(D), d_B(D))

/************* major indexing procedures *******************************/


#ifdef CELLBLAS	/* bypass procedures with serial equivlents */		
/* this optimization may be dangerous if G2L() or L2G() required to do 
 * non-unit block-striding within a cell. Currently in the DBLAS, 
 * non-unit block-strides are not possible on a 1x1 grid.
 */
 
#define G2Llg(N, D, B0, P0)		(N)
#define L2Glg(N, D, B0, P0)		(N)
#define G2LSubDim(N, D)			(N)
#define G2LSubDimM1(D)			(-1)
#define SubDim(D, i)			(D)
#define pSubDim_(D, D1, i)		(D)
#define pInitDim(D, B0, B, P0, P)	(D)
/* UnitDim, CommonDimAlignCheck  should not be accessed */
		
#else

#define G2Llg		GG2Llg
#define L2Glg		GL2Glg
#define G2LSubDim	GG2LSubDim
#define G2LSubDimM1	GG2LSubDimM1
#define SubDim		GSubDim
#define pSubDim_	GpSubDim_
#define pInitDim	GpInitDim
#define UnitDim		GUnitDim
#define	CommonDimAlignCheck	GCommonDimAlignCheck

/* invariant: if P0 = RelPrId(i, B0, B, P), 
 *		L2G(G2L(i, B0, B, P0, P), B0, B, P0, P) = i
 */

int G2Llg STDC_PARAMS( 
	(unsigned int N, unsigned int B0, unsigned int B, unsigned int P0, 
	 unsigned int P, unsigned int lgB, unsigned int lgP) 
);
/* pre:  B0<B; P0<P; lgB=D_INVAL_LG || 2**lgB=B; lgP=D_INVAL_LG || 2**lgP=P
 *	 a global vector A of block size B is distributed across P cells; 
 *	 starts at relative cell -P0; has block offset B0 for 1st block 
 * post: if N is size of global vector, returns the number of elements
 *	 in processor P0 from start of vector.
 *       This is ewquivalent to local offset of subvector starting from A[N].
 */	

int G2LSubDim STDC_PARAMS( (unsigned int N, pMatDim D) );
/* 
 *(post: returns G2Ld(N, D_pre); D = pSubDim(D, D_pre, N)
 * notes:this saves repeated integer operations comon to both G2Ld() & pSubDim()
 */	
	
int G2LSubDimM1 STDC_PARAMS( (pMatDim D) );
/* post: as for G2LSubDim(-1, D) 
 * notes:returns 0 or -1
 */	
	
int L2Glg STDC_PARAMS( 
	(unsigned int N, unsigned int B0, unsigned int B, unsigned int P0,
	 unsigned int P, unsigned int lgB, unsigned int lgP) 
);
/* pre:  B0 < B; P0 < P; lgB=D_INVAL_LG || 2**lgB=B; lgP=D_INVAL_LG || 2**lgP=P
 *	 a global vector A of block size B is distributed across P cells; 
 *	 starts at relative cell -P0; has block offset B0 for 1st block 
 * post: returns the global index corresponding to local index i on cell P0
 */
	
extern MatDim UnitDim;		/* {0, 1, 0, 1} */

pMatDim pInitDim STDC_PARAMS( (pMatDim D, int B0, int B, int P0, int P ) );
/* pre:  0<=B0<B && 0<=P0<P
 * post: returns D, *D = {B0, B, P0, P}
 */

MatDim SubDim STDC_PARAMS( (MatDim D, int i) );
/* post: returns distributed Matrix dimension from offseting D with i
 */	
#endif /*CELLBLAS*/

#define pSubDim(D, i)	pSubDim_(D, D, i)

pMatDim pSubDim_ STDC_PARAMS( (pMatDim RV, pMatDim D, int i) );
/* post: returns RV; *RV = SubDim(*D, i)
 */ 


int CommonDimAlignCheck STDC_PARAMS(
	(int N, int MIncA, pMatDim DNA, int MIncB, pMatDim DNB, int *ParamNums)
);
/* pre: DNA & DNB (MIncA & MIncB) are the dimensions (local matrix strides)
 *      in 2 distrubted matrices A & B corresponding (opposite) to N.
 *      *ParamNums = 0x(p[0])(p[1])(p[2]), where p[0..2] denotes the
 *      parameters positions of N, A & B respectively in an outer-level
 *      DBLAS call.
 * post: the value of (*ParamNums) signfifies the following error conditions:
 *         p[0]:                N<0
 *         p[1] (p[2]):         *DNA (*DNB) are invalid MatDim structures, or
 *                              MIncA (MIncB) are invalid local matrix strides
 *         0x(p[0])(p[1]):      MIncA < G2Ld(N, DNA)
 *         0x(p[0])(p[2]):      MIncB < G2Ld(N, DNB)
 *         0:                   none of the above holds
 *       returns whether DNA and DNB are block-aligned with repsect to each
 *       other.
 *	 DNB may be NULL, in which case all checking associated with MIncB
 *	 is suppressed.
 * note: this routine performs most checking associated with a particular
 *       logical dimension in a distribteed computation.
 *       This should enable efficient error checking in both time and code
 *       size.
 *       The hex encoding enables convenient & efficient indentification of
 *       the numbers of erroneous parameters (provided <= 15 parameters used)
 */

 
#endif /*__MatDim_INCL__*/
