/* @(#) All Right Reserved, Copyright (c) 1997 Australian National University */
/* @(#)                     Copyright (c) 1997 Fujitsu Laboratories Ltd. */
/*  */
/*        */
/*                        Permission Notice */
/*  */
/*                Australian National University */
/*                Canberra ACT 0200 Australia */
/*                           and */
/*                Fujitsu Laboratories Ltd. */
/*                4-1-1 Kamikodanaka Nakahara */
/*                Kawasaki 211-88 Japan */
/*  */
/*   Permission to use, copy, modify and distribute this software (including  */
/*   its documentation) without fee is hereby granted under the following terms */
/*   and conditions: */
/*  */
/*   1. This permission notice and all of the copyright notices in this */
/*      software must appear in all copies and derived versions of all */
/*      or a part of this software. */
/*  */
/*   2. Sales of this software, whether incorporated in other products as a */
/*      part thereof or not, is prohibited. */
/*  */
/*   3. On publication or presentation in any form of the results obtained */
/*      by use of this software, an explicit reference to this software and */
/*      its copyright owner shall be made. */
/*  */
/*   4. THIS SOFTWARE IS PROVIDED 'AS IS' AND THE AUSTRALIAN NATIONAL UNIVERSITY */
/*      AND FUJITSU LABORATORIES LTD. MAKE NO WARRANTIES, EITHER EXPRESSED OR */
/*      IMPLIED, AS TO ANY MATTER WHATSOEVER REGARDING THIS SOFTWARE,  */
/*      INCLUDING, WITHOUT LIMITATION, WARRANTIES OF MERCHANTABILITY OR FITNESS */
/*      FOR ANY PARTICULAR PURPOSE. */
/*      IN NO EVENT SHALL THE AUSTRALIAN NATIONAL UNIVERSITY OR FUJITSU */
/*      LABORATORIES LTD. BE LIABLE FOR ANY CLAIM FROM A THIRD PARTY, OR */
/*      FOR SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES WHATSOEVER RESULTING */
/*      FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, */
/*      NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION */
/*      WITH THE USE, COPYING OR MODIFICATION OF THIS SOFTWARE. */
/*  */

/************************************************************************/
/*   SpreadMatrix: distributed matrix spread procedures used by DBLAS	*/
/************************************************************************/
/* SCCS INFO @(#)SpreadMatrix.c 4.8         last modified P. Strazdins, 97/04/24
   first created: P. Strazdins 93/15/01
 */
/* compilation: macro FDATATYPE must be defined;
 *              if macro ASSERT defined, assertion checking turned on
 *		if macro TRACE defined, parameter value tracing enabled 
 */
#include "SpreadMatrix.h"
#include "BLASOptions.h"
#include "FDATAOps.h"
#include "LocMat.h"
#include "FDATAAssert.h"

#include "MatDim.h" 
#include "DistMat.h"
#include "FDATACellMatrix.h"
#include "FDATACellBlkMatrix.h"
#include "FDATACommun.h"
#include "FDATADistMatrix.h"


/****************  macros for spread optimizations *********************/

/* SpreadTrivial(K, DK, M, DM)) returns whether a spread of length K
 *	across dimension DK (to be received be cells spanned by M with 
 *	dimenssion DN) requires communication. 
 *	This is for efficiency; thus it is not essential for correctness
 *	that this macro be conservative.
 *	note: DK & DM are orientated in the same direction.
 * eg:  !TransB:		SpreadTrivial(K, BDK, M, DM)
 *	TransB && !KPreSpread:	SpreadTrivialP0(K, BDK, d_P0(BDN), M, DM)
 *	TransB && KPreSpread:	TrKPreSpreadTrivial(N, BDN, M, DN, DN)
 */  
#define SpreadTrivial(K, DK, M, DM)					\
	SpreadTrivialP0(K, DK, d_P0(DK), d_P(DK), M, DM)
#define SpreadTrivialP0(K, DK, P0, P, M, DM)				\
	((P)==1 || SpreadDestAligned(K, DK, P0, M, DM))
#define SpreadDestAligned(K, DK, P0, M, DM)				\
	 (Blk1d_(M, DM) && Blk1d_(K, DK) && (P0)==d_P0(DM))

#define TrKPreSpreadTrivial(N, BDN, M, DM, DN)				\
	((d_P(BDN)==1 && d_P(DN)==1) || 				\
	 (d_P0(DN)==0 && SpreadDestAligned(N, BDN, d_P0(BDN), M, DM)))
/* note: the 2 most important cases here are:
 *	 1. d_P(DN)==1 (ie. d_P(BDK)==1)
 *	 2. B was created by a previous spread in the K-dimension
 *	    and there is a single block contained in this column
 *	    (eg. in a level 2 LLT). Here we can generally assume 
 *	    d_P0(BDN)==d_P0(DM) (note that block alignment restrictons
 *	    ensure the other attributes of BDN & DM must be the same).
 *	 2a.
 *	    if d_P(DN)>1 && d_P(BM)==1, non-trivial block selection
 *	    occurs if there are > d_P(DN) blocks, even though no communication
 *	    is actually required. Here a workspace is needed, so the spread
 *	    should be regarded as NOT being trivial.
 *	 2b.
 *	    However, the spread will be still trivial if there are <=
 *	    d_P(DN) blocks (hence <= 1 block per column) and 
 *	    d_P0(BDN) = d_P0(DM)+d_P0(DN) modulo d_P(BDN).
 *	    because of the cost to compute this condition, this optimization
 *	    is not used here.
 */


/* DirectSoleSend() needs only to be valid for a (non-trivial) spread.
 *      for a spread of length K across dimension DK involving communication,
 *	returns true iff there is a single cell that is the source
 *	of the spread, and whether it could have been be sent directly from lB.
 *	In this case, on the source cell, no workspace is required.
 *	
 *	On the receiving cells, this can be used to tell if the spread
 *	has been scaled by alpha. Currently, with the DBLAS, the
 *	receiving cell can locally determine this using the assumption that
 *	whether lB is row or column major is consistent across all cells.
 *	
 */ 

#define DirectSoleSend(K, DK, RowCol, Trans, lB)			\
	(Blk1d_(K, DK) && LM_Inc(1-((Trans)^(RowCol)), lB)==1)

#define OwnerBlk0(DK)			(d_P0(DK)==0)

#define	KPreSprDirectSoleSend(BDN, DN, RowCol, K, n, lB)		\
	(d_P(BDN)==d_P(DN) && LM_Contiguous(1-(RowCol), K, n, lB))
	/* || Blk1d_(N, BDN) && LM_Contiguous(1-(RowCol), K, N, lB) */
 	/* strictly, n = G2Ld(N, BDN); however G2Ld(N, DN) can be
	   safely used if d_P(BDN)==d_P(DN) and KPreSprOwnerBlk0(). 
	   Also, note that Blk1d_(N, BDN) ==> N = n */          

#define KPreSprOwnerBlk0(BDN, DN) 	(d_P0(BDN)==d_P0(DN))
	/* strictly, d_P0(BDN)==d_P0(DN)%d_P(BDN) is the correct expression.
	 * However, the only cases of interest to optimize are where 
	 * d_P(DN)==1 or d_P(DN)==d_P(BDN) where the above is accurate.
         */

#define TransB  	DM_Trans(B)
#define BDK 		DM_D(RowCol^TransB, B)
#define BDN 		DM_D(1-(RowCol^TransB), B)
#define lB 		DM_Loc(B)
#define KPreSpread 	(d_P(BDK)==1)

/***************** low-level `spread' routines ********************/  

/* to calculate the the tasks of determining spread workspaces size, 
 * to initialize them, and to do the spread itself all require the calculation 
 * of several `spread factors'. The spread task itself in turn
 * requires the other two to be performed. To avoid expensive recalculation,
 * and yet preserve modularity, the CalcSpreadFactors() is made as an
 * internal routine. Making it inline reduces significant call overheads.
 */


#define SPREAD_FACTORS_LIST					\
	n_DN, n_WS, NeedRot, TrivialSpread, ScaledSpread, WSsize
#define SPREAD_FACTORS_ALIST					\
	&n_DN, &n_WS, &NeedRot, &TrivialSpread, &ScaledSpread, &WSsize
#define SPREAD_FACTORS_DECLN					\
	int n_DN, n_WS, NeedRot, TrivialSpread, ScaledSpread, WSsize
#define SPREAD_FACTORS_ADECLN					\
	int *n_DN, *n_WS, *NeedRot, *TrivialSpread, *ScaledSpread, *WSsize

static inline void CalcSpreadFactors
	(CalcNOnly, ForceWS, RowCol, K, N, B, DN, M, DM, SPREAD_FACTORS_LIST)
  int CalcNOnly, ForceWS, RowCol, K, N, M; DistMat B; pMatDim DN, DM;
  SPREAD_FACTORS_ADECLN;
{ int NeedWS = ForceWS;
  *n_DN = G2Ld(N, DN);	
  *NeedRot = !TransB && (d_P0(DN)!=d_P0(BDN));
  *n_WS = *NeedRot? G2Ld_Mx(N, DN): *n_DN;
  if (!CalcNOnly && !TransB) {
    *TrivialSpread = SpreadTrivial(K, BDK, M, DM);
    if (*NeedRot || ForceWS)
      NeedWS = *ScaledSpread = 1;
    else if (*TrivialSpread) 
      NeedWS = *ScaledSpread = 0;
    else {  
      int Direct1Send = DirectSoleSend(K, BDK, RowCol, 0, lB);
      NeedWS = !(Direct1Send && OwnerBlk0(BDK));
      *ScaledSpread = !Direct1Send; 
    }
  } else if (!CalcNOnly) {
    if (KPreSpread) {
      *TrivialSpread = TrKPreSpreadTrivial(N, BDN, M, DM, DN);
      NeedWS = !*TrivialSpread && 
	       !(KPreSprDirectSoleSend(BDN, DN, RowCol, K, *n_DN, lB) &&
		 KPreSprOwnerBlk0(BDN, DN));
      *ScaledSpread = 0;
    } else {
      *TrivialSpread = SpreadTrivialP0(K, BDK, d_P0(BDN), d_P(BDN), M, DM);
      *ScaledSpread = ForceWS ||
		      !((d_P(BDK)==1 && d_P(BDN)==1) || 
			(*TrivialSpread && /* a trivial transposition, ie. */ 
			 Blk1d_(N, DN) && Blk1d_(K, BDK) &&
              	         d_P0(BDK)==d_P0(DN) /*nb. d_P0(BDN)==d_P0(BTK)*/));
      NeedWS = *ScaledSpread;	
    }		    	
  }  
  *WSsize = (ForceWS || NeedWS)? FAST_MUL1(K, *n_WS): -1; 
} /* CalcSpreadFactors() */


static inline void InitSpreadWS_(WSsize, n_WS, RowCol, K, B, DN, sB)  
  int WSsize, n_WS; int RowCol, K; DistMat B; pMatDim DN; pLocMat sB; 
{ if (WSsize==-1) {		  
    LM_Copy(TransB^RowCol, *sB, lB);
  } else {			/* for safety, use the largest valid strides */
    if (WSsize==0 && K>0) {	
      /* if later NxK A=Spread2DistMat(...,1,*sB), will ensure valid stride A */
      LM_Inc(0, *sB) = 2;
      LM_Inc(1, *sB) = 1;
    } else {
      LM_Inc(0, *sB) = MAX(n_WS, 1);
      LM_Inc(1, *sB) = 1;
    }
    if (WSsize>0 && LM_ADR(*sB)==NULL)
      MatMalloc_F("InitSpreadWS", 1, &LM_ADR(*sB), WSsize);
   /* else assume >= WSsize words have been pre-allocated at LM_ADR(*sB) */
  }
} /* InitSpreadWS_() */


/***************** high-level `spread' operations ********************/  

int SizeSpreadWS(RowCol, K, N, B, DN, ForceWS, M, DM)
  DistMat B;
  int RowCol, K, N, ForceWS, M; pMatDim DN, DM;
{ SPREAD_FACTORS_DECLN;
  CalcSpreadFactors(ForceWS, ForceWS, RowCol, K, N, B, DN, M, DM,
  		    SPREAD_FACTORS_ALIST);  		    		     
  return(WSsize);
} /*SizeSpreadWS()*/


int InitSpreadWS(RowCol, K, N, B, DN, sB, ForceWS, M, DM)
  DistMat B; pMatDim DN, DM;
  int RowCol, K, N, ForceWS, M; 
  pLocMat sB;
{ SPREAD_FACTORS_DECLN;
  CalcSpreadFactors(ForceWS, ForceWS, RowCol, K, N, B, DN, M, DM,
  		    SPREAD_FACTORS_ALIST);
  InitSpreadWS_(WSsize, n_WS, RowCol, K, B, DN, sB);
  LM_Trans(RowCol, *sB);
  return(WSsize>=0? n_DN: -n_DN);
} /* InitSpreadWS() */
  

int SpreadMat(RowCol, K, N, B, alpha, DN, sB, ForceWS, M, DM)
  DistMat B; 
  int RowCol, K, N, ForceWS, M; pMatDim DN, DM;
  FDATA *alpha; pLocMat sB;
{ SPREAD_FACTORS_DECLN;
  LocMat lBt; 			/* lBt is k x n */
  LM_Copy(RowCol^TransB, lBt, lB);  	

  COMMON_ASSERT_TR_E(
    (TraceStr, "(%d,%dx%d,%s,%1.1f,%d %d:%d %d,_,%d,%d,%d %d:%d %d)", 
     RowCol, K, N, PrintDistMat(B, TmpTraceStrs[0]), *alpha, d_B0(DN), d_B(DN),
     d_P0(DN), d_P(DN), ForceWS, M, d_B0(DM), d_B(DM), d_P0(DM), d_P(DM)),
    RCOK(RowCol) && CompatBlk(BDN,DN) && K>0 && N>0 && M>=0, 
    0, "SpreadMat", 2, TRACE_ALL
  );

  CalcSpreadFactors(0, ForceWS, RowCol, K, N, B, DN, M, DM, 
  		    SPREAD_FACTORS_ALIST);
  InitSpreadWS_(WSsize, n_WS, RowCol, K, B, DN, sB);
  
  if (TransB && KPreSpread) {
    if ((!TrivialSpread && n_DN>0) || WSsize>0)
      /* assert LM_ADR(sB) != NULL */
      TransReSpreadMat(RowCol, K, N, &lBt, BDN, NULL, DN, sB, WSsize==-1);
      /* note: only source cell can accurately determine if scaling was applied
         in this case; thus NULL (no scaling) is passed rather than alpha */
  } else {
    pLocMat sBp; LocMat sBp_WS;		/*1st part of sB this cell contributes*/
    MatDim BDTKs;
    pMatDim BTDK = !TransB? BDK: 
    		   pInitDim(&BDTKs, d_B0(BDK), d_B(BDK), d_P0(BDN), d_P(BDN));

    sBp = K==1? sB: pLocSubMat_(&sBp_WS, sB, BlkIxP0d(BTDK), 0);

    if (!TransB && WSsize>0) {		/* assert LM_ADR(*sB) != NULL */
      int k = G2Ld(K, BDK);		int n = NeedRot? G2Ld(N, BDN): n_DN;
      if (k==1)				/* optimize for low overhead */
        MatCopy(1, n, &lBt, alpha, sBp);
      else if (k > 0)  
        BlkMatStrideUnpack(0, k, n, &lBt, alpha, sBp, 
             	           BlkB0(d_P0(BDK),d_B0(BDK)),d_B(BDK),d_P(BDK), 0,1,1);
      if (NeedRot) 
        RotMat(1-RowCol, d_P0(BDN)-d_P0(DN), k, BDK, n_WS, pLM_ADR(sBp));
    }
    if (TransB && WSsize!=-1)
      /*nb. if WSsize=0, sBp wont be accessed, but may still send from lB */
      TransMat(1-RowCol, 0, K, N, &lBt, BDK, BDN, 1, alpha,  
     	       sBp, d_P0(DN), d_P0(BDN), d_P(BDN));

    if (!TrivialSpread && n_WS>0) 
      /* assert LM_ADR(sB) != NULL */
      if (K==1)				/* reduce overhead */
	BrdMat(RowCol, d_P0(BTDK), 1, n_WS, pLM_ADR(sB), n_WS); 
      else 
        GatherMat(RowCol, BTDK, K, n_WS, sB);
    if (ScaledSpread)
      *alpha = FD_1;			/*source has scaled data by *alpha */
  }

  if (PrintRes && (WSsize>0 || ForceWS)) {
    char sBName[4];			
    sBName[0] = 's';			sBName[1] = RowCol? 'A': 'B';
    sBName[2] = TransB? 'T': ' ';	sBName[3] = '\0';
    MatPrint(sBName, MIN(PrintRes, K), MIN(PrintRes, n_DN), *sB); 	
  }
  LM_Trans(RowCol, *sB);
  TRACE_EXIT;
  return(WSsize>=0? n_DN: -n_DN);
} /* SpreadMat() */


/***************** high-level `reduce' operations ********************/  

/* the reduce operation is in many senses the complement of the
 * spread. However, one main difference is that its workspace
 * must be established before the reduce  proper.
 * Factors required for the optimization of reduce operations are
 * somewhat simpler than for the spread; this, and the fact that
 * the workspace must be intitialzed separately (hence the calculations
 * required by both cant be optimized), gives the implmentation below.
 */

#define SoleOwner(K, DK)	(Blk1d(K, DK) && OwnerBlk0(DK))

 
static inline int CalcReduceWSSize
	(ForceWS, RowCol, K, N, B, DN, M, DM, n_WS, n_DN)
  int ForceWS, RowCol, K, N, M; DistMat B; pMatDim DN, DM; int *n_WS, *n_DN;
{ int NeedWS = ForceWS;
  int NeedRot = !TransB && (d_P0(DN)!=d_P0(BDN));
  *n_DN = G2Ld(N, DN);	
  *n_WS = NeedRot? G2Ld_Mx(N, DN): *n_DN;
  if (!NeedWS && !TransB) 
    NeedWS = NeedRot || (!SpreadTrivial(K, BDK, M, DM) && !SoleOwner(K, BDK));
  else if (!NeedWS) 
    NeedWS = !((d_P(BDK)==1 && d_P(BDN)==1) || 
	      (SpreadTrivialP0(K, BDK, d_P0(BDN), d_P(BDN), M, DM) && 
	       /* a trivial transposition, ie. */ 
	       Blk1d_(N, DN) && Blk1d_(K, BDK) &&
               d_P0(BDK)==d_P0(DN) /*nb. d_P0(BDN)==d_P0(BTK)*/));
  return (NeedWS? FAST_MUL1(K, *n_WS): -1); 
} /* CalcReduceWSSize() */


int SizeReduceWS(Source, RowCol, K, N, B, DN, ForceWS, M, DM)
  int Source, RowCol, K, N, ForceWS, M; DistMat B; pMatDim DN, DM;
{ int n_WS, n_DN;
  return (CalcReduceWSSize(ForceWS, RowCol, K, N, B, DN, M, DM, &n_WS, &n_DN));
} /*SizeReduceWS()*/

static struct {		
/* used to avoid recalculation of G2Ld(N, DN) between 
 * corresponding InitReduceWS(); ReduceMat() calls. 
 */
  FDATA *ADR_sB;		
  /* to check that n_DN was set by the corresp. InitReduceWS() call */  
  int Nval, nval;
} Reduce_n_DN = {NULL, -1, -1};

int InitReduceWS(RowCol, K, N, B, alpha, DN, sB, ForceWS, M, DM)
  DistMat B; FDATA *alpha; pMatDim DN, DM;
  int RowCol, K, N, ForceWS, M; 
  pLocMat sB;
{ int n_WS, n_DN;
  int WSsize = CalcReduceWSSize(ForceWS, RowCol, K, N, B, DN, M, DM, &n_WS, &n_DN);

  COMMON_ASSERT_TR_E(
    (TraceStr, "(%d,%dx%d,%s,%1.1f,%d %d,_,%d,%d %d %d)", 
     RowCol, K, N,  PrintDistMat(B, TmpTraceStrs[0]), pFD_VAL(alpha), 
     d_P0(DN), d_P(DN), M, d_B0(DM), d_B(DM), d_P(DM)),
    RCOK(RowCol) && CompatBlk(BDN,DN) && K>=0 && N>=0 && M>=0, 
    0, "InitReduceWS", 2, TRACE_ALL
  );

  if (WSsize==-1) {	     /* => there is a sole `owner' for the reduce opn */
    LM_Copy(TransB^RowCol, *sB, lB);
    if (d_P0(BDK)==0)			/* this cell row/col is `sole owner' */
      MatScale(K, n_WS, alpha, sB);	/* nb. K is also local length here */
  } else {
    int Ld = MAX(n_WS, 1);	
    LM_Inc(0, *sB) = Ld;
    LM_Inc(1, *sB) = 1; 
    if (WSsize>0) {  
      if (LM_ADR(*sB)==NULL)
        MatMalloc_F("InitReduceWS", 1, &LM_ADR(*sB), WSsize);
      /* else assume >= WSsize words have been pre-allocated at LM_ADR(*sB) */  
      if (K > 0)
        MatZero(K, Ld, sB);
    }  
  }
  LM_Trans(RowCol, *sB);		/* *sB is RowCol==0? K x n: n x K */
  Reduce_n_DN.ADR_sB = pLM_ADR(sB);	Reduce_n_DN.nval = n_DN;
  Reduce_n_DN.Nval = N;
  TRACE_EXIT;
  return(WSsize>=0? n_DN: -n_DN);
} /* InitReduceWS() */
  


void ReduceMat(RowCol, K, N, B, beta, DN, sB, M, DM)
  DistMat B; 
  int RowCol, K, N, M; pMatDim DN, DM;
  FDATA *beta; pLocMat sB;
{ int NoWSsB = LM_ADR(lB)==pLM_ADR(sB);
  int n = (pLM_ADR(sB)==Reduce_n_DN.ADR_sB && N==Reduce_n_DN.Nval)? 
  	  Reduce_n_DN.nval:  G2Ld(N, DN);
  LocMat sBt, lBt;			/* lBt, sBt are k x n */
      
  MatDim BDTKs;
  pMatDim BTDK = !TransB? BDK: 
    		 pInitDim(&BDTKs, d_B0(BDK), d_B(BDK), d_P0(BDN), d_P(BDN));

  Reduce_n_DN.Nval = -1;		/* invalidate Reduce_n_DN */

  COMMON_ASSERT_TR_E(
    (TraceStr, "(%d,%dx%d,%s,%1.1f,%d %d:%d %d,a%d %d %d,%d,%d %d %d)", 
     RowCol, K, N,  PrintDistMat(B, TmpTraceStrs[0]), pFD_VAL(beta),
     d_B0(DN), d_B(DN), d_P0(DN), d_P(DN), FD_ADR(pLM_ADR(sB)), 
     pLM_RInc(sB), pLM_CInc(sB), M, d_B0(DM), d_B(DM), d_P(DM)),
     RCOK(RowCol) && CompatBlk(BDN,DN) && K>=0 && N>=0 && M>=0,
     0, "ReduceMat", 2, TRACE_ALL
  );
  if (K==0 || N==0) {
    TRACE_RETURN;	
  }

  LM_Copy(RowCol, sBt, *sB);		LM_Copy(RowCol^TransB, lBt, lB);

  if (!SpreadTrivial(K, BTDK, M, DM) && n>0) {
    /* note: for a small optimization, d_P(BTDK) <- MIN(NBLKSd(K, BTDK), d_P(BTDK))
       as only this many cells contribute non-zero elements to the sum  */
    /* assert LM_ADR(sB) != NULL */   
    if (PrintRes)
      MatPrint("sC ", MIN(PrintRes, K), MIN(PrintRes, n), sBt);
    if (K==1)  
      SumMat(RowCol, d_P0(BTDK), d_P(BTDK), 1, n, &sBt);
    else  
      MultSumMat(RowCol, BTDK, K, n, &sBt);
  }
  if (PrintRes && LM_ADR(sBt)!=NULL)
    MatPrint("sC ", MIN(PrintRes, K), MIN(PrintRes, n), sBt);

  if (!NoWSsB) {			/* assert LM_ADR(sB) != NULL */
    int RotFactor = d_P0(DN)-d_P0(BDN),	AddMat = pFD_UNITY(beta) || !FD_EQ(*beta, FD_0);
    int kB = G2Ld(K, BDK),		nB = (RotFactor|TransB)? G2Ld(N, BDN):n;
    LocMat sBp_WS;         		/*1st part of sB this cell contributes*/
    pLocMat sBp = K==1? &sBt: pLocSubMat_(&sBp_WS, &sBt, BlkIxP0d(BTDK), 0);  

    if (AddMat && !pFD_UNITY(beta))
       MatScale(kB, nB, beta, &lBt);
    if (!TransB) {
      if (RotFactor && kB>0)	/* as N>0, assert LM_ADR(sB) != NULL  */
        RotMat(1-RowCol, RotFactor, kB, BDK, pLM_Ld(sBp), pLM_ADR(sBp));
      if (kB==1) 
        (AddMat? MatAdd: MatCopy)(1, nB, sBp, NULL, &lBt);
      else if (kB>0)   
        BlkMatStridePack(AddMat, kB, nB, sBp, NULL, &lBt,
      		         BlkB0(d_P0(BDK), d_B0(BDK)), d_B(BDK), d_P(BDK), 0,1,1);
    } else
      TransMat(RowCol, AddMat, K, N, sBp, BTDK, DN, d_P(BTDK), NULL,
     	       &lBt, d_P0(BDN), d_P0(BDK), 1);
  }

  if (PrintRes && LM_ADR(lBt)!=NULL) {
    int kB = G2Ld(K, BDK),		nB = G2Ld(N, BDN);
    char sBname[4];
    sBname[0] = 'C';			sBname[1] = RowCol==0? '0': '1';
    sBname[2] = TransB? 'T': ' ';	sBname[3] = '\0';
    MatPrint(sBname, MIN(PrintRes, kB), MIN(PrintRes, nB), lBt); 
  }   
  TRACE_EXIT;
} /* ReduceMat() */

#undef BDN
#undef BDK
#undef lB
#undef KPreSpread 


