/* @(#) All Right Reserved, Copyright (c) 1997 Australian National University */
/* @(#)                     Copyright (c) 1997 Fujitsu Laboratories Ltd. */
/*  */
/*        */
/*                        Permission Notice */
/*  */
/*                Australian National University */
/*                Canberra ACT 0200 Australia */
/*                           and */
/*                Fujitsu Laboratories Ltd. */
/*                4-1-1 Kamikodanaka Nakahara */
/*                Kawasaki 211-88 Japan */
/*  */
/*   Permission to use, copy, modify and distribute this software (including  */
/*   its documentation) without fee is hereby granted under the following terms */
/*   and conditions: */
/*  */
/*   1. This permission notice and all of the copyright notices in this */
/*      software must appear in all copies and derived versions of all */
/*      or a part of this software. */
/*  */
/*   2. Sales of this software, whether incorporated in other products as a */
/*      part thereof or not, is prohibited. */
/*  */
/*   3. On publication or presentation in any form of the results obtained */
/*      by use of this software, an explicit reference to this software and */
/*      its copyright owner shall be made. */
/*  */
/*   4. THIS SOFTWARE IS PROVIDED 'AS IS' AND THE AUSTRALIAN NATIONAL UNIVERSITY */
/*      AND FUJITSU LABORATORIES LTD. MAKE NO WARRANTIES, EITHER EXPRESSED OR */
/*      IMPLIED, AS TO ANY MATTER WHATSOEVER REGARDING THIS SOFTWARE,  */
/*      INCLUDING, WITHOUT LIMITATION, WARRANTIES OF MERCHANTABILITY OR FITNESS */
/*      FOR ANY PARTICULAR PURPOSE. */
/*      IN NO EVENT SHALL THE AUSTRALIAN NATIONAL UNIVERSITY OR FUJITSU */
/*      LABORATORIES LTD. BE LIABLE FOR ANY CLAIM FROM A THIRD PARTY, OR */
/*      FOR SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES WHATSOEVER RESULTING */
/*      FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, */
/*      NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION */
/*      WITH THE USE, COPYING OR MODIFICATION OF THIS SOFTWARE. */
/*  */
 
/************************************************************************/
/* TriMatSolve: DBLAS Tri Mat Solve alg					*/
/************************************************************************/
/* SCCS INFO @(#)TriMatSolve.c 1.2         last modified P. Strazdins, 97/04/24
   first created: P. Strazdins 97/03
*/
/* compilation: macro FDATATYPE must be defined;
 *              if macro ASSERT defined, assertion checking turned on
 *              if macro TRACE defined, parameter value tracing enabled
 */

#include "TriMatSolve.h"

#include "BLASOptions.h"
#include "FDATAOps.h"
#include "FDATAAssert.h"

#include "LocMat.h"
#include "MatDim.h"
#include "DistMat.h"
#include "MachineParams.h"
#include "MatDim.h"   
#include "GetBlockFactor.h"
#include "FDATACellMatrix.h"
#include "SpreadMatrix.h"
#include "TRSM.h"

#define ADM DM_DM(A)
#define ADN DM_DN(A)
#define BDM DM_DM(B)
#define BDN DM_DN(B)  
#define lA DM_Loc(A)
#define lB DM_Loc(B)

/* TriMatSolve() is suitable if B distributed over > 1 cell columns 
 *  ie >1 N-blocks.
 * It uses wide blocking is used to achieve load balance, as well as 
 * improved cell computation performance, with A partitioned vertically 
 * to also utilize load balamce from > 1 N-blocks.
 * The TR_M() routine is used to perform the final update on strips of B,
 * using the triangular parts (near the diagonal) of the A strips. 
 * Communication startups include 1 column b/c for each row of B, 
 * and (at most) 1 row b/c for each column of A; 
 * if DM_Trans(B)!=DM_Trans(A), there will also be a transpose
 * operation associated with the spread of A.
 * Note that the volumne of communication of A will be insignificant for
 * large M.
 */
#define CALL_TR_M
/* if CALL_TR_M defined, internal blocking is performed in the update of B
 * using the triangular pieces of A; this increases commun. startup 
 * (& s.w overheads) slightly but results in slighly less communication
 * and better cell computation 
 */

void TriMatSolve(UpLo, Diag, M, N, alpha, A, B)
  char UpLo, Diag; int M, N;
  FDATA *alpha; DistMat A, B; 
{ int n;                         	
  int TrA = DM_Trans(A);
  int UpLoTr = ((UpLo==Upper)^DM_Trans(B))^TrA;
  DistMatStore BN;

  COMMON_ASSERT_TR_E(
    (TraceStr, "(%c%c,%dx%d,%1.1f,%d a%d %d %d,%d a%d %d %d)",
     UpLo, Diag, M, N, *alpha, DM_Trans(A), FD_ADR(LM_ADR(lA)), LM_RInc(lA), 
     LM_CInc(lA), DM_Trans(B), FD_ADR(LM_ADR(lB)), LM_RInc(lB), LM_CInc(lB)),
     CompatBlk(BDM, ADN) && M>=0 && N>=0 , 0, "TriMatSolve", 1, TRACE_ALL
  ); 
 
  if (M==0 || N==0) {
    TRACE_RETURN;
  }
     
  n=0;
  while (n < N) {
    int dN = MIN(PAR_OUTER_VECSIZE, N-n), nB /* = G2Ld(dN, BDN)*/;
    LocMat BsN, BsNj;                	/* reduce workspaces of B */
    int BsNn;
    int j = UpLoTr?  M-1:0, 		jF = UpLoTr? -1: M;
    int mA;				/* = G2Ld(M, ADM) */ 
    
    LM_ADR(BsN) = NULL;
    BsNn = InitReduceWS(1-TrA, dN, M, TrA? Trans(B): B, NULL, ADM, &BsN, 0, 
    			M, ADN);
    mA = ABS(BsNn);
    LM_Trans(TrA, BsN);			/* make mA x dN */
    pLocSubMat_(&BsNj, &BsN, UpLoTr? mA: 0, 0);

#ifdef CALL_TR_M
    nB =  G2Ld(dN, BDN);
#endif

    while (j != jF) {
      int Mj = UpLoTr? j+1: M-j,	Kj = UpLoTr? M-1-j: j; 	
      int w = GenGetW(Mj, 2.0, d_B(ADN)*d_P(ADN), 
      		      MIN(Mj, OUTER_VECSIZE*d_P(ADN)), ADN, j, UpLoTr? -1: +1);
      int jL = !UpLoTr? j: j-w+1,	jR = !UpLoTr? j+w-1: j;         
      DistMat_RVAlloc(3);
      DistMat Ajw = SubMat(A, UpLoTr? 0: jR+1, jL),
      	      Ajj = SubMat(A, jL, jL),
      	      BjL = SubMat(B, jL, 0),	BjLt = TrA? Trans(BjL): BjL;
      LocMat lBjL;
      int w_m = G2Ld(w, DM_DM(Ajj)); 
      if (UpLoTr)
        pLocSubMat(&BsNj, -w_m, 0);
      if (Kj>0) {			/* must pack BsNj & reduce into BjL */
        pMatDim AjwDN;			MatDim AjjwDN;
        FDATA *W = NULL;		int Ld = MAX(w_m, 1); 
        int NeedRot = !DM_Trans(BjLt) && d_P0(DM_DM(Ajj))!=d_P0(DM_DM(BjL));
        LocMat BjLsN;			/* corresp. B[jL..jR, [:]] */
      	/* note: InitReduceWS() is not used here to set BjLsN because: 
      	 *       1) it would perform unnecessary scaling/clearing 
	 *	 2) the conditions for avoiding pack are slightly different
	 * nontheless, the requirements for BjLsN of ReduceMat() must be 
	 * carefully followed here
         */
        if ((dN==1 && !NeedRot) || BsNn<=0) { 
          LM_Copy(0, BjLsN, BsNj);	/* can avoid pack */
          if (BsNn>0)
            LM_CInc(BjLsN) = Ld,       LM_RInc(BjLsN) = 1;
        } else {
          if (NeedRot) 			/* need extra space for alignment */
            Ld = G2Ld_Mx(w, DM_DM(Ajj));	
	  MatMalloc_FT("TriMatSolve", 1, &W, Ld*dN);
      	  LM_Init(BjLsN, W, 1, Ld);
          MatCopy(w_m, dN, &BsNj, NULL, &BjLsN);
        }
        LM_Trans(TrA, BjLsN);		/* orient for ReduceMat() */
        AjwDN = UpLoTr? pSubDim_(&AjjwDN, DM_DN(Ajj), w): DM_DN(Ajw);
        ReduceMat(1-TrA, dN, w, BjLt, NULL, DM_DM(Ajj), &BjLsN, Kj, AjwDN);
        if (W!=NULL)
          MatFree(W);
      }
      if (!UpLoTr)
        pLocSubMat(&BsNj, w_m, 0);
      LM_Copy(DM_Trans(B), lBjL, DM_Loc(BjL));/* make w_m x nB */  
#ifdef CALL_TR_M
      TR_M(1, UpLo, Diag, w, dN, &FD_1, Ajj,  BjL);
      if (d_P0(DM_DM(Ajj))!=d_P0(DM_DM(BjL)) || d_P(DM_DM(Ajj))!=d_P(DM_DM(BjL)))
        w_m = G2Ld(w, DM_DM(BjL));	/* redefine w_m in terms of B*/
#else
      { int Side = DM_Trans(BjL);
        LocMat AsM, BsM;		/* spread workspaces for Ajj BjL */
        int AsMn, BsMn;                 /* = workspace lengths for AsM, BsM */
        FDATA AsMScale = FD_M1;

	LM_ADR(AsM) = NULL;
	AsMn = SpreadMat(1-Side, w, w, Ajj, &AsMScale, DM_DM(BjL), &AsM, 
		  	 0, dN, DM_DN(BjL));
        w_m = ABS(AsMn);		/* w_m is now in terms of BjL */
        LM_Trans(Side, AsM);  		/* AsM is w_m x w */
	LM_ADR(BsM) = NULL;
	BsMn = InitSpreadWS(Side, w, dN, BjL, BDN, &BsM, 0, w, DM_DM(BjL));
        nB = ABS(BsMn);
        LM_Trans(Side, BsM);   		/* BsM is w x nB */
                     
      	SpreadL2TR_M(1, UpLoTr, Diag, Side, w, w_m, nB, DM_DM(BjL),
                     &AsMScale, &AsM, &BsM, &lBjL, 1-Blk1d(w, DM_DN(BjL)));
                     
        if (AsMn > 0)
          MatFree(LM_ADR(AsM));            
        if (BsMn > 0)
          MatFree(LM_ADR(BsM));            
      }
#endif      
     
      if (Mj-w > 0) {
	int mA = G2Ld(Mj-w, DM_DM(Ajw));
        FDATA AsMScale = FD_M1;        
        LocMat BTsN;			/* corresp. B[jL..jR, [:]]^T */
        int BTsNn, w_n;			/* workspace lengths for BTsM */
        LM_ADR(BTsN) = NULL;
        BTsNn = SpreadMat(TrA, dN, w, Trans(BjLt), &AsMScale,
      			  DM_DN(Ajw), &BTsN, 0, Mj-w, DM_DM(Ajw)); 	
        w_n = ABS(BTsNn);			  

        LM_Trans(TrA, DM_Loc(Ajw));	/* make mA x w_n */
        LM_Trans(1-TrA, BTsN);		/* make w_n x dN */
        MatPartMult(mA, dN, w_n, &AsMScale, &DM_Loc(Ajw), &BTsN, NULL, 
      		    UpLoTr? &BsN: &BsNj);
        if (BTsNn>0)
          MatFree(LM_ADR(BTsN)); 
      }	

      MatScale(w_m, nB, alpha, DM_pLoc(BjL));
      
      j += (UpLoTr? -w: w);              
    } /* while (j...) */
     
    if (BsNn>0)
      MatFree(LM_ADR(BsN));      
    n += dN;
    if (n < N)
      B = SubMat_(&BN, B, 0, dN);
  } /* while (n...) */
  TRACE_EXIT;
} /*TriMatSolve()*/


/* TriMatSolve1Blk() communicates only B (a b/c & global sum for each row), 
 * and must split the computation in single blocks of B in the N-dimension. 
 * Effort has been expended to reduce software overheads also;
 * for this reason, spread/reduce calls are not directly used.
 */

#include "FDATAMalloc.h"
#include "FDATACommun.h"		/* BrdMat(), SumMat() */
#include "D_BLAS.h"   			/* D_ScaleAdd() */

static inline InnerProd(k, A, IncA, B, IncB, C)
  int k, IncA, IncB; FDATA *A, *B, *C;
/* pre:  k>0;
 * post: *C +=(Sum i: 0<=i<k: A[i]*B[i])
 */
{ FDATA c = *C;
  do {
    c += *A * *B;                       k--;
    A += IncA;                          B += IncB;
  } while (k > 0);
  *C = c;
} /* InnerProd() */

void TriMatSolve1Blk(UpLo, Diag, M, N, alpha, A, B)
  char UpLo, Diag; int M, N;
  FDATA *alpha; DistMat A, B; 
{ int n;                         	
  int TrA = DM_Trans(A);
  int UpLoTr = ((UpLo==Upper)^DM_Trans(B))^TrA;
  DistMatStore BN;

  COMMON_ASSERT_TR_E(
    (TraceStr, "(%c%c,%dx%d,%1.1f,%d a%d %d %d,%d a%d %d %d)",
     UpLo, Diag, M, N, *alpha, DM_Trans(A), FD_ADR(LM_ADR(lA)), LM_RInc(lA), 
     LM_CInc(lA), DM_Trans(B), FD_ADR(LM_ADR(lB)), LM_RInc(lB), LM_CInc(lB)),
     CompatBlk(BDM, ADN) && M>=0 && N>=0 , 0, "TriMatSolve1Blk", 1, TRACE_ALL
  ); 
 
  if (M==0 || N==0) {
    TRACE_RETURN;
  }
     
  n=0;
  while (n < N) {
    int dN = MIN(PAR_OUTER_VECSIZE, MIN(d_B(DM_DN(B))-d_B0(DM_DN(B)), N-n)); 
    LocMat BTsN,                	/* kA x dN spread workspace of B^T */
    	   BTsNj; 			/* -> row j in BTsN */
#define dj  (UpLoTr? -1: +1)
    int kA = G2Ld(M, ADN);
    int j = UpLoTr?  M-1: 0, 		jF = UpLoTr? -1: M;
    DistMat_RVAlloc(2);
    DistMat Ajj = SubMat(A, j, j), 	BT = SubMat(A, 0, 0);   
    int kj = 0;				/* local length of MM */
    
    LM_Init(BTsN, NULL, dN, 1);		
    MatMalloc("TriMatSolve", 1, &LM_ADR(BTsN), kA*dN);  
    pLocSubMat_(&BTsNj, &BTsN, UpLoTr? kA: -1, 0);

    /* the following is similar to SpreadMat(TrA, dN, M, TrA? Trans(B): B,...)
     * except this this would produce BTsN with the opposite storage 	*/     
    LM_Copy(1-TrA, DM_Loc(BT), BTsN); 
    pInitDim(DM_DM(BT), d_B0(DM_DN(B)),  d_B(DM_DN(B)), 
      			d_P0(DM_DM(BT)), d_P(DM_DM(BT)));
    D_ScaleAdd(M, dN, 1.0, B, 0.0, Trans(BT));	
    BrdMat(TrA, d_P0(DM_DM(BT)), kA, dN, LM_ADR(BTsN), dN); 

    while (j != jF) {
      /* invariants: kj = G2Ld(Kj, DM_DN(Ajw));     Ajj = SubMat(A, j, j);
       * 	     BTsNj = BTsN[UpLoTr? kA-kj-ki: kj+ki-1, 0] 
       */
      int Mj = UpLoTr? j+1: M-j,	Kj = UpLoTr? M-1-j: j; 	
      int w = GenGetW(Mj, 2.0, d_B(ADM)*d_P(ADM), 
      		      MIN(Mj, INNER_VECSIZE*d_P(ADM)), DM_DM(Ajj), j, dj);
      int jL = UpLoTr? j-w+1: j;
      DistMat_RVAlloc(1);
      DistMat Ajw = SubMat(A, jL, UpLoTr? j+1: 0);
      int w_m = G2Ld(w, DM_DM(Ajw));		
      LocMat BsN; 			/* w_m x dN reduce workspace of B[j,0]*/
      LocMat BsNj;			/* -> row |j-jL| in BsN */
      FDATA *BsNWS=NULL;
      int i, ki=0;			/* ki is MM length in i-loop */

      if ((Blk1d_(Kj+w, DM_DN(Ajw)) && d_P0(DM_DN(BT))==d_P0(DM_DN(Ajw))) ||
      	  (d_P(DM_D0(Ajw))==1 && d_P(DM_D1(Ajw))==1)) { 
        /* the reduction on BsN will be trivial */
        pLocSubMat_(&BsN, &BTsNj, UpLoTr? -w_m: 1, 0);
      } else {
        MatMalloc_F("TriMatSolve", 1, &BsNWS, w_m*dN);  
        LM_Init(BsN, BsNWS, dN, 1);
        MatZero(w_m, dN, &BsN);   
        pLocSubMat_(&BsNj, &BsN, UpLoTr? w_m-1: 0, 0);
      }  
      
      if (kj > 0) {
      	LocMat Amm;      		LM_Copy(TrA, Amm, DM_Loc(Ajw));
        MatPartMult(w_m, dN, kj, NULL, &Amm, UpLoTr? &BTsNj: &BTsN, NULL, &BsN);
      }
      
      for (i=0; i<w; i++) {
        if (d_P0(DM_DM(Ajj))==0 && ki>0) { /* final update for jth row of BsN */
          LocMat lAj;			LM_Copy(TrA, lAj, DM_Loc(Ajj));
          if (!BsNWS) 
            pLocSubMat_(&BsNj, &BTsNj, dj, 0);
          if (!UpLoTr) {
            pLocSubMat(&lAj, 0, -ki);	LM_ADR(BTsNj) -= (ki-1)*dN;
          } else if (d_P0(DM_DN(Ajj))==0) 
            pLocSubMat(&lAj, 0, 1); 	/* skip over diagonal */
	  if (dN==1)
	    InnerProd(ki, LM_ADR(lAj), LM_CInc(lAj), LM_ADR(BTsNj), 1, 
		      LM_ADR(BsNj));
	  else
            MatMult(1, dN, ki, NULL, &lAj, &BTsNj, NULL, &BsNj);
          if (!UpLoTr)
             LM_ADR(BTsNj) += (ki-1)*dN;
        }
        
	if (d_P0(DM_DN(Ajj))==0) {	/* only here will spread update BTsNj */
	  LM_ADR(BTsNj) += UpLoTr? -dN: dN;
	  ki++;
	}  

	if (d_P0(DM_DM(Ajj))==0) {
	  FDATA Ajje = -1.0;     
	  if (d_P0(DM_DN(Ajj))==0 && Diag!=Unit) 
	    Ajje = Ajje / *LM_ADR(DM_Loc(Ajj));
	  if (BsNWS) {	
	    SumMat(1-TrA, d_P0(DM_DN(Ajj)), d_P(DM_DN(Ajj)), 1, dN, &BsNj); 
	    if (d_P0(DM_DN(Ajj))==0)
	      if (dN==1) 
		*LM_ADR(BTsNj) = (*LM_ADR(BTsNj) + *LM_ADR(BsNj)) * Ajje;
	      else {
		MatAdd(1, dN, &BsNj, NULL, &BTsNj);
		MatScale(1, dN, &Ajje, &BTsNj);
	      }
            pLocSubMat(&BsNj, dj, 0);
	  } else if (d_P0(DM_DN(Ajj))==0) {
	    if (dN==1)
              *LM_ADR(BTsNj) *= Ajje;
	    else
	      MatScale(1, dN, &Ajje,  &BTsNj); 
	  } 	 	
	}   
	if (d_P0(DM_DN(Ajj))==0) 	
	  BrdMat(TrA, d_P0(DM_DM(Ajj)), 1, dN, LM_ADR(BTsNj), dN);

	Ajj = SubMat_(Ajj, Ajj, dj, dj); 
        j += dj;
      } /* for (i...) */
      kj += ki;

      if (BsNWS)	
        MatFree(BsNWS); 
    } /* while (j...) */

    D_ScaleAdd(M, dN, -pFD_VAL(alpha), Trans(BT), 0.0, B);	        
    MatFree(LM_ADR(BTsN));      
     
    n += dN;
    if (n < N)
      B = SubMat_(&BN, B, 0, dN);
  } /* while (n...) */
  TRACE_EXIT;
} /*TriMatSolve1Blk()*/



