/* @(#) All Right Reserved, Copyright (c) 1997 Australian National University */
/* @(#)                     Copyright (c) 1997 Fujitsu Laboratories Ltd. */
/*  */
/*        */
/*                        Permission Notice */
/*  */
/*                Australian National University */
/*                Canberra ACT 0200 Australia */
/*                           and */
/*                Fujitsu Laboratories Ltd. */
/*                4-1-1 Kamikodanaka Nakahara */
/*                Kawasaki 211-88 Japan */
/*  */
/*   Permission to use, copy, modify and distribute this software (including  */
/*   its documentation) without fee is hereby granted under the following terms */
/*   and conditions: */
/*  */
/*   1. This permission notice and all of the copyright notices in this */
/*      software must appear in all copies and derived versions of all */
/*      or a part of this software. */
/*  */
/*   2. Sales of this software, whether incorporated in other products as a */
/*      part thereof or not, is prohibited. */
/*  */
/*   3. On publication or presentation in any form of the results obtained */
/*      by use of this software, an explicit reference to this software and */
/*      its copyright owner shall be made. */
/*  */
/*   4. THIS SOFTWARE IS PROVIDED 'AS IS' AND THE AUSTRALIAN NATIONAL UNIVERSITY */
/*      AND FUJITSU LABORATORIES LTD. MAKE NO WARRANTIES, EITHER EXPRESSED OR */
/*      IMPLIED, AS TO ANY MATTER WHATSOEVER REGARDING THIS SOFTWARE,  */
/*      INCLUDING, WITHOUT LIMITATION, WARRANTIES OF MERCHANTABILITY OR FITNESS */
/*      FOR ANY PARTICULAR PURPOSE. */
/*      IN NO EVENT SHALL THE AUSTRALIAN NATIONAL UNIVERSITY OR FUJITSU */
/*      LABORATORIES LTD. BE LIABLE FOR ANY CLAIM FROM A THIRD PARTY, OR */
/*      FOR SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES WHATSOEVER RESULTING */
/*      FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, */
/*      NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION */
/*      WITH THE USE, COPYING OR MODIFICATION OF THIS SOFTWARE. */
/*  */

/************************************************************************/
/* SYR2KMult: DBLAS Symmetric Rank-2K Multiplication procedures		*/
/************************************************************************/
/* SCCS INFO @(#)SYR2KMult.c 4.2         last modified P. Strazdins, 97/04/22
   first created: P. Strazdins  95/04/02
*/
/* compilation: macro FDATATYPE must be defined;
 *              if macro ASSERT defined, assertion checking turned on
 *              if macro TRACE defined, parameter value tracing enabled
 */

#include "SYR2KMult.h"
#include "BLASOptions.h"
#include "FDATAOps.h"
#include "FDATAAssert.h"

#include "LocMat.h"
#include "DistMat.h"
#include "MachineParams.h"
#include "FDATACellTriMatrix.h"
#include "SpreadMatrix.h"


#define Mergeable(RC, K, A, B) 						\
	/* true if B & A[0..K-1] can be merged into a single matrix */	\
	(LM_RInc(A)==LM_RInc(B) && LM_CInc(A)==LM_CInc(B) && 		\
	 LM_ADR(A)+LM_Inc(RC, A)*(K)==LM_ADR(B))	

void DistSYR2KMult(N, K, alpha, A, B, beta, FormC, C)
  int N, K; FDATA *alpha, *beta; 
  DistMat A, B, C; char FormC;
/* notes: this algorithm exploits the SYRK computation in the following 2 ways:
 *	1. (parallel) explicit transpose of A, B is avoided as for SYRK.
 *		
 *	   However, if A & B were aligned (normally the case),
 *	   it would be possible to merge them before spreading them;
 *	   this would reduce commun. startup by a factor of 2. 
 *	   Because of the extra S/E difficulty, this optimization is not
 *	   curently used.
 *	2. (cellwise) if K is too small for optimal cell matrix multiply,
 *	   workspaces are forced for A and B which can be merged to form
 *	   a 2*K cell matrix multiply, using:
 *	     AB^T + BA^T = (A|B)(B^T|A^T) 	or A^TB + AB^T = (B^T|A^T)(A|B)
 *	   In this sense, the cell BLAS _SYR2K() is never utilized, but if 
 *	   d_P(CDM)==1, there is some added overhead of workpace copying
 *	   which could have been avoided had cell BLAS _SYR2K() been used.
 *	   Note however that functionality of the cell BLAS _SYR2K() is 
 *	   inadequate for the parallel case.
 */
#define lA DM_Loc(A)
#define lB DM_Loc(B)
#define lC DM_Loc(C)
#define CDM DM_DM(C)
#define CDN DM_DN(C)  
{ int k, MAXdK; 			/* max size of spread across K-dim */
  int Trans = DM_Trans(A);
  DistMatStore AK, BK;
  
  COMMON_ASSERT_TR_E(
    (TraceStr, "(%dx%dx%d,%1.1f,a%d %d %d,a%d %d %d,%1.1f,%c a%d %d %d)",
      N, N, K, pFD_VAL(alpha), FD_ADR(LM_ADR(lA)), LM_RInc(lA), LM_CInc(lA),
      FD_ADR(LM_ADR(lB)), LM_RInc(lB), LM_CInc(lB), pFD_VAL(beta),
      FormC, FD_ADR(LM_ADR(lC)), LM_RInc(lC), LM_CInc(lC)),
    N>=0 && K>=0 && DM_Trans(A)==DM_Trans(B) && UpLoOK(FormC), 
    0, "DistSYR2KMult", 1, TRACE_ALL
  );

  if (Trans^DM_Trans(C)) {
    DM_Transpose(C);			FormC = TransForm(FormC);
  }

  if (!pFD_UNITY(beta))
    TriMatScale(G2Ld(N, DM_DM(C)), G2Ld(N, DM_DN(C)), beta, FormC, C);

  if (K==0 || (FD_EQ(*alpha, FD_0)) || N==0) {
    TRACE_RETURN;
  }
  
  MAXdK=MIN(PAR_INNER_VECSIZE, K);

  k=0;
  while (k < K) { 
    /* C += alpha * (A[:,K..+dK]*B^T[K..+dK,:] + B[:,K..+dK]*A^T[:,K..+dK]) */
    int dK = MIN(MAXdK, K-k);
    int m, n;				/* local sizes for C */
    LocMat AsK, BsK, BTsK, ATsK;	/* spread workspaces of A, B */
    int AsKn, BsKn, BTsKn, ATsKn;
    DistMat At = Trans? Trans(A): A,	Bt = Trans? Trans(B): B;        
    DistMat AT, BT;			DistMat_RVAlloc(2);
          
    int MergeAB = (dK<=PAR_INNER_VECSIZE/2), MergeBTAT = MergeAB;
      
    LM_ADR(AsK) = LM_ADR(BsK) = NULL;
    if (MergeAB) { 			/* try to make AsK, BsK mergeable */
      int ForceWS = !Mergeable(1-Trans, dK, DM_Loc(At), DM_Loc(Bt)) || 
      		    d_P0(DM_D(1-Trans, A))!=d_P0(DM_D(1-Trans, B));
      /*nb. val(ForceWS) only matters if this cell is sole source for spreads */
      int AsKSz = SizeSpreadWS(1-Trans, dK, N, At, CDM, ForceWS, N, CDN), 
          BsKSz = SizeSpreadWS(1-Trans, dK, N, Bt, CDM, ForceWS, N, CDN);
      if (AsKSz != BsKSz || 		/* exactly 1 of A,B not aligned CDM */
          AsKSz <= 0)			/* sole source && lA,lB mergeable */ 
        MergeAB = 0; 
      else {
        MergeAB = MatMalloc_F("DistSYR2KMult", 0, &LM_ADR(AsK), AsKSz+BsKSz);
        if (MergeAB) 
          LM_ADR(BsK) = LM_ADR(AsK) + AsKSz;
      }  
    }        
    AsKn = SpreadMat(1-Trans, dK, N, At, &FD_1, CDM, &AsK, MergeAB, N, CDN);
    BsKn = SpreadMat(1-Trans, dK, N, Bt, &FD_1, CDM, &BsK, MergeAB, N, CDN);
    m = ABS(AsKn);  
    BT = Spread2DistMat(Trans, 1, CDM, &BsK);   
    AT = Spread2DistMat(Trans, 1, CDM, &AsK);
    LM_Trans(Trans, AsK);		/* make m x dK */
    LM_Trans(Trans, BsK);

    MergeBTAT = MergeBTAT && Mergeable(1, dK, AsK, BsK);
    LM_ADR(BTsK) = LM_ADR(ATsK) = NULL;
    if (MergeBTAT) {
      /* nb. AT, BT must be same except for the local array address; 
             thus BTsKn=ATsKn */
      int BTsKSz = SizeSpreadWS(Trans, dK, N, BT, CDN, 1, N, CDM);
      if (BTsKSz<=0) 
        MergeBTAT = 0;
      else {	
        MergeBTAT = MatMalloc_F("DistSYR2KMult", 0, &LM_ADR(BTsK), 2*BTsKSz);
        if (MergeBTAT)
          LM_ADR(ATsK) = LM_ADR(BTsK) + BTsKSz;
      }
    }                    
    BTsKn = SpreadMat(Trans, dK, N, BT, &FD_1, CDN, &BTsK, MergeBTAT, N, CDM);
    ATsKn = SpreadMat(Trans, dK, N, AT, &FD_1, CDN, &ATsK, MergeBTAT, N, CDM);
    LM_Trans(Trans, ATsK);		/* make dK x n */
    LM_Trans(Trans, BTsK);
    n = ABS(ATsKn);
    
    if (MergeBTAT) {
      TriMatPartMult(m, n, 2*dK, alpha, &AsK, &BTsK, NULL, FormC, C);
      MatPrint("ATsK", 0, 0, ATsK);
      COMMON_ASSERT(Mergeable(0,dK,BTsK,ATsK), 0, "DistSYR2KMult: merge", 6, TRACE_ALL);
    } else {
      TriMatPartMult(m, n, dK, alpha, &AsK, &BTsK, NULL, FormC, C);
      TriMatPartMult(m, n, dK, alpha, &BsK, &ATsK, NULL, FormC, C);
    }
    if (PrintRes) 
      MatPrint("C", MIN(PrintRes, m), MIN(PrintRes, n), LocMatTr(Trans(C), lC));

    if (!MergeBTAT && ATsKn>0)
      MatFree(LM_ADR(ATsK));        
    if (MergeBTAT || BTsKn>0)
      MatFree(LM_ADR(BTsK));    
    if (!MergeAB && BsKn>0)  
      MatFree(LM_ADR(BsK));  
    if (MergeAB || AsKn>0)
      MatFree(LM_ADR(AsK));
      
    k += dK;
    if (k < K) {
      A = SubMat_(&AK, A, 0, dK);	B = SubMat_(&BK, B, 0, dK);
    }  
  } /* while (k...) */

  TRACE_EXIT;
} /*DistSYR2KMult*/

#undef lA 
#undef lB 
#undef lC 
